import hashlib
import json
import uuid

import redis
from itsdangerous import TimedJSONWebSignatureSerializer
from flask import request
from sqlalchemy import text
from sqlalchemy.orm import noload
from sqlalchemy.orm.exc import NoResultFound

from tic.auth.auth_configuration import AuthConfiguration
from tic.auth.auth_db import AuthDbRepository
from tic.auth.auth_util import AuthUtil
from tic.shared.service import TICService, AuthException
from tic.auth.db_model import (
    UserSchema,
    UserCredentialSchema,
    AuthTokenSchema,
    UserCredential,
    CustomerServer,
    User,
)
from tic.shared.aws.kms_client import KmsClient
from tic.shared.utils.util import to_bool


# Notes:
# With auth running on local VM and talking to db in AWS dev:
# Getting customer auth token - takes about 350 ms.
# Getting user auth token - takes about 200 ms.

# Uncomment the following to see the queries that are being executed and the time taken for each execution
# See: http://docs.sqlalchemy.org/en/latest/faq/performance.html

# from sqlalchemy import event
# from sqlalchemy.engine import Engine
# from tic.shared.utils import log_util

# logger = log_util.get_logger("auth", 'DEBUG')
#
# @event.listens_for(Engine, "before_cursor_execute")
# def before_cursor_execute(conn, cursor, statement,
#                         parameters, context, executemany):
#     conn.info.setdefault('query_start_time', []).append(time.time())
#     logger.debug("Start Query: %s" % statement)
#
# @event.listens_for(Engine, "after_cursor_execute")
# def after_cursor_execute(conn, cursor, statement,
#                         parameters, context, executemany):
#     total = time.time() - conn.info['query_start_time'].pop(-1)
#     logger.debug("Query Complete!")
#     logger.debug("Total Time: %f" % total)


class AuthService(TICService):
    def __init__(self, config, principal):
        super(AuthService, self).__init__(config, principal=principal)
        self.auth_util = AuthUtil(logger=self.logger, boto_session=self.boto_session)
        self.auth_repository = None
        self.auth_token_key = None
        self.redis_conn = None
        self.setup()
        self._setup_routes()

    def setup(self):
        aws_region = self.config.get("auth", "aws_region")

        if aws_region is None:
            raise Exception("Failed to get config value for aws_region")

        kms_client = KmsClient(self.boto_session)

        self.auth_token_key = kms_client.decrypt_string(
            self.config.get("auth", "encrypted_token_key")
        )

        self.auth_repository = AuthDbRepository(
            self.logger,
            self.config.auth_conn_str,
            connection_args={},
            statement_timeout=self.config.statement_timeout_ms,
        )

        cache_host = self.config.get("cache", "redis_host")
        if self.config.commit_sha is not None and len(self.config.commit_sha) > 0:
            cache_host = cache_host.format(self.config.commit_sha)

        # connection to redis cache
        self.redis_conn = redis.StrictRedis(
            cache_host,
            self.config.get("cache", "redis_port"),
            ssl=to_bool(self.config.get("cache", "redis_encrypt", default=False)),
        )

    def _setup_routes(self):
        self.app.add_url_rule(
            "/v1/customer-auth-token",
            view_func=self.get_customer_auth_token,
            methods=["GET"],
        )
        self.app.add_url_rule(
            "/v1/user-auth-token", view_func=self.get_user_auth_token, methods=["GET"]
        )
        self.app.add_url_rule(
            "/v1/api-key-auth-token/<api_key>",
            view_func=self.get_api_key_auth_token,
            methods=["GET"],
        )

        self.app.add_url_rule(
            "/v1/users", view_func=self.handle_users, methods=["POST", "GET"]
        )
        self.app.add_url_rule(
            "/v1/users/<int:id>",
            view_func=self.handle_user,
            methods=["GET", "PUT", "DELETE"],
        )

        self.app.add_url_rule(
            "/v1/users/<int:user_id>/credentials",
            view_func=self.create_user_credentials,
            methods=["POST"],
        )
        self.app.add_url_rule(
            "/v1/users/<int:user_id>/credentials/<user_key>",
            view_func=self.update_user_credentials,
            methods=["PUT"],
        )
        self.app.add_url_rule(
            "/v1/credentials/<user_key>/<user_secret>",
            view_func=self.validate_user_credentials,
            methods=["GET"],
        )
        self.app.add_url_rule(
            "/v1/credentials/<user_key>",
            view_func=self.get_user_credentials,
            methods=["GET"],
        )

        self.app.add_url_rule(
            "/v1/apikey/customer/<customer_token>",
            view_func=self.get_user_api_key,
            methods=["GET"],
        )
        self.app.add_url_rule(
            "/v1/apikey/key/<api_key>", view_func=self.get_api_key, methods=["GET"]
        )

        self.app.add_url_rule(
            "/v1/customer-token/<value>",
            view_func=self.get_customer_token,
            methods=["GET"],
        )

        # Expect to find one of the following in the query parameters:
        #   (1) customer token
        #   (2) user-key and user-secret
        self.app.add_url_rule(
            "/v1/cache", view_func=self.get_token_from_cache, methods=["GET"]
        )
        self.app.add_url_rule(
            "/v1/cache", view_func=self.delete_token_from_cache, methods=["DELETE"]
        )

        # self.app.add_url_rule('/v1/credentials', view_func=self.create_user_credentials, methods=['POST'])
        # validate_credentials will be called before each request
        # self.app.before_request(self.validate_credentials)
        return self.app

    def get_customer_token(self, value):
        http_code = 200
        result = {"status": "Success"}

        try:
            if not value:
                raise Exception("Missing customer token value")

            with self.auth_repository.auth_session() as session:
                user = self.auth_repository.get_user_with_customer_token(value, session)
                if user:
                    result["customer_token"] = value
                else:
                    result["message"] = "Cannot find customer token {0}".format(value)
                    http_code = 404

        except Exception as e:
            self.logger.exception("Failed to retrieve customer token.")
            http_code = 500
            result = {"status": "Failed", "message": str(e)}

        return jsonify_types(result), http_code

    def get_customer_auth_token(self):
        """
        The request must have the following query parameters:
        customer-token and an optional server-token

        Return opaque token with user id, permissions, roles, etc.
        The token expiry is specified in the config.

        :return:

        :http_code 200: OK
        :http_code 404: User not found
        :http_code 500: Internal Server Error
        """

        # Check for customer-token, server-token
        customer_token = request.args.get("customer-token")
        server_token = request.args.get("server-token")
        impersonate_customer_token = request.args.get("impersonate-customer-token")

        token, http_code = self.get_auth_token_for_customer(
            customer_token, server_token, impersonate_customer_token
        )

        return jsonify_types({"token": token}), http_code

    def get_api_key_auth_token(self, api_key):
        """
        The request must have the following query parameters:
        api-token

        Return opaque token with user id, permissions, roles, etc.
        The token expiry is specified in the config.

        :return:

        :http_code 200: OK
        :http_code 404: User not found
        :http_code 500: Internal Server Error
        """

        # Check for customer-token, server-token
        self.logger.debug("validating api key")
        token = {}
        key, customer_token, http_code = self.validate_api_key(api_key)

        self.logger.debug("validated api key")
        if http_code == 200:
            token, http_code = self.get_auth_token_for_customer(customer_token)
            self.logger.debug("collected auth token")

        return jsonify_types({"token": token}), http_code

    def get_auth_token_for_customer(
        self, customer_token, server_token=None, impersonate_customer_token=None
    ):
        """
        The request must have the following query parameters:
        api_key

        Return opaque token with user id, permissions, roles, etc.
        The token expiry is specified in the config.

        :return:

        :http_code 200: OK
        :http_code 404: User not found
        :http_code 500: Internal Server Error
        """
        http_code = 200
        token = {}

        self.logger.debug(
            "Retrieving auth token for customer token: {0} and server_token: {1} and impersonate_customer_token: {2}".format(
                customer_token, server_token, impersonate_customer_token
            )
        )

        cache_key = self.get_customer_auth_token_cache_key(
            customer_token, server_token, impersonate_customer_token
        )

        try:
            if customer_token is None:
                raise Exception("Missing customer_token")

            # Retrieve from cache
            auth_token = self.get_auth_token_from_cache(cache_key)
            if auth_token is not None:
                return auth_token, http_code

            self.logger.debug("auth token not cached")

            with self.auth_repository.auth_session() as session:
                # Query the user table to get the user_id and permissions - no need to load the customer_servers.
                db_user = self.auth_repository.get_user_with_customer_token(
                    customer_token, session, noload("customer_servers")
                )
                if not db_user:
                    raise NoResultFound
                self.validate_user(db_user)
                self.logger.debug("validated db user")

                impersonated_by_token = None

                if impersonate_customer_token is not None:
                    impersonated_by_token = customer_token
                    customer_token = impersonate_customer_token
                    impersonate_permissions = list(
                        filter(
                            lambda x: x.action.name == "impersonate"
                            and x.resource.name == "user",
                            db_user.permissions,
                        )
                    )

                    can_impersonate = len(impersonate_permissions) > 0

                    if can_impersonate:
                        db_user = self.auth_repository.get_user_with_customer_token(
                            customer_token, session, noload("customer_servers")
                        )
                    else:
                        http_code = 404

                if http_code == 200:
                    # add customer server token, if it doesn't exist
                    if server_token is not None:
                        self.logger.debug("checking for existing server token")
                        server_token = server_token.strip()

                        # Query with the user_id to see if the customer server token exists
                        found_server_token = self.auth_repository.get_server_token(
                            db_user.id, server_token, session
                        )

                        if not found_server_token:
                            self.logger.debug("adding server token")
                            self.auth_repository.insert_server_token(
                                server_token, db_user, session
                            )
                            self.logger.debug("added server token")

                    expiry_in_min = self.config.get("user_token", "expiry_in_min")
                    serializer = TimedJSONWebSignatureSerializer(
                        self.auth_token_key,
                        expires_in=self.config.cache_expiry_in_seconds
                        + int(expiry_in_min) * 60,
                        algorithm_name="HS256",
                    )
                    user_schema = AuthTokenSchema(
                        exclude=("credentials", "customer_servers")
                    )
                    user = user_schema.dump(db_user)

                    token = serializer.dumps(
                        {
                            "user_id": user["id"],
                            "customer_token": customer_token,
                            "server_token": server_token,
                            "credentials": user,
                            "impersonated_by_token": impersonated_by_token,
                        }
                    )

                    # Cache
                    self.logger.debug("caching auth token")
                    self.cache_auth_token(cache_key, token)
                    self.logger.debug("auth token cached")

        except NoResultFound:  # raised by sqlalchemy when user is not found
            self.logger.exception(
                "Failed to get auth token for customer token: {0} and server_token: {1} and impersonate_customer_token: {2}".format(
                    customer_token, server_token, impersonate_customer_token
                )
            )
            http_code = 404

        except Exception:
            self.logger.exception(
                "Failed to get auth token for customer token: {0} and server_token: {1} and impersonate_customer_token: {2}".format(
                    customer_token, server_token, impersonate_customer_token
                )
            )
            http_code = 500

        return token, http_code

    def get_user_api_key(self, customer_token):
        """
        The request must have the following  parameters:
        customer-token

        Return api key to be used in portal.

        :return:

        :http_code 200: OK
        :http_code 404: User not found
        :http_code 500: Internal Server Error
        """
        http_code = 200
        key = {}

        try:
            # Must find customer token
            if customer_token is None:
                raise Exception("Missing customer token")

            # Query the user table to get the user_id. No need to load relationships.
            with self.auth_repository.auth_session() as session:
                db_user = self.auth_repository.get_user_with_customer_token(
                    customer_token, session, noload("*")
                )

                if not db_user:
                    raise NoResultFound

                if not db_user.enabled:
                    raise Exception(
                        "The user with customer_token {0} is not active".format(
                            customer_token
                        )
                    )

                db_cred = self.auth_repository.get_user_api_key_with_user_id(
                    db_user.id, session, noload("*")
                )

                if not db_cred:
                    api_key = str(uuid.uuid4())
                    self.auth_repository.insert_user_api_key(api_key, db_user, session)
                else:
                    api_key = db_cred.api_key

                key = api_key

        except NoResultFound:  # raised by sqlalchemy when user is not found
            self.logger.exception("get_customer_api_key: Failed to get api_key.")
            http_code = 404

        except Exception:
            self.logger.exception("get_customer_api_key: Failed to get api_key.")
            http_code = 500

        return jsonify_types({"api_key": key}), http_code

    def get_api_key(self, api_key):
        """
        The request must specify the api_key

        Return api token that can be used in portal.

        :return:

        :http_code 200: OK
        :http_code 404: User not found
        :http_code 500: Internal Server Error
        """
        key, customer_token, http_code = self.validate_api_key(api_key)
        return jsonify_types({"api_key": key}), http_code

    def validate_api_key(self, api_key):
        """
        The request must specify the api_key

        Return api token that can be used in portal.

        :return:

        :http_code 200: OK
        :http_code 404: User not found
        :http_code 500: Internal Server Error
        """
        http_code = 200
        key = (None,)
        customer_token = None

        try:
            with self.auth_repository.auth_session() as session:
                customer_api_key = self.auth_repository.get_user_api_key_with_api_key(
                    api_key, session
                )
                self.logger.debug("api key found")
                if not customer_api_key.enabled:
                    raise Exception("The api key {0} is not active".format(api_key))

                if customer_api_key.user is None:
                    raise Exception(
                        "There is no user associated with api key {0}".format(api_key)
                    )

                if not customer_api_key.user.enabled:
                    raise Exception(
                        "The user with id {0} associated with api key {1} is not active".format(
                            customer_api_key.user.id, api_key
                        )
                    )

                key = customer_api_key.api_key
                customer_token = customer_api_key.user.customer_token
                self.logger.debug(
                    f"api key had corresponding customer_token={customer_token}"
                )

        except NoResultFound:  # raised by sqlalchemy when user is not found
            self.logger.exception("get_api_key: Failed to validate api key.")
            http_code = 404

        except Exception:
            self.logger.exception("get_api_key: Failed to validate api key.")
            http_code = 500

        return key, customer_token, http_code

    def get_user_auth_token(self):
        """
        The request must have the following query parameters:
        user-key, user-secret

        Return opaque token with user id, permissions, roles, etc.
        The token expiry is specified in the config.

        :http_code 200: OK
        :http_code 404: User not found
        :http_code 500: Internal Server Error
        """
        http_code = 200
        token = {}
        user_key = None

        try:

            # Check for user-key, user-secret
            user_key = request.args.get("user-key")
            user_secret = request.args.get("user-secret")

            if user_key is None or user_secret is None:
                raise Exception("Missing query parameters")

            self.logger.debug("Retrieving auth token for userkey {0}".format(user_key))

            # Retrieve from cache
            cache_key = self.get_user_key_user_secret_auth_token_cache_key(
                user_key, user_secret
            )
            auth_token = self.get_auth_token_from_cache(cache_key)
            if auth_token is not None:
                if isinstance(auth_token, bytes):
                    auth_token = auth_token.decode()
                return jsonify_types({"token": auth_token}), http_code

            # validate user_key, user_secret
            if not self.is_valid_user_key_and_secret(user_key, user_secret):
                raise AuthException(
                    "Failed to validate user_key {0} and user_secret {1}".format(
                        user_key, user_secret
                    )
                )

            with self.auth_repository.auth_session() as session:
                db_user = self.auth_repository.get_user_with_access_key(
                    user_key, session
                )

                if db_user is None:
                    self.logger.exception(
                        "Cannot find user for user-key {0}".format(user_key)
                    )
                    http_code = 404
                else:
                    self.validate_user(db_user)
                    user_schema = AuthTokenSchema(
                        exclude=("credentials", "customer_servers")
                    )
                    user = user_schema.dump(db_user)
                    expiry_in_min = self.config.get("user_token", "expiry_in_min")
                    serializer = TimedJSONWebSignatureSerializer(
                        self.auth_token_key,
                        expires_in=self.config.cache_expiry_in_seconds
                        + int(expiry_in_min) * 60,
                        algorithm_name="HS256",
                    )

                    token = serializer.dumps(
                        {
                            "user_id": user["id"],
                            "customer_token": user["customer_token"],  # user_key,
                            "server_token": None,
                            "credentials": user,
                        }
                    )

                    # Cache
                    self.cache_auth_token(cache_key, token)

        except AuthException:
            self.logger.exception(
                "Failed to get auth token for userkey {0}".format(user_key)
            )
            http_code = 404

        except Exception:
            self.logger.exception(
                "Failed to get auth token for userkey {0}".format(user_key)
            )
            http_code = 500

        return jsonify_types({"token": token}), http_code

    def handle_users(self):
        """
        ..  http:post:: /users
            Creates an user.

            **Example request**:

            :POST /cb-customers HTTP/1.1
            Host: example.com
            Accept: application/json
            {
                "role-name": cb-customer     (optional. If it is present - a Role with the specified name must be found in the database. Otherwise, error)
                "customer-token": customer-a (required, if role-name is cb-customer)
                "name": name-a               (required if role-name is None or if role-name is not cb-customer)
                "enabled":True               (optional, default = True)
                "create-credentials":False   (optional, default = False. If this is true, a user-key and user-secret will be generated and returned)
            }

            **Example response**:

            .. sourcecode:: http

                HTTP/1.1 200 OK
                Content-Type: application/json
                {
                    "user-id": 1234,
                    "customer-token":customer-a
                    "name": name-a
                    "user-key": generated-user-key
                    "user-secret": generated-user-secret
                    "enabled": True,
                    "date_created": "01/01/2015"
                }

        :return:
            A json object containing the input data

        The following http status codes may be returned from this call:
        :statuscode 200: OK
        :statuscode 400: Bad Request
        :statuscode 500: Internal Server Error
        """
        if request.method == "POST":
            return self.create_user()
        elif request.method == "GET":
            return self.get_users()
        else:
            msg = "Cannot handle request method:%s" % request.method
            return jsonify_types({"reason": msg}), 400

    def create_user(self):
        http_code = 200
        user_data = {}

        try:
            # post_data will be None if content-type is not set to 'application/json'
            post_data = request.get_json()
            if post_data is None:
                raise AuthException(
                    "Missing post data or content-type not set to application/json"
                )

            role_name = post_data.get("role-name")
            customer_token = post_data.get("customer-token")
            server_token = post_data.get("server-token")
            name = post_data.get("name")
            enabled = to_bool(post_data.get("enabled", True))
            description = post_data.get("description")
            create_credentials = post_data.get("create-credentials", False)

            role = None

            with self.auth_repository.auth_session() as session:
                if role_name is not None:
                    role = self.auth_repository.get_role_with_name(role_name, session)

                    if role_name == "cb-customer" and customer_token is None:
                        user_data = {"status": "Missing customer token"}
                        raise AuthException("Missing customer token in post data")

                    if role_name != "cb-customer" and name is None:
                        user_data = {"status": "Missing name token"}
                        raise AuthException("Missing name in post data")

                if customer_token is not None and name is None:
                    name = customer_token

                user = User(
                    name=name,
                    customer_token=customer_token,
                    enabled=enabled,
                    description=description,
                )

                existing_user = self.auth_repository.get_user_with_customer_token_name(
                    customer_token, name, session
                )

                if not existing_user:
                    if server_token is not None:
                        customer_server = CustomerServer(
                            server_token=server_token, user=user
                        )
                        user.customer_servers.append(customer_server)

                    if role is not None:
                        user.roles.append(role)

                    if create_credentials:
                        (
                            access_key_id,
                            secret_access_key,
                            secret_access_key_hash,
                            secret_access_key_b64_encoded,
                        ) = self.auth_util.generate_credentials()
                        user_credential = UserCredential(
                            access_key_id=access_key_id,
                            secret_access_key_hash=secret_access_key_hash,
                        )
                        user.credentials.append(user_credential)

                    if not existing_user:
                        session.add(user)
                        existing_user = (
                            self.auth_repository.get_user_with_customer_token_name(
                                customer_token, name, session
                            )
                        )

                user_schema = UserSchema()
                user_data = user_schema.dump(existing_user)

                self.logger.info(user_data)
                if create_credentials:
                    user_data["user-secret"] = secret_access_key_b64_encoded
                    user_data["user-secret-raw"] = secret_access_key

        except AuthException:
            self.logger.exception("Failed to create cb customer.")
            http_code = 400

        except Exception:
            self.logger.exception("Failed to create cb customer.")
            user_data = {"status": "Failure"}
            http_code = 500

        return jsonify_types(user_data), http_code

    def get_users(self):
        """
        Return users for the specified role-name ordered by the specified users.column

        Query Parameters:
        page_num: Defaults to page 1. Returns empty results - if page_num > maximum pages.
        items-per-page: Defaults to 20
        role-name: Defaults to 'cb-customer'
        order-by: Defaults to 'date_created'

        :return:
            An array of json objects

        The following http status codes may be returned from this call:
        :statuscode 200: OK
        :statuscode 500: Internal Server Error
        """
        http_code = 200
        users = []

        try:
            with self.auth_repository.auth_session() as session:
                role_name = request.args.get("role-name")
                if role_name is None:
                    result = session.execute(
                        "select id, customer_token, name, enabled, description, date_created from users"
                    )
                else:
                    sql = text(
                        """
                                  select u.id, u.customer_token, u.name, u.enabled, u.description, u.date_created
                                  from users u
                                  join user_role ur on ur.user_id = u.id
                                  join roles r on r.id = ur.role_id
                                  where r.name = :role_name;
                               """
                    )
                    result = session.execute(sql, params={"role_name": role_name})

                for row in result:
                    users.append(
                        {
                            "id": row["id"],
                            "name": row["name"],
                            "customer_token": row["customer_token"],
                            "enabled": row["enabled"],
                            "description": row["description"],
                            "date_created": str(row["date_created"]),
                        }
                    )

                return json.dumps(users), http_code

        except Exception:
            self.logger.exception("Failed to get_users")
            http_code = 500

        return json.dumps(users), http_code

    def create_user_credentials(self, user_id):
        """
        Create and store user credentials (key/secret) for the specified user_id.
        Return the generated credentials

        :param user_id:
        :return:
            user credentials
        """
        http_code = 200
        credentials = {}

        try:
            with self.auth_repository.auth_session() as session:
                # lookup the user
                db_user = self.auth_repository.get_user_with_id(user_id, session)

                if db_user is None:
                    raise AuthException("Cannot find user with id: %d" % user_id)

                credentials = self._create_user_credentials(db_user, session)

        except AuthException:
            self.logger.exception("Failed to create user credentials.")
            http_code = 400

        except Exception:
            self.logger.exception("Failed to create user credentials.")
            http_code = 500

        return jsonify_types(credentials), http_code

    def _create_user_credentials(self, db_user, session):
        (
            access_key_id,
            secret_access_key,
            secret_access_key_hash,
            secret_access_key_b64_encoded,
        ) = self.auth_util.generate_credentials()

        user_credential = self.auth_repository.insert_user_credential(
            access_key_id, secret_access_key_hash, db_user, session
        )

        credential_schema = UserCredentialSchema()
        credentials = credential_schema.dump(user_credential)
        credentials["secret_access_key"] = secret_access_key_b64_encoded

        return credentials

    def update_user_credentials(self, user_id, user_key):
        http_code = 200
        result = {
            "user_id": user_id,
            "user_key": user_key,
            "status": "Successfully updated",
        }

        try:
            with self.auth_repository.auth_session() as session:
                # lookup the user
                user = self.auth_repository.get_user_with_id(user_id, session)

                if user is None:
                    result["status"] = "User not found"
                    raise AuthException("Cannot find user with id: %d" % user_id)

                db_credential = next(
                    (x for x in user.credentials if x.access_key_id == user_key), None
                )
                if db_credential is None:
                    result["status"] = "Cannot find user-key"
                    raise AuthException(
                        "Cannot find user key {0} for user id {1}".format(
                            user_key, user_id
                        )
                    )

                # get query parameters
                enabled = request.args.get("enabled", default="")
                description = request.args.get("description")

                update = False

                if enabled.lower() in ["true", "false"]:
                    if enabled.lower() == "true":
                        enabled = True
                    else:
                        enabled = False

                    db_credential.enabled = enabled
                    update = True

                if description is not None:
                    db_credential.description = description
                    update = True

                if update:
                    session.add(db_credential)
                else:
                    result["status"] = "Nothing to update"

        except AuthException:
            self.logger.exception("Failed to create user credentials.")
            http_code = 400

        except Exception:
            self.logger.exception("Failed to create user credentials.")
            http_code = 500

        return jsonify_types(result), http_code

    def validate_user_credentials(self, user_key, user_secret):
        http_code = 200
        validated = {"valid": False}

        try:
            validated["valid"] = self.is_valid_user_key_and_secret(
                user_key, user_secret
            )
            validated["user-key"] = user_key
            validated["user-secret"] = user_secret
        except AuthException:
            self.logger.exception("Failed to validate user credentials.")
            http_code = 404

        except Exception:
            self.logger.exception("Failed to validate user credentials.")
            http_code = 500

        return jsonify_types(validated), http_code

    def is_valid_user_key_and_secret(self, user_key, user_secret):

        if user_key is None or user_secret is None:
            raise AuthException("Missing user-key or user-secret or both")

        with self.auth_repository.auth_session() as session:
            user_credential = self.auth_repository.get_user_credential_with_access_key(
                user_key, session
            )
            if user_credential is None:
                raise AuthException(
                    "Cannot find credentials for user-key {}".format(user_key)
                )
            if not user_credential.enabled:
                raise AuthException("User key {0} is disabled".format(user_key))

            is_valid = self.auth_util.validate_client_access_keys_given_hash(
                access_key_id=user_key,
                secret_access_key=user_secret,
                secret_access_key_hash=user_credential.secret_access_key_hash,
            )

            return is_valid

    def get_user_credentials(self, user_key):
        http_code = 200
        credential_data = {}

        try:
            if user_key is None:
                raise AuthException("Missing user-key")

            with self.auth_repository.auth_session() as session:
                user_credential = (
                    self.auth_repository.get_user_credential_with_access_key(
                        user_key, session
                    )
                )
                if user_credential is None:
                    raise AuthException(
                        "Cannot find credentials for user-key %s" % user_key
                    )

                credential_schema = UserCredentialSchema()
                credential_data = credential_schema.dump(user_credential)

        except AuthException:
            self.logger.exception("Failed to validate user credentials.")
            http_code = 400

        except Exception:
            self.logger.exception("Failed to validate user credentials.")
            http_code = 500

        return jsonify_types(credential_data), http_code

    def handle_user(self, id):
        if request.method == "GET":
            return self.get_user(id)
        elif request.method == "PUT":
            return self.update_user(id)
        elif request.method == "DELETE":
            return self.delete_user(id)
        else:
            msg = "Cannot handle request method:%s" % request.method
            return jsonify_types({"reason": msg}), 400

    def get_user(self, id):
        http_code = 200
        user = {"id": id, "status": "not found"}

        try:
            with self.auth_repository.auth_session() as session:
                db_user = self.auth_repository.get_user_with_id(id, session)
                if db_user is not None:
                    user_schema = UserSchema()
                    user = user_schema.dump(db_user)

                # query = get_debug_queries()

        except Exception:
            self.logger.exception("Failed to get user with id {0}.".format(id))
            user = {"id": id, "status": "failed to retrieve"}
            http_code = 500

        return jsonify_types(user), http_code

    def update_user(self, id):
        http_code = 200
        user = {}

        try:
            put_data = request.get_json()
            if put_data is None:
                raise AuthException(
                    "Missing post data or content-type not set to application/json"
                )
            with self.auth_repository.auth_session() as session:
                db_user = self.auth_repository.get_user_with_id(id, session)
                if db_user is None:
                    raise AuthException("Cannot find user with id: {0}".format(id))

                if "enabled" in put_data:
                    db_user.enabled = put_data["enabled"]
                if "description" in put_data:
                    db_user.description = put_data["description"]
                if "customer-token" in put_data:
                    db_user.customer_token = put_data["customer-token"]
                if "name" in put_data:
                    db_user.name = put_data["name"]

                session.add(db_user)

                user_schema = UserSchema()
                user = user_schema.dump(db_user)

        except Exception:
            self.logger.exception("Failed to update user with id {0}.".format(id))
            user = {"id": id, "status": "failed to update"}
            http_code = 500

        return jsonify_types(user), http_code

    def delete_user(self, id):
        http_code = 200
        user = {}

        try:
            with self.auth_repository.auth_session() as session:
                db_user = self.auth_repository.get_user_with_id(id, session)
                if db_user is None:
                    raise AuthException("Cannot find user with id: {0}".format(id))

                session.delete(db_user)

            user = {"id": id, "status": "deleted"}

        except Exception:
            self.logger.exception("Failed to delete user with id {0}.".format(id))
            user = {"id": id, "status": "failed to delete"}
            http_code = 500

        return jsonify_types(user), http_code

    def get_auth_token_from_cache(self, cache_key):
        auth_token = None
        try:
            # auth_token = self.config.cache.get(cache_key)
            auth_token = self.redis_conn.get(cache_key)
        except Exception:
            self.logger.exception(
                "Failed to get auth token from redis at {0}:{1} for key {2}".format(
                    self.config.get("cache", "redis_host"),
                    self.config.get("cache", "redis_port"),
                    cache_key,
                )
            )

        return auth_token

    def cache_auth_token(self, cache_key, auth_token):
        try:
            # self.config.cache[cache_key] = auth_token
            self.redis_conn.setex(
                name=cache_key,
                value=auth_token,
                time=self.config.cache_expiry_in_seconds,
            )
        except Exception:
            self.logger.exception(
                "Failed to cache auth token to redis at {0}:{1} for for key {2}".format(
                    self.config.get("cache", "redis_host"),
                    self.config.get("cache", "redis_port"),
                    cache_key,
                )
            )

    def get_customer_auth_token_cache_key(
        self, customer_token, server_token=None, impersonate_customer_token=None
    ):
        st = server_token if server_token else ""
        ict = impersonate_customer_token if impersonate_customer_token else ""
        return customer_token + ":" + st + ":" + ict

    def get_user_key_user_secret_auth_token_cache_key(self, user_key, user_secret):
        return hashlib.sha256(
            (user_key + ":" + user_secret).encode("utf-8")
        ).hexdigest()

    def validate_user(self, db_user):
        if not db_user.enabled:
            raise Exception(
                "The user with customer token {0} is not enabled".format(
                    db_user.customer_token
                )
            )

    def get_token_from_cache(self):
        http_code = 200
        tokens = []
        found_in_cache = False

        try:
            user_key = request.args.get("user-key")
            user_secret = request.args.get("user-secret")
            customer_token = request.args.get("customer-token")

            if customer_token:
                cache_key = self.get_customer_auth_token_cache_key(
                    customer_token=customer_token
                )
                keys = self.redis_conn.keys(cache_key + "*")

            elif user_key and user_secret:
                cache_key = self.get_user_key_user_secret_auth_token_cache_key(
                    user_key, user_secret
                )
                keys = self.redis_conn.keys(cache_key)
            else:
                raise Exception(
                    "Failed to find wither [customer-token] or [user-key, user-secret] query parameters"
                )

            if len(keys) == 0:
                return (
                    jsonify_types({"tokens": tokens, "found-in-cache": found_in_cache}),
                    http_code,
                )

            for key in keys:
                tokens.append(self.redis_conn.get(key))

            found_in_cache = True

        except Exception:
            self.logger.exception("get_token_from_cache: Failed to get auth token.")
            http_code = 500

        return (
            jsonify_types({"tokens": tokens, "found-in-cache": found_in_cache}),
            http_code,
        )

    def delete_token_from_cache(self):
        """
        If customer-token is specified, all keys with that customer token are deleted.
        If user-key/user-secret
        :return:
        """
        http_code = 200
        keys_deleted = []
        error_msg = None

        try:
            user_key = request.args.get("user-key")
            user_secret = request.args.get("user-secret")
            customer_token = request.args.get("customer-token")

            if (user_key and not user_secret) or (user_secret and not user_key):
                raise Exception(
                    "Found either user-key {0} or user-secret {1} but not both".format(
                        user_key, user_secret
                    )
                )

            if customer_token:
                cache_key = self.get_customer_auth_token_cache_key(
                    customer_token=customer_token
                )
                keys = self.redis_conn.keys(cache_key + "*")
                if len(keys) == 0:
                    self.logger.debug(
                        "No cache keys found for customer token {0}".format(
                            customer_token
                        )
                    )

            elif user_key and user_secret:
                cache_key = self.get_user_key_user_secret_auth_token_cache_key(
                    user_key, user_secret
                )
                keys = self.redis_conn.keys(cache_key)
                if len(keys) == 0:
                    self.logger.debug(
                        "No cache keys found for user-key {0} and user-secret {1}".format(
                            user_key, user_secret
                        )
                    )
            else:
                key_count = self.redis_conn.dbsize()
                self.redis_conn.flushdb()
                self.logger.debug(
                    "Did not find user-key/user-secret  or customer-token. Removing all - {0} keys from cache.".format(
                        key_count
                    )
                )
                return (
                    jsonify_types(
                        {"num_keys_deleted": key_count, "error_msg": error_msg}
                    ),
                    http_code,
                )

            for key in keys:
                # result should be 1
                result = self.redis_conn.delete(key)
                if result == 1:
                    keys_deleted.append(key)
                else:
                    raise Exception("Failed to delete key: {0}".format(key))

        except Exception as e:
            self.logger.exception("Failed to delete token from cache.")
            error_msg = str(e)
            http_code = 500

        return (
            jsonify_types({"keys_deleted": keys_deleted, "error_msg": error_msg}),
            http_code,
        )


def jsonify_types(obj):
    if isinstance(obj, bytes):
        obj = obj.decode()
    elif isinstance(obj, dict):
        for key, val in obj.items():
            obj[key] = jsonify_types(val)
    elif isinstance(obj, list):
        for i, item in enumerate(obj):
            obj[i] = jsonify_types(item)
    return obj


def start_service(port):
    config = AuthConfiguration(port)
    service = AuthService(config, principal=None)
    return service.app
