from contextlib import contextmanager
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from tic.auth.db_model import (
    User,
    UserCredential,
    UserApiKey,
    CustomerServer,
    Role,
)


class AuthDbRepository(object):
    def __init__(
        self,
        logger,
        conn_str=None,
        connection_args=None,
        statement_timeout=0,
    ):
        statement_timeout = int(statement_timeout)
        connection_args.setdefault(
            "options", f"-c statement_timeout={statement_timeout}"
        )

        self.logger = logger

        if conn_str:
            # Records connections (generic sql only, but returns dicts):
            self.db = sessionmaker(
                bind=create_engine(conn_str, connect_args=connection_args)
            )

    @contextmanager
    def auth_session(self):
        """
        Provide a transactional scope around a writer instance.
        """
        session = scoped_session(self.db)
        try:
            yield session
            session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

    def get_user_with_access_key(self, access_key_id, session):
        """
        @param access_key_id: The access key id of the user that is being queried for
        @param session: a scoped session
        @return: the matched user object correlating to the provided user_key if exists
        """
        result = (
            session.query(User)
            .join(UserCredential.user)
            .filter(UserCredential.access_key_id == access_key_id)
            .first()
        )

        return result

    def get_user_credential_with_access_key(self, access_key_id, session):
        """
        @param access_key_id: The access key id of the user ccccccccccccccredential that is being queried for
        @param session: a scoped session
        @return: the matched user credential object correlating to the provided user_key if exists
        """
        result = (
            session.query(UserCredential)
            .filter_by(access_key_id=access_key_id)
            .one_or_none()
        )

        return result

    def get_user_with_id(self, user_id, session):
        """
        @param user_id: The id of the user that is being queried for
        @param session: a scoped session
        @return: the matched user object correlating to the provided user_key if exists
        """
        result = session.query(User).filter_by(id=user_id).first()

        return result

    def get_user_with_customer_token(self, customer_token, session, *opts):
        """
        @param customer_token: The customer_token of the user that is being queried for
        @param session: a scoped session
        @param opts: options to query with
        @return: the matched user object correlating to the provided user_key if exists
        """
        if opts:
            result = (
                session.query(User)
                .filter_by(customer_token=customer_token)
                .options(*opts)
                .one_or_none()
            )
        else:
            result = (
                session.query(User)
                .filter_by(customer_token=customer_token)
                .one_or_none()
            )

        return result

    def get_user_with_customer_token_name(self, customer_token, name, session):
        """
        @param customer_token: The customer_token of the user that is being queried for
        @param session: a scoped session
        @param opts: options to query with
        @return: the matched user object correlating to the provided user_key if exists
        """
        result = (
            (session.query(User).filter_by(customer_token=customer_token).first())
            if customer_token
            else (session.query(User).filter_by(name=name).first())
        )

        return result

    def get_role_with_name(self, role_name, session):
        """
        @param role_name: name of the role that's being queried for
        @param session: a scoped session
        @return: role object
        """
        result = session.query(Role).filter_by(name=role_name).one()
        return result

    def get_user_api_key_with_user_id(self, user_id, session, *opts):
        """
        @param user_id: user_id associate with the desired api_key
        @param session: a scoped session
        @param opts: options to query with
        @return:
        """
        if opts:
            result = (
                session.query(UserApiKey)
                .filter_by(user_id=user_id)
                .options(*opts)
                .first()
            )
        else:
            result = session.query(UserApiKey).filter_by(user_id=user_id).first()

        return result

    def get_user_api_key_with_api_key(self, api_key, session, *opts):
        """
        @param api_key: the value of the api_key
        @param session: a scoped session
        @param opts: options to query with
        @return:
        """
        if opts:
            result = (
                session.query(UserApiKey)
                .filter_by(api_key=api_key)
                .options(*opts)
                .first()
            )
        else:
            result = session.query(UserApiKey).filter_by(api_key=api_key).first()

        return result

    def get_server_token(self, user_id, server_token, session):
        """
        @param user_id: the id of the user associated with the server token
        @param server_token: the value of the server token
        @param session: a scoped session
        @return: the server token object if exists, else none
        """
        result = (
            session.query(CustomerServer)
            .filter_by(user_id=user_id, server_token=server_token)
            .one_or_none()
        )
        return result

    def insert_server_token(self, server_token, user, session):
        """
        @param server_token: generated server token
        @param user: the attached user for this server token
        @param session: a scoped session
        @return:
        """
        customer_server = CustomerServer(server_token=server_token, user=user)
        session.add(customer_server)

    def insert_user_api_key(self, api_key, user, session):
        """
        @param api_key: the value of the api_key
        @param user: the attached user for this api_key
        @param session: a scoped session
        @return:
        """
        customer_api_key = UserApiKey(api_key=api_key, user=user)
        session.add(customer_api_key)

    def insert_user(self, user, session):
        """
        @param user: the user object to insert
        @param session: a scoped session
        @return:
        """
        session.add(user)

    def insert_user_credential(
        self, access_key_id, secret_access_key_hash, user, session
    ):
        """
        @param access_key_id: id of the access key
        @param secret_access_key_hash: secret hash generated paired with the access_key_id
        @param user: the associated user
        @param session: a scoped session
        @return:
        """
        user_credential = UserCredential(
            access_key_id=access_key_id,
            secret_access_key_hash=secret_access_key_hash,
            user=user,
        )
        session.add(user_credential)
        return user_credential
