import inspect
import uuid
from base64 import b64encode, b64decode

from itsdangerous import BadTimeSignature
from itsdangerous import TimedJSONWebSignatureSerializer
from passlib.hash import sha256_crypt

from tic.shared.aws.kms_client import KmsClient
from tic.shared.utils import log_util


class AuthUtil(object):
    def __init__(self, logger=None, boto_session=None):
        if not boto_session:
            raise Exception("Boto Session is None")
        self.boto_session = boto_session
        if logger is None:
            self.logger = log_util.get_logger(__name__)
        else:
            self.logger = logger

    @staticmethod
    def generate_credentials():
        access_key_id = str(uuid.uuid4())
        secret_access_key = str(uuid.uuid4())
        secret_access_key_hash = sha256_crypt.hash(secret_access_key).encode("utf-8")
        secret_access_key_b64_encoded = b64encode(secret_access_key.encode("utf-8"))

        return (
            access_key_id,
            secret_access_key,
            secret_access_key_hash,
            secret_access_key_b64_encoded,
        )

    def validate_client_access_keys_given_hash(
        self, access_key_id, secret_access_key, secret_access_key_hash
    ):
        secret_access_key_decoded = None
        _secret_access_key = secret_access_key.strip()

        try:
            secret_access_key_decoded = b64decode(_secret_access_key)
            valid_client = sha256_crypt.verify(
                secret_access_key_decoded, secret_access_key_hash
            )

        except Exception as e:
            self.logger.exception(
                "Failed to validate credentials for access_key_id:[%s], secret_access_key:[%s] and decoded_secret_access_key:[%s]"
                % (access_key_id, _secret_access_key, secret_access_key_decoded)
            )
            raise e

        return valid_client

    def create_auth_token_encryption_key(self, kms_key_alias, aws_region):
        """
        This will create a symmetric key (guid) and encrypt it using the
        specified kms key alias which should be associated with the specified aws_region

        The encrypted key needs to be stored in TICBaseConfig.

        The original key (guid) is used to encrypt and decrypt the auth token
        See the following google doc for details:
        https://docs.google.com/a/bit9.com/document/d/1GRwwOM_W1JLxx45yUOADVZyo3lwCxGcP7vnkZSiwcB8/edit

        :param kms_key_alias:
        :param aws_region:
        :return:
        """

        try:
            key = str(uuid.uuid4())
            kms_client = KmsClient(self.boto_session)
            encrypted_key = kms_client.encrypt_string(key, kms_key_alias)

        except Exception as e:
            self.logger.exception(
                "Failed to create auth_token encryption key using kms key alias %s for region %s"
                % (kms_key_alias, aws_region)
            )
            raise e

        return encrypted_key

    def get_auth_token_data(self, token, token_key):
        """
        Return user credential data for the user specified by the token
        """
        token_data = None
        try:
            serializer = TimedJSONWebSignatureSerializer(
                token_key, algorithm_name="HS256"
            )
            token_data = serializer.loads(token)
            if "credentials" not in token_data:
                raise Exception("Missing credentials from auth token")

        except BadTimeSignature as e:  # will also catch SignatureExpired
            fq_fn_name = __name__ + ":" + inspect.currentframe().f_code.co_name
            self.logger.exception(fq_fn_name + ": Token signature has expired %s" % e)

        except Exception as e:
            fq_fn_name = __name__ + ":" + inspect.currentframe().f_code.co_name
            self.logger.exception(fq_fn_name + ": Error occurred %s" % e)

        return token_data
