from datetime import datetime
from marshmallow import fields
from marshmallow import Schema
from sqlalchemy import (
    Sequence,
    Integer,
    Column,
    String,
    Boolean,
    DateTime,
    ForeignKey,
    LargeBinary,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship


Base = declarative_base()
metadata = Base.metadata


class User(Base):
    __tablename__ = "users"
    id = Column(Integer, Sequence("users_id_seq"), primary_key=True)
    # Size of a cert common name seems to be 64
    # Search for the string 'ub-common-name-length INTEGER ::= 64' in http://www.ietf.org/rfc/rfc5280.txt
    customer_token = Column(
        String(64), nullable=True
    )  # 64 seems to be the limit on the cert cn
    name = Column(String(50))
    enabled = Column(Boolean, default=True)
    description = Column(String(200))
    date_created = Column(DateTime, default=datetime.utcnow())

    # many to many User<->Role
    roles = relationship("Role", secondary="user_role", backref="users")
    # many to many User<->Permission
    permissions = relationship(
        "Permission", secondary="user_permission", backref="users"
    )
    # one-to-many
    credentials = relationship("UserCredential", cascade="all, delete-orphan")
    # one-to-many
    customer_servers = relationship("CustomerServer", cascade="all, delete-orphan")


class Role(Base):
    __tablename__ = "roles"
    id = Column(Integer, Sequence("roles_id_seq"), primary_key=True)
    name = Column(String(50))
    description = Column(String(200))
    date_created = Column(DateTime, default=datetime.utcnow())

    # many to many Role<->Permission
    permissions = relationship(
        "Permission", secondary="role_permission", backref="roles"
    )


class UserCredential(Base):
    __tablename__ = "user_credentials"
    id = Column(Integer, Sequence("user_credentials_id_seq"), primary_key=True)
    user_id = Column(Integer, ForeignKey("users.id"), nullable=True)
    access_key_id = Column(String(50))

    secret_access_key_hash = Column(LargeBinary)
    date_created = Column(DateTime, default=datetime.utcnow())
    enabled = Column(Boolean, default=True)
    description = Column(String(200))
    user = relationship("User", foreign_keys="UserCredential.user_id")


class Action(Base):
    __tablename__ = "actions"
    id = Column(Integer, Sequence("actions_id_seq"), primary_key=True)
    name = Column(String(50))
    description = Column(String(255))
    date_created = Column(DateTime, default=datetime.utcnow())


class Resource(Base):
    __tablename__ = "resources"
    id = Column(Integer, Sequence("resources_id_seq"), primary_key=True)
    name = Column(String(50))
    description = Column(String(255))
    date_created = Column(DateTime, default=datetime.utcnow())


class Permission(Base):
    __tablename__ = "permissions"
    id = Column(Integer, Sequence("permissions_id_seq"), primary_key=True)
    name = Column(String(50))
    resource_id = Column(Integer, ForeignKey("resources.id"))
    action_id = Column(Integer, ForeignKey("actions.id"))
    description = Column(String(255))
    date_created = Column(DateTime, default=datetime.utcnow())

    resource = relationship("Resource", foreign_keys="Permission.resource_id")
    action = relationship("Action", foreign_keys="Permission.action_id")


class CustomerServer(Base):
    __tablename__ = "customer_servers"
    id = Column(Integer, Sequence("customer_servers_id_seq"), primary_key=True)
    user_id = Column(Integer, ForeignKey("users.id"), nullable=True)
    server_token = Column(String(50))
    description = Column(String(255))
    date_created = Column(DateTime, default=datetime.utcnow())
    enabled = Column(Boolean, default=True)

    user = relationship("User", foreign_keys="CustomerServer.user_id")


class UserRole(Base):
    __tablename__ = "user_role"
    id = Column(Integer, Sequence("user_role_id_seq"), primary_key=True)
    user_id = Column(Integer, ForeignKey("users.id"))
    role_id = Column(Integer, ForeignKey("roles.id"))


class RolePermission(Base):
    __tablename__ = "role_permission"
    id = Column(Integer, Sequence("role_permission_id_seq"), primary_key=True)
    role_id = Column(Integer, ForeignKey("roles.id"))
    permission_id = Column(Integer, ForeignKey("permissions.id"))


class UserPermission(Base):
    __tablename__ = "user_permission"
    id = Column(Integer, Sequence("user_permission_id_seq"), primary_key=True)
    user_id = Column(Integer, ForeignKey("users.id"))
    permission_id = Column(Integer, ForeignKey("permissions.id"))


class UserApiKey(Base):
    __tablename__ = "user_api_key"
    user_id = Column(Integer, ForeignKey("users.id"), primary_key=True)
    api_key = Column(String(50), index=True)

    date_created = Column(DateTime, default=datetime.utcnow())
    enabled = Column(Boolean, default=True)
    user = relationship("User", foreign_keys="UserApiKey.user_id")


# region #### SCHEMA (Marshmallow) ####
# https://marshmallow.readthedocs.org/en/latest/
class ActionSchema(Schema):
    id = fields.Integer()
    name = fields.String()
    description = fields.String(allow_none=True, default=None)
    date_created = fields.DateTime()


class ResourceSchema(Schema):
    id = fields.Integer()
    name = fields.String()
    description = fields.String(allow_none=True, default=None)
    date_created = fields.DateTime()


class PermissionSchema(Schema):
    date_created = fields.DateTime()
    action = fields.Nested(ActionSchema, many=False)
    resource = fields.Nested(ResourceSchema, many=False)


class RoleSchema(Schema):
    id = fields.Integer()
    name = fields.String()
    description = fields.String(allow_none=True, default=None)
    date_created = fields.DateTime()

    permissions = fields.Nested(PermissionSchema, many=True)


class UserCredentialSchema(Schema):
    id = fields.Integer()
    access_key_id = fields.String()
    secret_access_key_hash = fields.String()
    enabled = fields.Boolean()
    description = fields.String(allow_none=True, default=None)
    date_created = fields.DateTime()
    user_id = fields.Integer()


class CustomerServerSchema(Schema):
    id = fields.Integer()
    server_token = fields.String()
    description = fields.String(allow_none=True, default=None)
    date_created = fields.DateTime()
    enabled = fields.Boolean()


class UserSchema(Schema):
    id = fields.Integer()
    name = fields.String()
    customer_token = fields.String(allow_none=True, default=None)
    enabled = fields.Boolean()
    description = fields.String(allow_none=True, default=None)
    date_created = fields.DateTime()

    roles = fields.Nested(RoleSchema, many=True)
    permissions = fields.Nested(PermissionSchema, many=True)
    credentials = fields.Nested(UserCredentialSchema, many=True)
    customer_servers = fields.Nested(CustomerServerSchema, many=True)


class CustomerApiKeySchema(Schema):
    user_id = fields.Integer()
    api_key = fields.String()
    enabled = fields.Boolean()
    date_created = fields.DateTime()


class AuthTokenPermissionSchema(Schema):
    date_created = fields.DateTime()
    action = fields.Nested(ActionSchema, many=False, only=("name",))
    resource = fields.Nested(ResourceSchema, many=False, only=("name",))


class AuthTokenRoleSchema(Schema):
    name = fields.String()

    permissions = fields.Nested(
        AuthTokenPermissionSchema, many=True, only=("action", "resource")
    )


class AuthTokenSchema(Schema):
    id = fields.Integer()
    name = fields.String()
    customer_token = fields.String(allow_none=True, default=None)
    enabled = fields.Boolean()

    roles = fields.Nested(AuthTokenRoleSchema, many=True, only=("name", "permissions"))
    permissions = fields.Nested(
        AuthTokenPermissionSchema, many=True, only=("action", "resource")
    )
    credentials = fields.Nested(UserCredentialSchema, many=True)
    customer_servers = fields.Nested(CustomerServerSchema, many=True)
