import json
from collections import deque

# from multiprocessing import get_context
from enum import Enum
from random import randint
from time import time
from typing import Dict, Set, List, Tuple, Optional

import requests

# Third party imports
# import gevent
from flask import jsonify
from flask import request
from requests.adapters import HTTPAdapter

from tic.defense.defense_config import DefenseConfig
from tic.defense.defense_principal import DefensePrincipal

# Program imports
from .dtos.request_metadata import Metadata
from .iris_api_client import IrisApiClient

# from .queue_worker import QueueWorker, WorkerType
from ..reputation.shared.dtos.reputation_cache_item import ReputationCacheItem
from ..reputation.shared.dtos.reputation_item import ReputationItem
from ..reputation.shared.enums.reputation_channel import ReputationChannel
from ..reputation.shared.enums.reputation_status import ReputationStatus
from ..reputation.shared.repository.dynamodb_repository import ReputationDbRepository
from ..shared.aws.kinesis_client import KinesisClient
from ..shared.dto import Dto
from ..shared.service import TICConfiguration
from ..shared.service import TICService, TicPrincipal, read_defense_data_permission
from ..shared.ubs.dtos.analysis.pipeline_summary import PipelineSummary
from ..shared.ubs.enums.analysis.pipeline import UbsAnalysisPipeline
from ..shared.ubs.enums.analysis.stage import UbsAnalysisStage
from ..shared.ubs.enums.analysis.status import UbsAnalysisStatus
from ..shared.utils.date_util import utc_now
from ..shared.utils.parsers.hash import determine_hash_type, HashType
from ..shared.utils.util import to_bool

# import sys


EXT_DENYLIST = {
    "nsf",
    "pem",
    "hta",
    "html",
    "htm",
    "php",
    "json",
    "pdf",
    "xls",
    "xlm",
    "xlt",
    "xlsx",
    "xltx",
    "xlsm",
    "xltm",
    "xlsb",
    "xlw",
    "ppt",
    "pot",
    "potx",
    "potm",
    "ppsx",
    "ppsm",
    "pps",
    "pptx",
    "pptm",
    "sldx",
    "sldm",
    "doc",
    "docx",
    "docm",
    "docb",
    "rtf",
    "txt",
    "csv",
    "tsv",
    "wbx",
    "pub",
    "xps",
    "oxps",
    "mdm",
    "accdb",
    "accde",
    "accdt",
    "accdr",
    "numbers",
    "pages",
    "key",
    "js",
    "css",
    "sh",
    "rb",
    "py",
    "pyc",
    "pl",
    "ksh",
    "csh",
    "bat",
    "bash",
    "ps",
    "cmd",
    "dot",
}


class QueueStatus(Dto):
    def __init__(self, timestamp, queue_size):
        self.timestamp = timestamp
        self.queue_size = queue_size


class QueueMonitor(Dto):
    encode_settings = {
        "queue_check": lambda x: [y.to_dict(True, True) for y in x],
        "queue_type": lambda x: x.name,
    }

    def __init__(self, queue, queue_type, max_size=100, num_intervals=5):
        self.queue = queue
        self.queue_type = queue_type
        self.max_size = max_size
        self.num_intervals = num_intervals - 1
        self.healthy = True
        self.queue_check = deque(maxlen=self.num_intervals)

    def health_check(self):
        item = QueueStatus(time(), self.queue.qsize())

        # Check history
        if len(self.queue_check) < self.num_intervals:
            self.queue_check.append(item)
        else:
            # Has the queue consistently increased over the last num_intervals?
            count = 0
            items = list(self.queue_check)
            items = items[::-1]
            items.append(item)
            prev_x = items.pop(0).queue_size
            for x in items:
                if x.queue_size > prev_x:
                    count += 1
                prev_x = x.queue_size

            if count == self.num_intervals:
                self.healthy = False
            else:
                self.healthy = True

        # Update the details
        self.queue_check.rotate()
        self.queue_check[0] = item

        # Safety check
        if item.queue_size > self.max_size:
            self.healthy = False
        return self.healthy


class BadRequest(Exception):
    pass


class TimerType(Enum):
    INIT = 0
    GET_REPUTATION = 1
    PUT_REPUTATION = 2
    PUT_APC = 3
    PUT_LOG = 4


class DefenseTimeSplits(Dto):
    def __init__(self):
        self.start()

    def start(self):
        self._start = time()
        self._last_stop = self._start
        self.init = None
        self.get_reputation = None
        self.put_reputation = None
        self.put_apc = None
        self.put_log = None
        self.total = None

    def stop_timer(self, timer_type):
        t = time()
        if timer_type == TimerType.INIT:
            self.init = t - self._last_stop
        elif timer_type == TimerType.GET_REPUTATION:
            self.get_reputation = t - self._last_stop
        elif timer_type == TimerType.PUT_REPUTATION:
            self.put_reputation = t - self._last_stop
        elif timer_type == TimerType.PUT_APC:
            self.put_apc = t - self._last_stop
        elif timer_type == TimerType.PUT_LOG:
            self.put_log = t - self._last_stop
        self._last_stop = t

    def stop(self):
        self._last_stop = time()
        self.total = self._last_stop - self._start

    # Ignore PyCharm
    def to_json(
        self, pretty=False, minimize=False, key_ignorelist=None, ensure_ascii=False
    ):
        return json.dumps(
            self.to_dict(True, True, key_ignorelist={"_start", "_last_stop"}),
            sort_keys=True,
            ensure_ascii=ensure_ascii,
        )


class ReputationLookupInfo:
    def __init__(
        self,
        sha256: str,
        metadata: Metadata,
        lookup_request: Dict,
        pipeline: UbsAnalysisPipeline = None,
        version: Tuple[int] = None,
        invalid_filetype: bool = False,
    ):
        self.sha256 = sha256
        self.metadata = metadata
        self.is_interesting = to_bool(lookup_request.get("is_interesting", False))
        self.is_candidate = to_bool(
            ReputationChannel(lookup_request.get("channel", "stable"))
            == ReputationChannel.CANDIDATE
        )
        self.pipeline = pipeline
        self.reputation_item = ReputationItem(sha256)
        self.cache_item = self.reputation_item.to_reputation_cache_item(
            is_interesting=self.is_interesting
        )
        self.lookup_request = lookup_request
        self.has_completed: Dict[UbsAnalysisStage, bool] = dict()
        self.version = version
        self.invalid_filetype = invalid_filetype


ReputationInfoMap = Dict[str, ReputationLookupInfo]
Sha256Set = Set[str]


class DefenseConfiguration(TICConfiguration):
    def __init__(self, service_port):
        super(DefenseConfiguration, self).__init__("defense", service_port)
        self.logging_stream = None
        self.kinesis_logging_enabled = False
        self.log_apc_reasons_enabled = False
        self.apc_file_size_limit = None
        self.new_reputation_cache_name = None
        self.defense_timeout = 0.3
        self.defense_batch_timeout = 1.0
        self.log_level_name = "INFO"
        self.new_cache = None
        self.submitted_rate = None
        self.ubs_analysis_url = self.get_healthy_service_uri("ubs_analysis", "v1")
        self.endpoint_url = None
        self.route_prefix = ""
        self.iris_api_invocation_enabled = False
        self.iris_api_timeout = 10  # in seconds
        self.iris_service_url = None
        self.iris_token = None
        self.config_changed()

    def config_changed(self):
        # Pull updated config values
        self.new_reputation_cache_name = self.get(
            "defense", "new_defense_reputation_dynamodb"
        )
        self.logging_stream = self.get("defense", "logging_stream")
        old_apc_file_size_limit = self.apc_file_size_limit
        self.apc_file_size_limit = int(self.get("defense", "apc_max_file_size"))
        if self.apc_file_size_limit != old_apc_file_size_limit:
            self.logger.debug(
                "apc max file size set to {} bytes".format(self.apc_file_size_limit)
            )

        old_kinesis_logging_enable = self.kinesis_logging_enabled
        self.kinesis_logging_enabled = to_bool(
            self.get("defense", "kinesis_logging_enabled")
        )
        if self.kinesis_logging_enabled != old_kinesis_logging_enable:
            self.logger.debug(
                "config changed - kinesis logging set to {}".format(
                    self.kinesis_logging_enabled
                )
            )

        self.log_apc_reasons_enabled = to_bool(
            self.get("defense", "log_apc_reasons_enabled")
        )
        self.log_level_name = self.get("logging", "log_level", default="INFO")

        self.submitted_rate = int(self.get("defense", "submitted_rate"))

        try:
            self.iris_api_invocation_enabled = to_bool(
                self.get("defense", "iris_api_invocation_enabled")
            )
            self.iris_api_timeout = int(self.get("defense", "iris_api_timeout"))
            self.iris_token = self.get("defense", "iris_token")
            self.iris_service_url = self.get("defense", "iris_service_url")
        except Exception:
            self.logger.exception("Incorrect Iris configuration")


class DefenseService(TICService):
    def __init__(self, config, principal):
        super(DefenseService, self).__init__(config, principal=principal)
        self.app.config["JSON_AS_ASCII"] = False  # Enable Unicode support
        self._setup_routes()
        self.config = config
        self.timer = DefenseTimeSplits()
        self.log_kinesis_client = None
        if config.iris_api_invocation_enabled:
            try:
                self.iris_client = IrisApiClient(config)
            except Exception:
                self.logger.exception("Unable to create iris client object")
        # self.log_queue = None
        # self.log_queue_worker = None
        # self.queues = None
        # self.queue_workers = None
        # self.queue_monitors = None

        # Use default boto settings - caller will deal with timeout on their own
        self.new_cache = ReputationDbRepository(
            self.config.new_reputation_cache_name,
            session=self.boto_session,
            endpoint_url=self.config.endpoint_url,
        )

        self._set_up_clients()
        # gevent.signal(signal.SIGTERM, self.sigterm_handler)  # Works regardless of queue existence

    # def _setup_queues(self):
    #     # Setup queues
    #     ctx = get_context('spawn')
    #     self.log_queue = ctx.Queue()
    #     self.log_queue_worker = QueueWorker(
    #         self.config, self.log_queue, WorkerType.LOG, daemon=True, log_level_name=self.config.log_level_name)
    #     self.queues = [self.log_queue]
    #     self.queue_workers = [self.log_queue_worker]
    #
    #     # Setup queue monitoring
    #     self.queue_monitors = [QueueMonitor(self.log_queue, WorkerType.LOG)]
    #
    #     # Starting queue workers
    #     for qw in self.queue_workers:
    #         qw.start()
    #
    # def _teardown_queues(self):
    #     if self.queue_workers is None:
    #         self.queues = None
    #         self.log_queue = None
    #         return
    #
    #     # Send poison pill
    #     for q in self.queues:
    #         q.put(None)
    #
    #     # Wait for workers to finish up
    #     for qw in self.queue_workers:
    #         qw.join()

    def _set_up_clients(self):
        self.ubs_analysis_session = requests.Session()
        self.ubs_analysis_session.mount("http://", HTTPAdapter())
        if self.config.kinesis_logging_enabled:
            self.log_kinesis_client = KinesisClient(
                self.boto_session, self.config.logging_stream
            )

    # def sigterm_handler(self, *_):
    #     self.logger.info("Sigterm handler - start")
    #     super(DefenseService, self).sigterm_handler(None, None)
    #     self._teardown_queues()
    #     self.logger.info("Sigterm handler - end")
    #     sys.exit(0)

    # def queue_health_check(self):
    #     for qw in self.queue_workers:
    #         if not qw.is_alive():
    #             self.logger.warning("health_check: worker is not alive, sending SIG_TERM")
    #             os.kill(os.getpid(), signal.SIGTERM)
    #             return False
    #     return True

    # def ready_check(self):
    #     """
    #     Used by Kubernetes in the readiness probe. In Defense - we check the sizes and health of the queues as well.
    #
    #     :statuscode 200: Ready
    #     :statuscode 503: Not Ready
    #     """
    #
    #     r, base_response_code = super().ready_check()
    #     if base_response_code != 200 or not self.config.kinesis_logging_enabled:
    #         return r, base_response_code
    #
    #     sick = False
    #     for qm in self.queue_monitors:
    #         sick |= not qm.health_check()
    #     qms = [qm.to_dict(True, True, key_ignorelist='queue') for qm in self.queue_monitors]
    #
    #     if sick:
    #         self.logger.warning('ready_check: sick {}'.format(json.dumps(qms, sort_keys=True, ensure_ascii=False)))
    #
    #     tic_status.healthy = self.queue_health_check()
    #     tic_status.ready = not sick
    #
    #     if tic_status.healthy is True and tic_status.ready is True:
    #         return jsonify({'ready': 'True'}), 200
    #     else:
    #         return jsonify({'ready': 'False'}), 503

    def _setup_routes(self):
        """
        Adds routes to defense service
        """
        self.app.add_url_rule(
            rule=f"{self.config.route_prefix}/v1/reputation/<sha256>",
            view_func=self.get_reputation,
            methods=["GET"],
        )
        self.app.add_url_rule(
            rule=f"{self.config.route_prefix}/v1/reputation",
            view_func=self.get_reputation_bulk,
            methods=["POST"],
        )

        # So nothing needs to change when hitting the service directly
        self.app.add_url_rule(
            rule=f"{self.config.route_prefix}/api/v1/defense/reputation/<sha256>",
            view_func=self.get_reputation,
            methods=["GET"],
        )
        self.app.add_url_rule(
            rule=f"{self.config.route_prefix}/api/v1/defense/reputation",
            view_func=self.get_reputation_bulk,
            methods=["POST"],
        )
        self.app.add_url_rule(
            rule=f"{self.config.route_prefix}/v2/reputation",
            view_func=self.get_reputation_bulk_v2,
            methods=["POST"],
        )
        self.app.add_url_rule(
            rule=f"{self.config.route_prefix}/api/v2/defense/reputation",
            view_func=self.get_reputation_bulk_v2,
            methods=["POST"],
        )

    @read_defense_data_permission.require(http_exception=403)
    def get_reputation(self, sha256):
        """
        Returns reputation for the requested sha256, submits for calculation if necessary
        :param sha256: the sha256 of the file that needs reputation
        :return: reputation for the hash, else submitted or error message
        """

        self.timer.start()
        is_interesting = to_bool(request.args.get("is_interesting"))
        try:
            lookup_results, http_code = self._attempt_analysis(
                [
                    {
                        "sha256": sha256,
                        "is_interesting": is_interesting,
                        "channel": request.args.get("channel", "stable"),
                    }
                ]
            )
        except BadRequest as e:
            self.logger.warning("Bad request.")
            return jsonify({"status": "failed", "error": str(e)}), 400

        lookup_result = lookup_results[0]
        self.timer.stop()

        if self.timer.total > self.config.defense_timeout:
            self.logger.warning(
                "Timeout exceeded in get_reputation: {}".format(self.timer.to_json())
            )
        else:
            self.logger.debug("GET request time: {0:.3f}".format(self.timer.total))

        return jsonify(lookup_result), http_code

    @read_defense_data_permission.require(http_exception=403)
    def get_reputation_bulk(self):
        """
        Returns reputation for the requested sha256s, submits for calculation if necessary.
        Requires a json list of dicts containing sha256, is_interesting (bool) keys in body.
        :return: a dict with a success/failure message and a list of reputation objects in the 'results' section
        """

        self.timer.start()
        http_code = 200
        lookup_results = []
        return_status = "success"
        response_error = None

        try:
            requests = request.get_json(force=True)  # dicts with sha256, is_interesting
            lookup_results, http_code = self._attempt_analysis(requests)
        except BadRequest as e:
            return_status = "failed"
            response_error = str(e)
            self.logger.warning("Bad request.")
        except Exception:
            return_status = "failed"
            response_error = "Unexpected server error. Please try again later."
            self.logger.exception("Failed to lookup items in cache.")

        self.timer.stop()

        if self.timer.total > self.config.defense_batch_timeout:
            self.logger.warning(
                "Timeout exceeded in put_reputation. Num Items: {}, Details: {}".format(
                    len(lookup_results), self.timer.to_json()
                )
            )
        else:
            self.logger.debug(
                "POST batch size: {}; total time: {:.3f}".format(
                    len(lookup_results), self.timer.total
                )
            )

        response = {"status": return_status, "results": lookup_results}
        if response_error is not None:
            response["error"] = response_error
        self.logger.debug(response)
        return jsonify(response), http_code

    @read_defense_data_permission.require(http_exception=403)
    def get_reputation_bulk_v2(self):
        """
        Returns reputation for the requested sha256,tlsh,publisher, and signature verification status.
        Requires a json list of dicts containing ssha256,tlsh,publisher, and signature verification status in body.
        :return: a dict with a success/failure message and a list of reputation objects in the 'results' section
        """

        self.timer.start()
        http_code = 200
        lookup_results = []
        return_status = "success"
        response_error = None

        try:
            requests = request.get_json(
                force=True
            )  # dicts with sha256,tlsh,publisher, and signature verification status
            lookup_results, http_code = self._attempt_analysis(requests)

            # Make Iris call here
            if self.config.iris_api_invocation_enabled:
                if self.iris_client is not None:
                    response = self.iris_client.query(requests, lookup_results)
                else:
                    self.logger.warning(
                        "Iris invocation is enabled but iris_client is null"
                    )

        except BadRequest as e:
            return_status = "failed"
            response_error = str(e)
            self.logger.warning("Bad request.")
        except Exception:
            return_status = "failed"
            response_error = "Unexpected server error. Please try again later."
            self.logger.exception("Failed to lookup items in cache.")

        self.timer.stop()

        if self.timer.total > self.config.defense_batch_timeout:
            self.logger.warning(
                "Timeout exceeded in put_reputation. Num Items: {}, Details: {}".format(
                    len(lookup_results), self.timer.to_json()
                )
            )
        else:
            self.logger.debug(
                "POST batch size: {}; total time: {:.3f}".format(
                    len(lookup_results), self.timer.total
                )
            )

        response = {"status": return_status, "results": lookup_results}
        if response_error is not None:
            response["error"] = response_error
        self.logger.debug(response)
        return jsonify(response), http_code

    @staticmethod
    def _determine_pipeline_for_hash(
        lookup_info: ReputationLookupInfo,
    ) -> Optional[UbsAnalysisPipeline]:
        """
        Determines the UBS Analysis pipeline to start for a given hash using
        the raw flags passed into reputation APIs
        :param lookup_request: Request args for a given file
        :return: Pipeline to start
        """
        if lookup_info.invalid_filetype:
            return None
        elif lookup_info.is_interesting:
            return UbsAnalysisPipeline.REPUTATION
        elif lookup_info.is_candidate:
            return UbsAnalysisPipeline.REPUTATION_NO_FILE_SHARE
        return None

    @staticmethod
    def _is_valid_file(metadata):
        """
        Determine if a lookup request is for a valid file
        :param metadata: File metadata from request
        :returns: True if the file is valid
        """
        return (metadata.file_name or "").split(".")[-1].lower() not in EXT_DENYLIST

    @staticmethod
    def _get_sha256(lookup_request):
        """
        Determine if the lookup request is for a valid sha256
        :returns: True if the request is for a valid sha256, sha256 passed in
        """
        sha256 = lookup_request["sha256"].lower()
        hash_type, _ = determine_hash_type(sha256)
        return hash_type == HashType.sha256, sha256

    def _get_kinesis_log(self, item: ReputationCacheItem, is_interesting: bool):
        """
        Adds requestor and time of request to response, then formats for s3 log storage through Kinesis/firehose logging
        :param item: The response ReputationCacheItem to format for s3 storage
        :param is_interesting: Whether the item is interesting
        :return: the formatted log entry
        """

        # Get the JSON
        item.requested_by = self.tic_principal.customer_token
        item.requested_at = utc_now().strftime("%Y-%m-%d %H:%M:%S.%f")
        item.is_interesting = is_interesting
        j = item.to_json()

        # Cleanup
        del item.requested_by
        del item.requested_at

        return j

    def _log_interesting_and_candidate_items(
        self, hashes: Sha256Set, info_map: ReputationInfoMap
    ):
        """
        Log hashes for APC analysis to Kinesis
        :param hashes: Hashes which were submitted for APC analysis and/or candidate
        :param info_map: Mapping from sha256 to ReputationLookupInfo
        """
        if not self.config.kinesis_logging_enabled:
            return
        aggregated_logs = list()
        for sha256 in hashes:
            lookup_info = info_map[sha256]
            if (
                lookup_info.is_interesting
                or lookup_info.is_candidate
                or lookup_info.reputation_item.apc is not None
            ):
                aggregated_logs.append(
                    self._get_kinesis_log(
                        lookup_info.cache_item, lookup_info.is_interesting
                    )
                )

        if len(aggregated_logs) > 0:
            self.log_kinesis_client.put_records_aggregated({None: aggregated_logs})
            self.timer.stop_timer(TimerType.PUT_LOG)

    def _validate_hashes(
        self, requests, info_map: ReputationInfoMap
    ) -> Tuple[int, Sha256Set]:
        """
        Find valid sha256s with valid filetypes. Populates info_map.
        :param requests: List of dicts representing lookup requests
        :param info_map: Mapping of sha256s to ReputationLookupInfo to populate
        :return: Set of valid sha256s to perform lookup for
        """
        http_code = 200
        valid_hashes = set()
        invalid_apc_filetype = 0
        non_executable_filetypes = 0
        for lookup_request in requests:
            is_sha256, sha256 = self._get_sha256(lookup_request)
            metadata = Metadata.from_dict(lookup_request.get("metadata") or dict())
            is_valid_file = self._is_valid_file(metadata)

            lookup_info = ReputationLookupInfo(sha256, metadata, lookup_request)
            info_map[sha256] = lookup_info

            if not is_valid_file:
                non_executable_filetypes += 1
                if lookup_info.is_interesting:
                    invalid_apc_filetype += 1
                    lookup_info.invalid_filetype = True

            if is_sha256:
                valid_hashes.add(sha256)
                continue

            if not is_sha256:
                lookup_info.cache_item.status = ReputationStatus.Error
                http_code = 400
        if invalid_apc_filetype > 0:
            self.statsd.incr(
                "defense.invalid_extension_apc.rejection", invalid_apc_filetype
            )
        if non_executable_filetypes > 0:
            self.statsd.incr(
                "defense.non_executable_filetypes", non_executable_filetypes
            )
        return http_code, valid_hashes

    def _mark_portion_submitted(
        self, not_found: Sha256Set, info_map: ReputationInfoMap
    ):
        """
        Mark a portion of the not found items as submitted
        to avoid sensor errors
        :param not_found: Set of not_found hashes
        :param info_map: Mapping of sha256 to ReputationLookupInfo to populate
        """
        for sha256 in not_found:
            lookup_info = info_map[sha256]
            caller_product = lookup_info.metadata.app_name or "services"
            caller_sensor_version = lookup_info.metadata.device_version
            caller_platform = lookup_info.metadata.device_type

            if (
                caller_product == "services"
                and self._check_sensor_version_for_resubmit_hint(
                    caller_sensor_version, caller_platform
                )
                and randint(0, 100) < self.config.submitted_rate
            ):
                lookup_info.cache_item.status = ReputationStatus.Submitted

    def _get_existing_reputation(
        self, valid_hashes: Sha256Set, info_map: ReputationInfoMap
    ) -> Sha256Set:
        """
        Get the existing reputation data from DynamoDB
        :param valid_hashes: Set of hashes to perform lookups for
        :param info_map: Mapping of sha256 to ReputationLookupInfo to populate
        :return: Hashes to check for further analysis results
        """
        if len(valid_hashes) == 0:
            return valid_hashes
        not_found = valid_hashes.copy()
        further_analysis = valid_hashes.copy()
        for item in self.new_cache.get(valid_hashes.copy()):
            if item is None:
                continue
            not_found.remove(item.sha256)
            lookup_info: ReputationLookupInfo = info_map[item.sha256]
            lookup_info.reputation_item = item
            lookup_info.cache_item = item.to_reputation_cache_item(
                is_interesting=lookup_info.is_interesting
            )

            if item.is_definitively_categorized():
                further_analysis.remove(item.sha256)
        self._mark_portion_submitted(not_found, info_map)
        return further_analysis

    def _group_by_pipeline(self, hashes: Sha256Set, info_map: ReputationInfoMap):
        """
        Group hashes according to the pipeline that should be run for them
        based on the lookup request
        :param hashes: Set of sha256s to group
        :param info_map: Mapping of sha256 to ReputationLookupInfo to populate
        :return: Mapping of pipeline type to set of Sha256s
        """
        grouped_hashes: Dict[UbsAnalysisPipeline, Sha256Set] = dict()
        for sha256 in hashes:
            lookup_info = info_map[sha256]
            pipeline = self._determine_pipeline_for_hash(lookup_info)
            lookup_info.pipeline = pipeline
            if pipeline not in grouped_hashes:
                grouped_hashes[pipeline] = set()
            grouped_hashes[pipeline].add(sha256)

        reputation_hashes = grouped_hashes.get(UbsAnalysisPipeline.REPUTATION, set())
        no_file_share_hashes = grouped_hashes.get(
            UbsAnalysisPipeline.REPUTATION_NO_FILE_SHARE, set()
        )

        if len(reputation_hashes) > 0:
            self.statsd.incr(
                "defense.pipeline.reputation_started", len(reputation_hashes)
            )

        if len(no_file_share_hashes) > 0:
            self.statsd.incr(
                "defense.pipeline.reputation_no_file_share_started",
                len(no_file_share_hashes),
            )
        return grouped_hashes

    def _get_hashes_with_stage_results(
        self, hashes: Sha256Set, stage: UbsAnalysisStage, info_map: ReputationInfoMap
    ):
        """
        Get set of hashes which have results for the given stage
        :param hashes: Set of sha256s to filter
        :param stage: Stage to check for results
        :param info_map: Mapping from sha256 to ReputationLookupInfo to use for reference
        """
        return set(
            filter(
                lambda sha256: info_map[sha256].has_completed.get(stage, False),
                hashes,
            )
        )

    def _start_pipelines(
        self,
        grouped_hashes: Dict[UbsAnalysisPipeline, Sha256Set],
        info_map: ReputationInfoMap,
    ) -> Tuple[Sha256Set, int]:
        """
        Start a pipeline for each sha256 in the set
        :param grouped_hashes: Mapping of UbsAnalysisPipeline to set of sha256
        :param info_map: Mapping of sha256 to ReputationLookupInfo to reference and populate
        """
        http_code = 200
        # Only start the REPUTATION pipelines as this is when
        # is_interesting=True
        pipeline = UbsAnalysisPipeline.REPUTATION
        hashes = grouped_hashes.get(pipeline)
        if hashes is None:
            return set(), http_code

        for sha256 in hashes:
            lookup_info = info_map[sha256]
            response = self.ubs_analysis_session.post(
                f"{self.config.ubs_analysis_url}pipeline/start/{pipeline.name}/{sha256}",
                headers=self.get_pass_through_header(),
            )
            if response.status_code != 200:
                self.logger.warning(
                    f"Received HTTP {response.status_code} from UBS analysis frontend."
                    f"Response of {response.text}"
                )
                lookup_info.cache_item.status = ReputationStatus.Error
                lookup_info.version = None
                http_code = 500
                continue
            pipeline_summary = PipelineSummary.from_json(response.content, True)
            lookup_info.version = pipeline_summary.version
            for stage_summary in pipeline_summary.stages:
                lookup_info.has_completed[stage_summary.name] = (
                    stage_summary.status == UbsAnalysisStatus.COMPLETE
                )
        return hashes, http_code

    def _set_response_metadata(self, info_map: ReputationInfoMap):
        """
        Set ttl and need_file for each hash that is waiting on a pipeline to complete
        :param info_map: Mapping of sha256 to ReputationLookupInfo to populate
        """
        for sha256, lookup_info in info_map.items():
            current_status = lookup_info.cache_item.status
            if current_status == ReputationStatus.Error:
                continue

            # Convert to response format
            lookup_info.cache_item = (
                lookup_info.reputation_item.to_reputation_cache_item(
                    lookup_info.is_interesting, lookup_info.is_candidate
                )
            )

            # Force back to SUBMITTED if it was overwritten
            if (
                current_status == ReputationStatus.Submitted
                and lookup_info.cache_item.status == ReputationStatus.NotFound
            ):
                lookup_info.cache_item.status = current_status

            lookup_info.cache_item.set_need_file(
                lookup_info.is_interesting,
                lookup_info.has_completed.get(UbsAnalysisStage.FILE_EXISTS, False),
                lookup_info.has_completed.get(
                    UbsAnalysisStage.APC_UPLOAD_ALLOWED, False
                ),
                lookup_info.reputation_item.is_definitively_categorized(),
            )
            lookup_info.cache_item.set_ttl(
                lookup_info.is_interesting,
                lookup_info.is_candidate,
                lookup_info.has_completed.get(UbsAnalysisStage.FILE_EXISTS, False),
                lookup_info.has_completed.get(
                    UbsAnalysisStage.APC_UPLOAD_ALLOWED, False
                ),
                lookup_info.reputation_item.apc,
            )

    def _build_ordered_response(
        self, requests, info_map: ReputationInfoMap
    ) -> List[Dict]:
        """
        Build response for each request using info from info_map
        :param requests: Dict of requests for reputation lookup
        :param info_map: Mapping from sha256 to ReputationLookupInfo to build response from
        :return: List of responses
        """
        responses = list()
        for lookup_request in requests:
            lookup_info: ReputationLookupInfo = info_map[
                lookup_request["sha256"].lower()
            ]
            lookup_info.cache_item.channel = (
                ReputationChannel.CANDIDATE
                if lookup_info.is_candidate
                else ReputationChannel.STABLE
            )
            lookup_info.cache_item.version = (
                lookup_info.version
                if lookup_info.cache_item.channel == ReputationChannel.CANDIDATE
                else (1, 0)
            )
            responses.append(lookup_info.cache_item.to_response(True))
        return responses

    def _attempt_analysis(self, requests):
        """
        Performs cache lookup and categorization submission if necessary
        :param requests: a list of dicts with sha256, is_interesting (bool), is_candidate (bool) keys
        :return: Results to return to requester
        """
        info_map: ReputationInfoMap = dict()
        http_code, valid_hashes = self._validate_hashes(requests, info_map)
        self.timer.stop_timer(TimerType.INIT)

        further_analysis = self._get_existing_reputation(valid_hashes, info_map)
        self.timer.stop_timer(TimerType.GET_REPUTATION)

        grouped_hashes = self._group_by_pipeline(further_analysis, info_map)

        hashes_started, new_http_code = self._start_pipelines(grouped_hashes, info_map)
        self.timer.stop_timer(TimerType.PUT_APC)
        if http_code == 200:
            http_code = new_http_code

        candidate_hashes = set(
            filter(lambda sha256: info_map[sha256].is_candidate, hashes_started)
        )
        self.statsd.incr("defense.submitted.candidate", len(candidate_hashes))

        # APC hashes will include everything that is is_interesting
        # and everything with channel = "candidate" due to the fact
        # that we run the APC_LOOKUP stage on either of those conditions
        self._log_interesting_and_candidate_items(valid_hashes, info_map)

        self._set_response_metadata(info_map)
        return self._build_ordered_response(requests, info_map), http_code

    @staticmethod
    def _check_sensor_version_for_resubmit_hint(version_to_check, platform):
        """
        '.'.join(affected_parts) is the fixed version of the sensor.
        Anything less than that needs random resubmit on not found.
        Iterate parts of provided version number to check for a smaller version (major.minor.patch.build)
        @param version_to_check: the version number to check for resubmit eligibility
        @param platform: version numbers differ by platform
        @return: False if does not need resubmit else True
        """
        # TODO: right now, mac and unknown always need resubmit. Will need an update when mac sensor is patched.
        if platform is None or version_to_check is None:
            return True
        if platform.lower() == "linux":
            return False
        elif platform.lower() != "windows":
            return True

        affected_parts = [3, 4, 0, 0]

        split_check = version_to_check.split(".")
        for index, part in enumerate(split_check):
            if int(part) < affected_parts[index]:
                return True
        return False


def start_service(service_port):
    config = DefenseConfiguration(service_port)
    config.init_boto_session()
    principal = TicPrincipal(config)
    service = DefenseService(config, principal)
    return service.app


# Entry point for mosaic2
def start_service_mcs2(port):
    config = DefenseConfig(port)
    principal = DefensePrincipal()
    service = DefenseService(config, principal)
    return service.app
