import requests

from tic.defense.dtos.iris_request_payload import IrisRequestPayload
from tic.shared.aws.kms_client import KmsClient


class IrisApiClient(object):
    """
    Client for working with the Iris api. The results are not used in response as this is observation mode for now.
    """

    def __init__(self, config):
        """
        Initializes iris api config from defense config
        """
        self.kms_client = KmsClient(config.boto_session)
        self.iris_token = self.kms_client.decrypt_string(config.iris_token)
        self.api_timeout = config.iris_api_timeout
        self.logger = config.logger
        self.url = f"{config.iris_service_url}/iris-api/v1/internal/file-reputations"

    def query(self, cdc_requests, cdc_lookup_results):
        """
        Invokes Iris api with reputation request parameters and cdc reputation value
        """

        # Construct Iris api request payload
        payload = IrisRequestPayload.generate(cdc_requests, cdc_lookup_results)

        # no payload if no TLSH found in the input
        if payload is None:
            return None

        headers = {"Content-Type": "application/json", "X-Auth-Token": self.iris_token}
        try:
            self.logger.debug(f"Iris call payload={payload}")
            response = requests.request(
                "POST",
                self.url,
                headers=headers,
                data=payload,
                timeout=self.api_timeout,
            )
            if response.status_code == 200:
                self.logger.debug(response.content)
                return response
            self.logger.warn(f"Iris call failed status code = {response.status_code}")
        except requests.exceptions.ReadTimeout:
            self.logger.warn(
                f"Iris call timed out. current timeout = {self.api_timeout} "
            )
        except Exception:
            self.logger.exception("Iris call failed")

        return None
