from multiprocessing.queues import Empty
from multiprocessing import Process
import time
from enum import Enum
import signal
import sys

import boto3
import gevent

from ..shared.utils import log_util
from ..shared.aws.kinesis_client import KinesisClient


class WorkerType(Enum):
    LOG = 2


class QueueWorker(Process):
    def __init__(self, tic_config, queue, worker_type, log_level_name="INFO", **kwargs):
        super(QueueWorker, self).__init__(**kwargs)
        self.queue = queue
        self.worker_type = worker_type
        self.max_wait = 0.3
        self.poll_time = self.max_wait
        self.soft_put = False
        self.logger = log_util.get_logger(
            "defense_qw_{}".format(worker_type.name),
            log_level_name=log_level_name,
        )
        self.items = []
        self.aws_region = tic_config.aws_region

        if self.worker_type == WorkerType.LOG:
            self.stream_name = tic_config.logging_stream
            self.put_method = self._put_logs
            self.max_wait = 300
        else:
            msg = "Queue Worker: Invalid queue type"
            self.logger.error(msg)
            raise Exception(msg)

        self.boto_session = None
        self.kinesis_client = None

        gevent.signal_handler(signal.SIGTERM, self.sigterm_handler)

    def init_boto_session(self):
        boto_session = boto3.Session(region_name=self.aws_region)
        total_seconds = 0
        num_retries = 0

        # retry 25s- k8s waits 30 before calling first ready_check, 60 before first health_check
        while total_seconds < 25:
            try:
                if boto_session.get_credentials() is not None:
                    self.boto_session = boto_session
                    break
            except Exception:
                num_retries += 1

            wait_time = 2**num_retries
            total_seconds += wait_time
            time.sleep(wait_time)

        if self.boto_session is None:
            raise Exception("Queue Worker: Unable to create boto session.")
        self.kinesis_client = KinesisClient(self.boto_session, self.stream_name)

    def sigterm_handler(self, *_):
        print("Queue Worker: Sigterm handler - start")

        self.soft_put = False
        self._put_to_kinesis()
        if len(self.items) > 0:
            self.logger.warning(
                "Queue Worker: Unable to process final queue items."
                " {} items will be dropped on the floor.".format(len(self.items))
            )

        print("Queue Worker: Sigterm handler - end")

        sys.exit(0)

    def run(self):
        self.init_boto_session()

        t = time.time()
        t2 = time.time()
        do_work = False
        while True:
            # Write records to Kinesis if time has elapsed or we're told to
            remaining = self.max_wait - (time.time() - t)
            remaining_poll = self.poll_time - (time.time() - t2)
            self.soft_put = False
            if do_work or remaining < 0.001:  # It's been a long time, do a full flush
                self._put_to_kinesis()
                remaining = self.max_wait
                t = time.time()

            elif (
                remaining_poll < 0.001
            ):  # Do a frequent partial flush (better batch size)
                self.soft_put = True
                self._put_to_kinesis()
                t2 = time.time()
                if (
                    len(self.items) == 0
                ):  # Refresh full flush timer if a full flush was done
                    t = time.time()

            try:
                data = self.queue.get(timeout=remaining)

                # Check for poison pill and cleanup, as needed
                if data is None:
                    self.logger.warning("Queue Worker: Poison Pill Swallowed.")
                    self.soft_put = False
                    self._put_to_kinesis()
                    if len(self.items) > 0:
                        self.logger.warning(
                            "Queue Worker: Unable to process final queue items."
                            " {} items will be dropped on the floor.".format(
                                len(self.items)
                            )
                        )
                    break

                self.items.extend(data)
                do_work = False
            except Empty:  # Timeout reached - put items to kinesis and move on
                do_work = True

        self.logger.warning("Queue Worker: Exit While Loop.")

    def _put_to_kinesis(self):
        num_items = len(self.items)
        if num_items > 0:
            t = time.time()
            self.logger.debug("Queue Worker: Putting items to Kinesis")

            if self.put_method():
                self.logger.debug(
                    "Queue Worker: Put {} items to Kinesis in {}".format(
                        num_items, time.time() - t
                    )
                )
            else:
                return False

        return True

    def _put_logs(self):
        return self._put_error_handler(self.items)

    def _put_error_handler(self, items):
        try:
            self.items = self.kinesis_client.put_records_aggregated({None: items})
            # min_batch_length = (None if self.soft_put is False else 51200)
            # soft put temporarily unsupported. This is the only place it was used.
            return len(items) == 0
        except Exception:
            self.logger.exception(
                "Queue Worker: Failed put to Kinesis for {}. Will retry later.".format(
                    self.worker_type.name
                )
            )
            return False
