import logging
import time
from os import environ

import boto3

from tic.reputation.shared.dtos.reputation_item import ReputationItem
from tic.reputation.shared.dtos.success_responses import (
    ReputationResponseInternal,
)
from tic.shared.aws.kinesis_client import KinesisClient


# Broadcast messages from reputation-categorize-loader-results-* stream into new format.
# Stream with new standard format which is used by NSX and DSS
class RepUpdateBroadcaster(object):
    def __init__(self, rep_stream_name):
        # output reputation stream for new standard format consumed by NSX, AppC DSS and other teams
        # Eventually all teams will use to get rep updates
        self.rep_stream_name = rep_stream_name
        self.logger = logging.getLogger()
        self._set_up_clients()

    def _set_up_clients(self):
        boto_session = boto3.Session(region_name=environ["AWS_DEFAULT_REGION"])
        self.rep_stream = KinesisClient(boto_session, self.rep_stream_name)

    def process_message(self, data):

        # rep_Item is which is instance ReputationResponse, and it is new standard format
        rep_items = self.extract_transform(data)

        # Use sha256 as partition key for KS stream
        rep_records = {}
        for item in rep_items:
            rep_records[item.file_metadata.sha256] = [item.to_json(minimize=True)]

        failed_data = []
        # try replaying failed data
        for i in range(2):
            failed_data = (
                self.rep_stream.put_records_and_get_failed_records_with_error_code(
                    rep_records
                )
            )
            if len(failed_data) == 0:
                self.logger.warning(f"processed {len(rep_records)}.")
                break

            warning = f"try {i} - Failed to add {len(failed_data)} out of {len(rep_records)} records  {failed_data}."
            self.logger.warning(warning)
            # sleep three seconds
            time.sleep((i + 1) * 3)

            # only replay failed records
            failed_partition_keys = []
            for failed_rec in failed_data:
                p = failed_rec["failed-record"]["PartitionKey"]
                failed_partition_keys.append(p)
            rep_records.clear()
            for item in rep_items:
                if item.file_metadata.sha256 in failed_partition_keys:
                    rep_records[item.file_metadata.sha256] = [
                        item.to_json(minimize=True)
                    ]

        return failed_data

    # Extracts RepItesm and transform them new format
    def extract_transform(self, data):
        # Legacy format for gov cloud
        psc_rep_items = self.extract(data)

        # new format for AppC and NSX
        rep_items = list()
        for rep_item in psc_rep_items:
            try:
                rep_response = ReputationResponseInternal().from_reputation_item(
                    rep_item
                )
                rep_items.append(rep_response)
            except Exception as e:
                self.logger.error(e)
                self.logger.error(
                    f"Error occurred while transforming {rep_item.to_json(minimize=True)}"
                )
                raise Exception(e)
        return rep_items

    # deserialize json objects into rep items
    @staticmethod
    def extract(data):
        return [
            ReputationItem.from_json(x)
            for x in data.decode("utf-8").strip().split("\n")
            if x is not None and x != ""
        ]
