import gzip
from copy import deepcopy
from io import BytesIO
import json
from json.decoder import JSONDecodeError
from threading import Thread
from time import sleep
from typing import Iterable, List, Tuple, Optional
from uuid import uuid4

from boto3 import Session
from botocore.config import Config

from tic.reputation.gc_persister.gc_persister_config import RepConsumerPersisterConfig
from tic.reputation.gc_persister.gc_persister_principal import GcPersisterPrincipal
from tic.reputation.shared.dtos.reputation_item import ReputationItem
from tic.reputation.shared.repository.dynamodb_repository import ReputationDbRepository
from tic.shared.aws.dynamodb import ConditionalWriteException
from tic.shared.aws.kinesis_client import KinesisClient
from tic.shared.aws.s3_client import S3Client
from tic.shared.aws.sqs_client import SqsClient
from tic.shared.aws.util import create_boto_session, IMDSException
from tic.shared.service import TICService, SQSQueue
from tic.shared.utils.date_util import to_iso_8601, utc_now
from tic.shared.utils.util import chunk


class GcDynamoWriteThread(Thread):
    def run(self):
        self.exception = None
        self.ret = None
        try:
            self.ret = self._target(*self._args, **self._kwargs)
        except Exception as e:
            self.exception = e

    def join(self, *args):
        super().join(*args)
        if self.exception:
            raise deepcopy(self.exception)
        return self.ret


def gc_retry_conditional_write(func):
    def wrapper(self, *args, **kwargs):
        retry_occurred = False
        attempts = 0
        while attempts < 3:
            try:
                result = func(self, *args, **kwargs)
                if retry_occurred:
                    self.statsd.incr("retry.success")
                return result
            except ConditionalWriteException as e:
                retry_occurred = True
                self.logger.warning("Failed a conditional write.")
                if attempts == 0:
                    self.statsd.incr("retry.occurred")
                attempts += 1
                if attempts < 3:
                    self.statsd.incr("retry.total")
                    sleep(0.2)
                else:
                    self.logger.warning("Failed to persist to write after 3 tries.")
                    self.statsd.incr("retry.failed")
                    raise e

    return wrapper


class RepConsumerPersister(TICService):
    def __init__(self, config, principal):
        super().__init__(config, principal)

        # sqs on the Commercial side
        self.logger.info(
            f"comm_sqs_name={config.comm_sqs_name}, "
            f"comm_sqs_endpoint={config.comm_sqs_endpoint_url}, "
            f"comm_s3_endpoint={config.comm_s3_endpoint_url}, "
            f"comm_aws_region={config.comm_aws_region}, "
        )
        self.logger.info(
            f"persister_dlq ={config.persister_dlq} "
            f"expiration_bucket ={config.psc_expiration_bucket} "
            f"history kinesis stream : {config.kinesis_results_name} "
            f"repDB: {config.reputation_table_name}"
        )

        self.sqs_queue = RepConsumerPersisterQueue(
            config.comm_sqs_name, config.comm_aws_region, config
        )


class RepConsumerPersisterQueue(SQSQueue):
    def __init__(self, queue_name, queue_region, config):
        super().__init__(
            queue_name=queue_name,
            queue_region=queue_region,
            config=config,
            access_key_id=config.comm_access_key_id,
            access_key=config.comm_access_key,
            sqs_endpoint_url=config.comm_sqs_endpoint_url,
        )
        self._set_up_clients()

    def _set_up_clients(self):
        # rep item bucket in Commercial
        comm_boto_config = Config(s3={"addressing_style": "virtual"})
        self.s3_client = S3Client(
            boto_session=self.process_thread_boto_session,
            boto_config=comm_boto_config,
            endpoint_url=self.config.comm_s3_endpoint_url,
        )

        # govcloud session
        self.gc_process_thread_boto_session = Session(
            region_name=self.config.aws_region
        )

        # expiration bucket in govcloud
        self.logger.info(
            f"gc fips_s3_endpoint from global = {self.config.fips_s3_endpoint}"
        )
        fips_s3_endpoint = None
        if self.config.fips_s3_endpoint:
            fips_s3_endpoint = "https://" + self.config.fips_s3_endpoint

        self.gc_s3_client = S3Client(
            boto_session=self.gc_process_thread_boto_session,
            endpoint_url=fips_s3_endpoint,
        )

        self.logger.info(
            f"gc s3 endpoint_url={self.gc_s3_client.s3_client.meta.endpoint_url}"
        )

        # dlq in govcloud
        self.gc_sqs_client = SqsClient(self.gc_process_thread_boto_session)

        # history stream->bucket in govcloud
        self.kinesis_client = KinesisClient(
            self.gc_process_thread_boto_session, self.config.kinesis_results_name
        )

        # use the main session in govcloud for the first repository
        self.repositories = [
            ReputationDbRepository(
                table_name=self.config.reputation_table_name,
                session=self.gc_process_thread_boto_session,
            )
        ]

        while len(self.repositories) < self.config.thread_count:
            self.repositories.append(
                ReputationDbRepository(
                    table_name=self.config.reputation_table_name,
                    session=create_boto_session(region=self.config.aws_region),
                )
            )

    def process_malformatted_msg(self, message, error):
        # send to persister dlq in govCloud
        msg = [json.dumps({"error": error, "message_body": message["Body"]})]

        self.gc_sqs_client.send_message_batch(
            queue=self.gc_sqs_client.get_queue_url(
                queue_name=self.config.persister_dlq
            ),
            messages=msg,
        )

        # delete the messaege from the sqs in Commercial
        self.sqs_client.delete_message(self.queue_url, message["ReceiptHandle"])

    def process_message(self, message):
        # prevents starting processing before all repositories are live
        if len(self.repositories) < self.config.thread_count:
            sleep(5)
            return

        process_start = utc_now()
        metrics = dict()
        self.logger_extra["message_id"] = message.get("MessageId")

        raw_msg = message["Body"]
        # try collect which message type it is
        try:
            sqs_msg_body = json.loads(raw_msg)
        except JSONDecodeError as err:
            self.logger.exception(f"JSONDecodeError for message received:{str(err)}")
            self.process_malformatted_msg(message, str(err))
            return
        except Exception as ex:
            self.logger.exception(f"Exception captured for message received:{str(ex)}")
            self.process_malformatted_msg(message, str(ex))
            return

        full_message_succeeded = True
        categorized_items = list()
        sha256s = list()
        try:
            msg = json.loads(sqs_msg_body["Message"])
            thread_function = None
            thread_input = None

            if "Records" in msg:
                # this implies s3 source for messages
                items_from_feed = list()
                # one prefix is accepted at a time
                for record in msg["Records"]:
                    try:
                        items_from_feed = self.extract_items(
                            s3_bucket=record["s3"]["bucket"]["name"],
                            s3_key=record["s3"]["object"]["key"],
                        )
                    except Exception as ex:
                        self.logger.exception(
                            f"failed to extract rep items from history bucket - {str(ex)}"
                        )
                        self.process_malformatted_msg(message, str(ex))
                        return
                shas = list()
                for item in items_from_feed:
                    shas.append(item.sha256)
                self.logger.debug(f"shas received:{shas}")

                if len(items_from_feed) == 0:
                    metrics["empty_updates"] = 1
                else:
                    metrics["hashes.reputation_feed"] = len(items_from_feed)
                    thread_function = self._handle_feed_updates
                    thread_input = items_from_feed
                    if len(items_from_feed) > 10000:
                        self.sqs_client.change_message_visibility(
                            self.queue_url,
                            message["ReceiptHandle"],
                            timeout=(len(items_from_feed) // 10000) * 500,
                        )
            else:
                self.logger.exception("Unknown message format")
                self.process_malformatted_msg(message, "unknown message format")
                return

            if thread_function is not None:
                # gives us batches - one chunk of reputation items for each repository per batch
                for batch in chunk(
                    chunk(
                        thread_input
                        if isinstance(thread_input, list)
                        else [thread_input],
                        100,
                    ),
                    self.config.thread_count,
                ):
                    batch_categorized = list()
                    batch_sha256s = list()

                    # now we are in a batch - execute each chunk in a thread.
                    threads = list()
                    for ix, c in enumerate(batch):
                        t = GcDynamoWriteThread(
                            target=thread_function, args=(c, self.repositories[ix])
                        )
                        t.start()
                        threads.append(t)

                    for t in threads:
                        chunk_categorized, chunk_sha256s = t.join()
                        batch_categorized.extend(chunk_categorized)
                        batch_sha256s.extend(chunk_sha256s)

                    categorized_items.extend(batch_categorized)
                    sha256s.extend(batch_sha256s)

        except IMDSException as e:
            self.statsd.incr("imds_failure")
            self.logger.warning("imds_failure" + str(e))
            full_message_succeeded = False
            sleep(10)
        except Exception as e:
            self.logger.exception(f"Failure persisting reputation data: {str(e)}")
            full_message_succeeded = False

            if "Parameter validation failed" in str(e):
                self.process_malformatted_msg(message, "unknown message format")

            raise e

        finally:
            self.logger.debug(
                f"number of categorised items: {len(categorized_items)}, number pf shas: {len(sha256s)}"
            )
            self.logger.debug(f"processining shas: {sha256s}")
            if len(categorized_items) > 0:
                failed_msg = self.kinesis_client.put_records_aggregated(
                    {None: [item.to_json(minimize=True) for item in categorized_items]}
                )

                if len(failed_msg) > 0:
                    self.logger.warning(
                        "failed to write all rep items to history bucket"
                    )
                    full_message_succeeded = False

                self._publish_psc_cache_expiration(sha256s)

            if full_message_succeeded:
                self.sqs_client.delete_message(self.queue_url, message["ReceiptHandle"])

            for name, value in metrics.items():
                self.statsd.incr(name, value)
            self.statsd.timing("process_time", utc_now() - process_start)

    @gc_retry_conditional_write
    def _handle_feed_updates(
        self, items_from_feed: List[ReputationItem], repository: ReputationDbRepository
    ) -> Tuple[List[ReputationItem], Iterable[str]]:

        # map of sha256 to ReputationItem.
        items_map = {item.sha256: item for item in items_from_feed}
        # find all existing items, coalesce, and update items_map
        for existing_item in repository.get(list(items_map.keys())):
            items_map[existing_item.sha256] = self.merge_feed_items(
                feed_item=items_map[existing_item.sha256],
                existing_item=existing_item,
            )
        items_to_categorize = [item for item in items_map.values() if item is not None]

        return self.categorize_and_publish(items_to_categorize, repository)

    @staticmethod
    def merge_feed_items(
        feed_item: ReputationItem, existing_item: ReputationItem
    ) -> Optional[ReputationItem]:
        if existing_item.feed_date >= feed_item.feed_date:
            return
        feed_item.apc = existing_item.apc
        feed_item.version = existing_item.version
        return feed_item

    def extract_items(self, s3_bucket: str, s3_key: str) -> List[ReputationItem]:
        """
        Returns json objects contained in S3 object from Reputation Feed
        @param s3_bucket: The bucket to pull rep feed artifacts from
        @param s3_key: The object to extract records from
        @return: List of json objects extracted from rep feed file
        """
        with BytesIO() as zf:
            self.s3_client.download_fileobj(
                file_obj=zf, bucket_name=s3_bucket, key_name=s3_key
            )
            return [
                ReputationItem.from_json(x)
                for x in gzip.decompress(zf.getvalue()).decode("utf-8").split("\n")
                if x is not None and x != ""
            ]

    def categorize_and_publish(
        self, items: Iterable[ReputationItem], repository: ReputationDbRepository
    ) -> Tuple[List[ReputationItem], Iterable[str]]:
        """
        Categorize and save each reputation item
        @param items: ReputationItems to categorize
        @param repository: the repository to use to perform updates
        """
        categorized_items = list()
        sha256s = set()
        for item in items:
            sha256s.add(item.sha256)
            item.categorize()
            item.date_modified = utc_now()
            categorized_items.append(item)

        if len(categorized_items) > 0:
            self.persist_to_dynamodb(categorized_items, repository)
        return categorized_items, sha256s

    def _publish_psc_cache_expiration(self, sha256s: Iterable[str]) -> None:
        """
        Publishes sha256s as flat file to s3, which pushes notification to clear PSC caches.
        @param sha256s: the sha256s to expire
        """
        with BytesIO() as f:
            f.write("\n".join(sha256s).encode("utf-8"))
            f.seek(0)
            self.gc_s3_client.upload_fileobj(
                file_obj=f,
                bucket_name=self.config.psc_expiration_bucket,
                key_name=f'{to_iso_8601(utc_now()).replace(":", "-")}_{str(uuid4())}',
            )

    @staticmethod
    def persist_to_dynamodb(
        reputation_items: Iterable[ReputationItem], repository: ReputationDbRepository
    ) -> None:
        for item in reputation_items:
            item.version += 1

        repository.update(reputation_items)


# Entry point for mosaic2
def start_service_mcs2(port):
    config = RepConsumerPersisterConfig(port)
    principal = GcPersisterPrincipal()
    service = RepConsumerPersister(config, principal)
    return service.app
