"""
step2 of the rep-db migration:
1) process the messages from sqs that contains the rep item file prefixe in S3
2) fetch rep item files from S3
3) translate the dynamodb json formatted rep items into rep-item dto
4) write to destination rep-db (fail to dlq)
"""
import gzip
import json
import os
from copy import deepcopy
from json.decoder import JSONDecodeError
from threading import Thread
from time import sleep
from typing import List, Iterable

from tic.reputation.gc_repdb_migration.rep_importer.gc_repdb_importer_config import (
    GcRepDbimporterConfig,
)
from tic.reputation.gc_repdb_migration.rep_importer.gc_repdb_importer_principal import (
    GcRebDbimporterPrincipal,
)
from tic.reputation.shared.dtos.reputation_item import ReputationItem
from tic.reputation.shared.repository.dynamodb_repository import ReputationDbRepository
from tic.shared.aws.dynamodb import ConditionalWriteException
from tic.shared.aws.s3_client import S3Client
from tic.shared.aws.sqs_client import SqsClient
from tic.shared.aws.util import create_boto_session, IMDSException
from tic.shared.service import TICService, SQSQueue
from tic.shared.utils.date_util import utc_now
from tic.shared.utils.util import chunk


class GcDynamoWriteThread(Thread):
    def run(self):
        self.exception = None
        self.ret = None
        try:
            self.ret = self._target(*self._args, **self._kwargs)
        except Exception as e:
            self.exception = e

    def join(self, *args):
        super().join(*args)
        if self.exception:
            raise deepcopy(self.exception)
        return self.ret


def retry_repdb_write(func):
    def wrapper(self, *args, **kwargs):
        retry_occurred = False
        attempts = 0
        while attempts < 3:
            try:
                result = func(self, *args, **kwargs)
                if retry_occurred:
                    self.logger.info("retry success")
                return result
            except ConditionalWriteException as e:
                retry_occurred = True
                self.logger.warning("Failed a conditional write.")
                attempts += 1
                if attempts < 3:
                    sleep(0.2)
                else:
                    self.logger.warning("Failed to persist to write after 3 tries.")
                    raise e

    return wrapper


class GcRebDbimporter(TICService):
    def __init__(self, config, principal):
        super().__init__(config, principal)
        self.logger.info(
            f"importer_dlq ={config.repdb_importer_dlq} "
            f"sqs queue: {config.rep_sqs_name} "
            f"expiry queue: {config.repdb_expiry_sqs}"
            f"repDB: {config.reputation_table_name}"
        )

        self.sqs_queue = RepDbimporterQueue(
            config.rep_sqs_name, config.aws_region, config
        )


class RepDbimporterQueue(SQSQueue):
    def __init__(self, queue_name, queue_region, config):
        if config.gc_sqs_endpoint:
            super().__init__(
                queue_name=queue_name,
                queue_region=queue_region,
                config=config,
                sqs_endpoint_url=config.gc_sqs_endpoint,
            )
        else:
            super().__init__(
                queue_name=queue_name, queue_region=queue_region, config=config
            )

        self._set_up_clients()

    def _set_up_clients(self):
        # rep S3 bucket
        self.logger.info(
            f"gc fips_s3_endpoint from global = {self.config.fips_s3_endpoint}"
        )
        fips_s3_endpoint = None
        if self.config.fips_s3_endpoint:
            fips_s3_endpoint = "https://" + self.config.fips_s3_endpoint

        self.rep_s3_client = S3Client(
            boto_session=self.process_thread_boto_session, endpoint_url=fips_s3_endpoint
        )

        self.logger.info(
            f"rep s3 endpoint_url={self.rep_s3_client.s3_client.meta.endpoint_url}"
        )

        # sqs client
        self.rep_sqs_client = (
            SqsClient(self.process_thread_boto_session)
            if not self.config.gc_sqs_endpoint
            else SqsClient(
                boto_session=self.process_thread_boto_session,
                endpoint_url=self.config.gc_sqs_endpoint,
            )
        )
        self.expiry_sqs_url = self.rep_sqs_client.get_queue_url(
            queue_name=self.config.repdb_expiry_sqs
        )
        self.dlq_url = self.rep_sqs_client.get_queue_url(
            queue_name=self.config.repdb_importer_dlq
        )
        self.logger.info(
            f"expiry-sqs-url = {self.expiry_sqs_url} "
            f"importer_dql url = {self.dlq_url} "
            f"migration-sqs-url = { self.queue_url}"
        )

        # repdb
        self.repositories = [
            ReputationDbRepository(
                table_name=self.config.reputation_table_name,
                session=self.process_thread_boto_session,
            )
        ]

        while len(self.repositories) < self.config.thread_count:
            self.repositories.append(
                ReputationDbRepository(
                    table_name=self.config.reputation_table_name,
                    session=create_boto_session(region=self.config.aws_region),
                )
            )

    def send_to_expiry(self, in_msg, info):
        msg = [json.dumps({"info": info, "message": in_msg})]

        # send to dlq
        self.rep_sqs_client.send_message_batch(
            queue=self.expiry_sqs_url,
            messages=msg,
        )

    def send_to_dlq(self, in_msg, error):
        msg = [json.dumps({"error": error, "message": in_msg})]

        # send to dlq
        self.rep_sqs_client.send_message_batch(
            queue=self.dlq_url,
            messages=msg,
        )

    def delete_msg_from_queue(self, message):
        # delete the messaege from the rep item queue
        self.rep_sqs_client.delete_message(self.queue_url, message["ReceiptHandle"])

    def process_message(self, message):
        process_start = utc_now()
        message_body = message["Body"]
        self.logger.info(
            f"Start to process message: {message_body}, number of thread: {self.config.thread_count} ..."
        )

        # prevents starting processing before all repositories are live
        if len(self.repositories) < self.config.thread_count:
            self.logger.info("wait 5 seconds for boto sessions creation complates ...")
            sleep(5)
            return

        self.logger_extra["message_id"] = message.get("MessageId")

        try:
            sqs_msg_body = json.loads(message_body)
        except JSONDecodeError as err:
            self.logger.exception(f"JSONDecodeError for message received:{str(err)}")
            self.send_to_dlq(message_body, str(err))
            self.delete_msg_from_queue(message)
            return
        except Exception as ex:
            self.logger.exception(f"Exception captured for message received:{str(ex)}")
            self.send_to_dlq(message_body, str(ex))
            self.delete_msg_from_queue(message)
            return

        full_message_succeeded = True
        cnt = 0
        try:
            # move message to expiry queue, delete the message from queue to avoid being grabbed by other importers
            self.send_to_expiry(message_body, "message processed")
            self.delete_msg_from_queue(message)

            bucket = sqs_msg_body["s3_bucket"]
            prefix = sqs_msg_body["prefix"]
            self.logger.info(
                f"Processing rep items in s3 bucket {bucket} object {prefix}..."
            )

            for rep_item_batch in self.extract_transform_items(
                s3_bucket=bucket,
                s3_key=prefix,
            ):
                cnt += 1
                self.logger.info(  # debug
                    f"processing #{cnt} batch of {len(rep_item_batch)} rep items ..."
                )

                thread_function = self._handle_rep_items

                if thread_function is not None:
                    # gives us batches - one chunk of reputation items for each repository per batch
                    for batch in chunk(
                        chunk(
                            rep_item_batch
                            if isinstance(rep_item_batch, list)
                            else [rep_item_batch],
                            1000,
                        ),
                        self.config.thread_count,
                    ):
                        # now we are in a batch - execute each chunk in a thread.
                        threads = list()
                        for ix, c in enumerate(batch):
                            t = GcDynamoWriteThread(
                                target=thread_function, args=(c, self.repositories[ix])
                            )
                            t.start()
                            threads.append(t)

                        for t in threads:
                            t.join()

        except IMDSException as ex:
            self.logger.warning("imds_failure" + str(ex))
            full_message_succeeded = False
            sleep(10)
        except Exception as ex:
            self.logger.exception(f"Failure persisting reputation data: {str(ex)}")
            full_message_succeeded = False
            raise ex

        finally:
            if not full_message_succeeded:
                self.send_to_dlq(message_body, "Exception persisting to repdb")

            self.logger.info(
                f"Processed {cnt} batches. Process_time: {utc_now() - process_start}"
            )

    @retry_repdb_write
    def _handle_rep_items(
        self, rep_items: List[ReputationItem], repository: ReputationDbRepository
    ):
        self.persist_to_dynamodb(rep_items, repository)

    def get_val(self, dynamo_json_obj):
        singlar_types = ["S", "N"]
        for key, val in dynamo_json_obj.items():
            if key in singlar_types:
                if key == "S":
                    return str(val)
                elif key == "N":
                    if val.isdigit():
                        return int(val)
                    else:
                        return float(val)
            elif key == "BOOL":
                return True if val == "True" else False
            elif key == "M":
                ret_val = dict()
                for key1, val1 in val.items():
                    ret_val[key1] = self.get_val(val1)
                return ret_val
            elif key == "L":
                ret_val = list()
                for val1 in val:
                    ret_val.append(self.get_val(val1))
                return ret_val

        return None

    def from_dynamo_json(self, dynamo_json):
        dynamo_json_object = json.loads(dynamo_json)
        trimmed_dynamo_json_obj = dynamo_json_object["Item"]
        result = dict()
        for key, val in trimmed_dynamo_json_obj.items():
            result[key] = self.get_val(val)

        if "apc" in result and "apcv4" not in result:
            result["apcv4"] = result["apc"]

        return result

    def extract_transform_items(
        self, s3_bucket: str, s3_key: str
    ) -> List[ReputationItem]:
        """
        Returns json objects transformed from S3 object in dynamodb json format
        @param s3_bucket: The bucket to pull rep dynamodb jsom from
        @param s3_key: The object to extract records from
        @return: trunk of json objects extracted from s3 dynamodb json
        """
        chunk_size = 5000
        filename = "/tmp/" + s3_key.split("/")[-1]
        with open(filename, "wb") as zf:
            self.rep_s3_client.download_fileobj(
                file_obj=zf, bucket_name=s3_bucket, key_name=s3_key
            )
            self.logger.info(
                f"Successfully downloaded file {s3_key} from S3, processing rep items..."
            )

            with gzip.open(filename, "rb") as f:
                rep_items = list()
                for line in f:
                    item = line.rstrip()
                    if item:
                        rep_item = json.dumps(self.from_dynamo_json(item))
                        # self.logger.debug(f"rep_item: {rep_item}")  #  debug
                        rep_items.append(ReputationItem.from_json(rep_item))

                        if len(rep_items) != chunk_size:
                            continue

                        yield rep_items
                        rep_items = list()

                if len(rep_items) > 0:
                    yield rep_items

                os.remove(filename)

    @staticmethod
    def persist_to_dynamodb(
        reputation_items: Iterable[ReputationItem], repository: ReputationDbRepository
    ) -> None:
        repository.update_unconditional(reputation_items)


def start_service_mcs2(port):
    config = GcRepDbimporterConfig(port)
    principal = GcRebDbimporterPrincipal()
    service = GcRebDbimporter(config, principal)
    return service.app
