"""
step one of rep-db migration app:
transport S3 prefixes to sqs that is to be consumed by gc_repdb_importer
pre-requisites:
- rep-db items have been exported to the S3
"""
import json

from boto3 import Session

from tic.reputation.gc_repdb_migration.rep_processor.gc_repdb_processor_config import (
    GcRepDbProcessorConfig,
)
from tic.reputation.gc_repdb_migration.rep_processor.gc_repdb_processor_principal import (
    GcRepDBProcessorPrincipal,
)
from tic.shared.aws.sqs_client import SqsClient
from tic.shared.service import TICService


class GcRepDBProcessor(TICService):
    def __init__(self, config, principal):
        super().__init__(config, principal)

        # boto session
        self.boto_session = Session(region_name=self.config.aws_region)

        # S3
        self.s3_bucket_name = self.config.rep_s3_bucket_name
        self.s3_prefix = self.config.rep_s3_prefix
        self.logger.info(
            f"gc fips_s3_endpoint from global = {self.config.fips_s3_endpoint}"
        )
        fips_s3_endpoint = None
        if self.config.fips_s3_endpoint:
            fips_s3_endpoint = "https://" + self.config.fips_s3_endpoint

        self.s3_client = self.boto_session.client(
            "s3",
            endpoint_url=fips_s3_endpoint,
        )

        self.logger.info(f"gc s3 endpoint_url={self.s3_client.meta.endpoint_url}")

        # sqs endpoint (gov-cloud enforces endpoint url)
        gc_sqs_endpoint = self.config.gc_sqs_endpoint
        # rep item sqs
        self.rep_queue = (
            SqsClient(self.boto_session)
            if not gc_sqs_endpoint
            else SqsClient(boto_session=self.boto_session, endpoint_url=gc_sqs_endpoint)
        )
        self.rep_queue_url = self.rep_queue.get_queue_url(
            queue_name=self.config.rep_sqs_name
        )
        self.logger.info(
            f"migration-sqs name = {self.config.rep_sqs_name}, migration-sqs url = {self.rep_queue_url}"
        )

        # dlq
        self.rep_processor_dlq = (
            SqsClient(self.boto_session)
            if not gc_sqs_endpoint
            else SqsClient(boto_session=self.boto_session, endpoint_url=gc_sqs_endpoint)
        )
        self.rep_processor_dlq_url = self.rep_processor_dlq.get_queue_url(
            queue_name=self.config.processor_dlq
        )
        self.logger.info(
            f"processor dlq name: {self.config.processor_dlq}, processor dlq url: {self.rep_processor_dlq_url} "
        )

        self.logger.info("Start to process repdb items in S3 ...")
        try:
            self.process_rep_items()
        except Exception as ex:
            self.logger.warn("error processing rep items from s3 to sqs" + str(ex))

    def process_rep_items(self):
        paginator = self.s3_client.get_paginator("list_objects_v2")

        pages = paginator.paginate(Bucket=self.s3_bucket_name, Prefix=self.s3_prefix)
        self.logger.info(f"number of pages: {len([pages])}")

        for page in pages:
            contents = page.get("Contents")
            filenames = list()
            for c in contents:
                pf = c.get("Key")
                if pf.endswith(".gz"):
                    pfs = pf.split("/")
                    filenames.append(pfs[-1])
            self.send_to_sqs(filenames)

    def send_to_sqs(self, filenames):
        message = dict()
        self.logger.info(f"sending to rep sqs, number of files: {len(filenames)}")
        for f in filenames:
            self.logger.info(f"processing file: {f}")
            message["s3_bucket"] = self.s3_bucket_name
            message["prefix"] = self.s3_prefix + f
            message_body = json.dumps(message)
            response = self.rep_queue.send_message(self.rep_queue_url, message_body)

            self.logger.info(f"send to sqs - response={response}")

            response_meta = response.get("ResponseMetadata") if response else None
            if response_meta:
                response_code = response_meta.get("HTTPStatusCode")
                self.logger.debug(f"response code: {response_code}")

                if response_code != 200:
                    self.rep_processor_dlq.send_message(
                        self.rep_processor_dlq_url, message_body
                    )


def start_service_mcs2(port):
    config = GcRepDbProcessorConfig(port)
    principal = GcRepDBProcessorPrincipal()
    service = GcRepDBProcessor(config, principal)
    return service.app
