import gzip
from copy import deepcopy
from io import BytesIO
import json
from threading import Thread
from time import sleep
from typing import Iterable, List, Dict, Tuple, Optional
from uuid import uuid4

from tic.reputation.persister.persister_config import RepPersisterConfig
from tic.reputation.persister.persister_principal import PersisterPrincipal
from tic.reputation.shared.dtos.reputation_item import ReputationItem, Apc
from tic.reputation.shared.repository.dynamodb_repository import ReputationDbRepository
from tic.shared.apc.dtos.apc_query import ApcQuery
from tic.shared.apc.utils import get_apc_description, get_apc_verdict
from tic.shared.aws.dynamodb import ConditionalWriteException
from tic.shared.aws.kinesis_client import KinesisClient
from tic.shared.aws.s3_client import S3Client
from tic.shared.aws.util import create_boto_session, IMDSException
from tic.shared.dto import Dto
from tic.shared.service import TICConfiguration, TicPrincipal, TICService, SQSQueue
from tic.shared.ubs.enums.analysis.stage import UbsAnalysisStage
from tic.shared.utils.date_util import to_iso_8601, utc_now
from tic.shared.utils.util import chunk


class ReputationPersisterConfig(TICConfiguration):
    def __init__(self, port):
        super().__init__("reputation_persister", port)
        self.aws_region = None
        self.reputation_table_name = None
        self.kinesis_results_name = None
        self.psc_expiration_bucket = None
        self.thread_count = 10
        self.config_changed()

    def config_changed(self):
        self.aws_region = self.get("AWS", "aws_region")
        self.kinesis_results_name = self.get(
            "reputation_persister", "kinesis_results_name"
        )
        self.reputation_table_name = self.get(
            "reputation_persister", "reputation_table_name"
        )
        self.psc_expiration_bucket = self.get(
            "reputation_persister", "psc_expiration_bucket"
        )


class DynamoWriteThread(Thread):
    def run(self):
        self.exception = None
        self.ret = None
        try:
            self.ret = self._target(*self._args, **self._kwargs)
        except Exception as e:
            self.exception = e

    def join(self, *args):
        super().join(*args)
        if self.exception:
            raise deepcopy(self.exception)
        return self.ret


def retry_conditional_write(func):
    def wrapper(self, *args, **kwargs):
        retry_occurred = False
        attempts = 0
        while attempts < 3:
            try:
                result = func(self, *args, **kwargs)
                if retry_occurred:
                    self.statsd.incr("retry.success")
                return result
            except ConditionalWriteException as e:
                retry_occurred = True
                self.logger.warning("Failed a conditional write.")
                if attempts == 0:
                    self.statsd.incr("retry.occurred")
                attempts += 1
                if attempts < 3:
                    self.statsd.incr("retry.total")
                    sleep(0.2)
                else:
                    self.statsd.incr("retry.failed")
                    raise e

    return wrapper


class ReputationPersister(TICService):
    def __init__(self, config, principal, is_mcs2=False):
        super().__init__(config, principal)

        sqs_name = (
            self.config.persister_sqs_name
            if is_mcs2
            else "reputation-categorize-and-load"
        )
        self.sqs_queue = ReputationPersisterQueue(
            sqs_name, self.config.aws_region, self.config
        )


class ReputationPersisterQueue(SQSQueue):
    def __init__(self, queue_name, queue_region, config):
        super().__init__(queue_name, queue_region, config)
        self._set_up_clients()

    def _set_up_clients(self):
        self.s3_client = S3Client(self.process_thread_boto_session)
        self.kinesis_client = KinesisClient(
            self.process_thread_boto_session, self.config.kinesis_results_name
        )

        # use the main session for the first repository
        self.repositories = [
            ReputationDbRepository(
                table_name=self.config.reputation_table_name,
                session=self.process_thread_boto_session,
            )
        ]

        while len(self.repositories) < self.config.thread_count:
            self.repositories.append(
                ReputationDbRepository(
                    table_name=self.config.reputation_table_name,
                    session=create_boto_session(region=self.config.aws_region),
                )
            )

    def process_message(self, message):
        # prevents starting processing before all repositories are live
        if len(self.repositories) < self.config.thread_count:
            sleep(5)
            return

        process_start = utc_now()
        metrics = dict()
        self.logger_extra["message_id"] = message.get("MessageId")
        # try collect which message type it is
        sqs_msg_body = json.loads(message["Body"])
        full_message_succeeded = True
        categorized_items = list()
        sha256s = list()
        try:
            msg = json.loads(sqs_msg_body["Message"])
            thread_function = None
            thread_input = None
            if "Records" in msg:
                items_from_feed = list()
                # this implies s3 source for messages
                for record in msg["Records"]:
                    items_from_feed = self.extract_items(
                        s3_bucket=record["s3"]["bucket"]["name"],
                        s3_key=record["s3"]["object"]["key"],
                    )
                if len(items_from_feed) == 0:
                    metrics["empty_updates"] = 1
                else:
                    metrics["hashes.reputation_feed"] = len(items_from_feed)
                    thread_function = self._handle_feed_updates
                    thread_input = items_from_feed
                    if len(items_from_feed) > 10000:
                        self.sqs_client.change_message_visibility(
                            self.queue_url,
                            message["ReceiptHandle"],
                            timeout=(len(items_from_feed) // 10000) * 500,
                        )
            elif "stage_name" in msg:
                stage_name = msg["stage_name"]
                if stage_name in {
                    UbsAnalysisStage.APC_LOOKUP.name,
                    UbsAnalysisStage.APC_LOOKUP_POLL.name,
                }:
                    metrics["hashes.apc"] = 1
                    msg["result"] = ApcQuery.from_json(msg["result"])
                    thread_function = self.merge_apc
                    thread_input = msg
                elif stage_name == UbsAnalysisStage.ROENTGEN.name:
                    metrics["hashes.roentgen"] = 1
                else:
                    raise Exception(f"No configuration for updates from {stage_name}")
            else:
                raise Exception("Unknown message format")

            if thread_function is not None:
                # gives us batches - one chunk of reputation items for each repository per batch
                for batch in chunk(
                    chunk(
                        thread_input
                        if isinstance(thread_input, list)
                        else [thread_input],
                        100,
                    ),
                    self.config.thread_count,
                ):
                    batch_categorized = list()
                    batch_sha256s = list()

                    # now we are in a batch - execute each chunk in a thread.
                    threads = list()
                    for ix, c in enumerate(batch):
                        t = DynamoWriteThread(
                            target=thread_function, args=(c, self.repositories[ix])
                        )
                        t.start()
                        threads.append(t)

                    for t in threads:
                        chunk_categorized, chunk_sha256s = t.join()
                        batch_categorized.extend(chunk_categorized)
                        batch_sha256s.extend(chunk_sha256s)

                    categorized_items.extend(batch_categorized)
                    sha256s.extend(batch_sha256s)

        except IMDSException as e:
            self.statsd.incr("imds_failure")
            self.logger.warning(str(e))
            full_message_succeeded = False
            sleep(10)
        except Exception as e:
            self.logger.exception(f"Failure persisting reputation data: {str(e)}")
            full_message_succeeded = False
            raise e

        finally:
            if len(categorized_items) > 0:
                failed_records = self.kinesis_client.put_records_aggregated(
                    {None: [item.to_json(minimize=True) for item in categorized_items]}
                )
                if len(failed_records) > 0:
                    full_message_succeeded = False

                self._publish_psc_cache_expiration(
                    sha256s
                )  # expire even if partially in repDB

            if full_message_succeeded:
                self.sqs_client.delete_message(self.queue_url, message["ReceiptHandle"])

            for name, value in metrics.items():
                self.statsd.incr(name, value)
            self.statsd.timing("process_time", utc_now() - process_start)

    def get_single_item_from_model(self, model: Dto) -> ReputationItem:
        try:
            item = next(self.repositories[0].get([model.sha256]))
        except StopIteration:
            item = ReputationItem(sha256=model.sha256)
        return item

    @retry_conditional_write
    def _handle_feed_updates(
        self, items_from_feed: List[ReputationItem], repository: ReputationDbRepository
    ) -> Tuple[List[ReputationItem], Iterable[str]]:
        # map of sha256 to ReputationItem.
        items_map = {item.sha256: item for item in items_from_feed}
        # find all existing items, coalesce, and update items_map
        for existing_item in repository.get(list(items_map.keys())):
            items_map[existing_item.sha256] = self.merge_feed_items(
                feed_item=items_map[existing_item.sha256],
                existing_item=existing_item,
            )
        items_to_categorize = [item for item in items_map.values() if item is not None]
        return self.categorize_and_publish(items_to_categorize, repository)

    @staticmethod
    def merge_feed_items(
        feed_item: ReputationItem, existing_item: ReputationItem
    ) -> Optional[ReputationItem]:
        if existing_item.feed_date >= feed_item.feed_date:
            return
        feed_item.apc = existing_item.apc
        feed_item.version = existing_item.version
        return feed_item

    @retry_conditional_write
    def merge_apc(
        self, apc_msg: List[Dict], repository: ReputationDbRepository
    ) -> Tuple[List[ReputationItem], Iterable[str]]:
        apc_msg = apc_msg[0]
        existing = self.get_single_item_from_model(apc_msg["result"])
        if existing.apc is not None:
            if existing.apc.stage_version > apc_msg["stage_version"] or (
                existing.apc.stage_version == apc_msg["stage_version"]
                and existing.apc.query_date >= apc_msg["result"].query_date
            ):
                return self.categorize_and_publish([], repository)

        existing.apc = Apc(
            stage_version=apc_msg["stage_version"],
            ttl=apc_msg[
                "result"
            ].ttl_absolute,  # store the absolute ttl in the db for services to know whether to request refresh
            status=apc_msg["result"].status,
            category=apc_msg["result"].category,
            verdict=get_apc_verdict(get_apc_description(apc_msg["result"].category)),
            query_date=apc_msg["result"].query_date,
        )

        return self.categorize_and_publish([existing], repository)

    def extract_items(self, s3_bucket: str, s3_key: str) -> List[ReputationItem]:
        """
        Returns json objects contained in S3 object from Reputation Feed
        @param s3_bucket: The bucket to pull rep feed artifacts from
        @param s3_key: The object to extract records from
        @return: List of json objects extracted from rep feed file
        """
        with BytesIO() as zf:
            self.s3_client.download_fileobj(
                file_obj=zf, bucket_name=s3_bucket, key_name=s3_key
            )
            return [
                ReputationItem.from_json(x)
                for x in gzip.decompress(zf.getvalue()).decode("utf-8").split("\n")
                if x is not None and x != ""
            ]

    def categorize_and_publish(
        self, items: Iterable[ReputationItem], repository: ReputationDbRepository
    ) -> Tuple[List[ReputationItem], Iterable[str]]:
        """
        Categorize and save each reputation item
        @param items: ReputationItems to categorize
        @param repository: the repository to use to perform updates
        """
        categorized_items = list()
        sha256s = set()
        for item in items:
            sha256s.add(item.sha256)
            item.categorize()
            item.date_modified = utc_now()
            categorized_items.append(item)

        if len(categorized_items) > 0:
            self.persist_to_dynamodb(categorized_items, repository)
        return categorized_items, sha256s

    def _publish_psc_cache_expiration(self, sha256s: Iterable[str]) -> None:
        """
        Publishes sha256s as flat file to s3, which pushes notification to clear PSC caches.
        @param sha256s: the sha256s to expire
        """

        with BytesIO() as f:
            f.write("\n".join(sha256s).encode("utf-8"))
            f.seek(0)
            self.s3_client.upload_fileobj(
                file_obj=f,
                bucket_name=self.config.psc_expiration_bucket,
                key_name=f'{to_iso_8601(utc_now()).replace(":", "-")}_{str(uuid4())}',
            )

    @staticmethod
    def persist_to_dynamodb(
        reputation_items: Iterable[ReputationItem], repository: ReputationDbRepository
    ) -> None:
        for item in reputation_items:
            item.version += 1

        repository.update(reputation_items)


def start_service(port):
    config = ReputationPersisterConfig(port)
    principal = TicPrincipal(config)
    service = ReputationPersister(config, principal)
    return service.app


# Entry point for mosaic2
def start_service_mcs2(port):
    config = RepPersisterConfig(port)
    principal = PersisterPrincipal()
    service = ReputationPersister(config=config, principal=principal, is_mcs2=True)
    return service.app
