# Native imports
import json
from retrying import retry
import traceback

# Third-party imports
import requests

# Program imports
from .apc_exceptions import InvalidApcResponse, ApcUnavailable
from .dtos.apc_query import ApcQuery
from .enums.apc_status import ApcStatus
from tic.shared.utils.exception_util import RetryableError
from ..utils.util import balanced_chunk


class ApcClient(object):
    """
    Client for working with the APC.
    """

    BASE_URI = "https://query-api.eu1.apc.avira.com"
    QUERY_URI = "{}/files/v1/query".format(BASE_URI)

    def __init__(self, api_key, http_session=None):
        """
        Initializes the class with the provided API key.

        @param api_key: The API key to be used to make requests to APC
        @param http_session: A requests HTTP session. If none is provided, one will be created.
        """

        self.api_key = api_key
        # avira would like a static ID so they can track our requests.
        self.user_randid = "b2f98b43-7721-4606-a409-72445a1973cb"

        if http_session is None:
            self.http_session = requests.Session()
            adapter = requests.adapters.HTTPAdapter(pool_maxsize=100)
            self.http_session.mount("https://", adapter)
            self.http_session.headers.update({"X-AVIRA-APIKEY": self.api_key})
        else:
            self.http_session = http_session

    def query(self, sha256s, set_flag=False, metadata=None, test=False):
        """
        Queries the APC with the provided SHA256s.

        @param sha256s: Iterable of SHA256s.
        @param set_flag: Sets flag to 1 indicating upload
        @param metadata: Any APC metadata.
        @param test: Optional override to provide a test key for APC tracking
        @return: List of APC DTOs.
        """

        headers = {"Content-Type": "application/json"}
        to_return = []

        # APC hard limit of 50 per request
        for chunk in balanced_chunk(sha256s, 50, allow_size_increase=False):
            body = {"sha256": {sha256.lower(): {} for sha256 in chunk}, "metadata": {}}
            if set_flag:
                body["flags"] = 1
            if metadata:
                body["metadata"] = metadata
            body["metadata"]["user_randid"] = (
                self.user_randid if not test else "cbtest-{}".format(self.user_randid)
            )

            try:
                response = self.http_session.post(
                    self.QUERY_URI, data=json.dumps(body), headers=headers
                )
                self._response_handler(response)
                try:
                    to_return.extend(ApcQuery.from_json(response.text))
                except Exception:
                    raise InvalidApcResponse(response)
            # Retries using retry stream (DO NOT use retrying because it will delay lambda and back up primary stream)
            except ApcUnavailable as e:
                raise RetryableError(e, traceback.format_exc())

        return to_return

    # retry if 502 bad gateway up to 3 times on exponential backoff (1.5s total)
    @retry(
        retry_on_exception=lambda exception: isinstance(exception, ApcUnavailable),
        stop_max_attempt_number=3,
        wait_exponential_multiplier=100,
    )
    def upload(self, fp, upload_url, metadata=None, gen_data=None, test=False):
        """
        Uploads a file to the APC. This must be done after calling query with upload_file=True.

        @param fp: A file-like object open in binary mode. This may be the file or the file
        compressed with bzip2.
        @param upload_url: URL of where to upload the file.
        @param metadata: Optional metadata.
        @param gen_data: Optional generic data.
        @param test: Optional override to provide a test key for APC tracking
        """

        files = [("file", fp)]
        optional_params = {}

        if metadata is not None:
            optional_params["metadata"] = metadata
        else:
            optional_params["metadata"] = {}
        if gen_data is not None:
            optional_params["gen_data"] = gen_data
        optional_params["metadata"]["user_randid"] = (
            self.user_randid if not test else "cbtest-{}".format(self.user_randid)
        )

        files.append(
            ("params", ("params.json", json.dumps(optional_params), "application/json"))
        )

        response = self.http_session.post(upload_url, files=files)
        self._response_handler(response)

    def query_and_upload(self, sha256, fp, metadata=None, gen_data=None, test=False):
        """
        Query the APC for the hash, and if necessary, upload the file.

        @param sha256: SHA256 of the file.
        @param fp: A file-like object open in binary mode. This may be the file or the file
        compressed with bzip2.
        @param metadata: Optional metadata.
        @param gen_data: Optional generic data.
        @param test: Optional override to provide a test key for APC tracking
        @return:
        """

        dto = self.query([sha256], True, metadata, test)[0]
        if dto.status == ApcStatus.UPLOAD:
            self.upload(fp, dto.upload_url, metadata, gen_data, test)
        return dto

    @staticmethod
    def _response_handler(response):
        if response.status_code != 200:
            raise InvalidApcResponse.new(response)
