import json
from typing import Dict, Iterable, Optional
from uuid import uuid4
from tic.shared.utils.util import message_chunk_batch


class KinesisClient(object):
    def __init__(self, boto_session, stream_name, boto_config=None, endpoint_url=None):
        """
        @param boto_session: Pre-existing boto session
        @param stream_name: A name to identify the stream
        @param boto_config: Optional boto config for the kinesis client
        """

        self.stream_name = stream_name
        self.kinesis_client = boto_session.client(
            "kinesis", config=boto_config, endpoint_url=endpoint_url
        )

    def put_records(
        self,
        records_map: Dict[Optional[str], Iterable],
        num_records_per_batch: int = 500,
    ):
        """
        Writes multiple data records into a Kinesis stream in a single call
        @param records_map: Map of partition_key: iterable[records to insert with key].
            Partition key uses random guid if not provided to evenly distribute records among shards by default.
        @param num_records_per_batch: The number of records to be sent in one put_records call
        @return any messages that failed
        """
        failed_data = list()

        kinesis_records = self._convert_to_kinesis_records(records_map)

        # Batch based on kinesis client put_records limit of 500 records or 5 MB limit
        # max message size = 5212880 = 5MB - (numRecords * partitionKeySize)
        # where random partitionKey size = 60, a provided partitionKey Size = 73 (orgkey_in_hex-sha256)
        #       numRecords = 500, 5MB = 5242880
        # this should only count the data section against the max message size, but this is an estimation.
        for batch in message_chunk_batch(
            kinesis_records, num_records_per_batch, max_message_size=5212880
        ):
            failed_data.extend(self._put_and_get_failed_items(batch))

        return failed_data

    def _convert_to_kinesis_records(self, records_map):
        kinesis_records = list()
        for pk, pk_records in records_map.items():
            kinesis_records.extend(
                [
                    {
                        "Data": record.decode("utf-8")
                        if isinstance(record, bytes)
                        else record,
                        "PartitionKey": pk or str(uuid4()),
                    }
                    for record in pk_records
                ]
            )
        return kinesis_records

    def put_records_and_get_failed_records_with_error_code(
        self,
        records_map: Dict[Optional[str], Iterable],
        num_records_per_batch: int = 500,
    ):
        """
        Writes multiple data records into a Kinesis stream in a single call
        @param records_map: Map of partition_key: iterable[records to insert with key].
            Partition key uses random guid if not provided to evenly distribute records among shards by default.
        @param num_records_per_batch: The number of records to be sent in one put_records call
        @return any messages that failed
        """
        failed_data = list()

        kinesis_records = self._convert_to_kinesis_records(records_map)

        # Batch based on kinesis client put_records limit of 500 records or 5 MB limit
        # max message size = 5212880 = 5MB - (numRecords * partitionKeySize)
        # where random partitionKey size = 60, a provided partitionKey Size = 73 (orgkey_in_hex-sha256)
        #       numRecords = 500, 5MB = 5242880
        # this should only count the data section against the max message size, but this is an estimation.
        for batch in message_chunk_batch(
            kinesis_records, num_records_per_batch, max_message_size=5212880
        ):
            failed_data.extend(self._put_and_get_failed_items_with_error_code(batch))

        return failed_data

    def put_records_aggregated(
        self, records_map: Dict[Optional[str], Iterable], delimiter="\n"
    ):
        """
        Aggregates and publishes records to kinesis
        @param records_map: Map of partition_key: iterable[records to insert with key].
            Partition key uses random guid if not provided to evenly distribute records among shards by default.
        @param delimiter: the delimiter to use during aggregation
        @return: List of records {"Data": record contents, "PartitionKey": pk} that failed to put to kinesis
        """

        records_to_send = list()
        for pk, pk_records in records_map.items():
            pk1 = str(uuid4()) if pk is None else pk
            record_struct = {"Data": "", "PartitionKey": pk1}

            # charged on 25k, but 1mb max size. If one item > 25k, put and continue
            record_max_size = 1048576 - len(
                json.dumps(record_struct, ensure_ascii=False).encode("utf-8")
            )
            record_target_size = 25600 - len(
                json.dumps(record_struct, ensure_ascii=False).encode("utf-8")
            )

            # handle any records that are already at max record size.
            pk_records_agg = list(pk_records)
            for record in pk_records:
                if self._check_for_record_size(
                    record, record_target_size, record_max_size
                ):
                    pk2 = str(uuid4()) if pk is None else pk
                    records_to_send.append({"Data": record, "PartitionKey": pk2})
                    pk_records_agg.remove(record)

            # chunk items into records of target cost size: 25k
            # If we chunk items in the largest possible record - don't we run the risk of overwhelming a single shard
            for batch in message_chunk_batch(
                pk_records_agg, max_message_size=record_target_size, delimiter=delimiter
            ):
                pk3 = str(uuid4()) if pk is None else pk
                records_to_send.append({"Data": "".join(batch), "PartitionKey": pk3})

        failed_items = list()
        # group into record batches based on total size and num records: 5MB/500 records per call
        for records_batch in message_chunk_batch(records_to_send, 500, 5212880):
            failed_items.extend(self._put_and_get_failed_items(records_batch))

        return failed_items

    @staticmethod
    def _check_for_record_size(record, target_size, max_size):
        if isinstance(record, dict):
            record = json.dumps(record, ensure_ascii=False)
        elif isinstance(record, bytes):
            record = record.decode("utf-8")

        if len(record) >= target_size:
            if len(record) >= max_size:
                raise Exception(
                    "Record is larger than specified max size: {}".format(record)
                )
            return True
        return False

    def _put_and_get_failed_items(self, records):
        """
        puts data to kinesis and parses for failed records
        @param records: the records to put to kinesis
        @return: any records that failed
        """
        failed_records = list()

        boto_resp = self.kinesis_client.put_records(
            Records=records, StreamName=self.stream_name
        )
        for i, record in enumerate(boto_resp["Records"]):
            if record.get("ErrorCode", None) is not None:
                failed_records.append(records[i])

        return failed_records

    def _put_and_get_failed_items_with_error_code(self, records):
        """
        puts data to kinesis and parses for failed records
        @param records: the records to put to kinesis
        @return: any records that failed with error code
        """
        failed_records = list()

        boto_resp = self.kinesis_client.put_records(
            Records=records, StreamName=self.stream_name
        )
        for i, record in enumerate(boto_resp["Records"]):
            if record.get("ErrorCode", None) is not None:
                error_code = record.get("ErrorCode", None)
                failed_records.append(
                    {"error": error_code, "failed-record": records[i]}
                )

        return failed_records
