import os

import boto3
import statsd

from .nop import Nop
from .tic_status import tic_status
from .statsd_aggregator import StatsDAggregator
from tic.shared.api_doc_settings import ApiDocSettings
from tic.shared.constants import TICEnvironment
from tic.shared.utils import log_util


class BaseConfig(object):
    """
    Base configuration class that is leveraged by all services.

    For mosaic v1 services TICConfiguration is used.

    For mosaic v2 services, this is the only class used.

    CAUTION: The get method will be overwritten in some cases.
             Do not call it directly in the init of this class.
             Instead add stuff to load_core or pre_init as needed.
             This is gross, but temporary until TICConfiguration can be
             nuked or heavily refactored.
    """

    def __init__(
        self,
        service_name,
        port,
        statsd_prefix=None,
        statsd_aggregation_time: int = 0,
        api_docs_settings: ApiDocSettings = None,
        pre_init=False,
        access_key_id=None,
        access_key=None,
        init_boto_session=True,
    ):
        # This ugliness is to keep things as DRY as possible w/o rewriting everything
        if not pre_init:
            self.commit_sha = ""
            self.raw_component_name = ""
            self.tic_environment = TICEnvironment.dev
            self.baseConfig = {}  # Required by TICService
            self.config = {}  # Required by TICService
            self.pre_init(service_name)

        # These are the core values that are expected to be overwritten
        # with the calling method.
        self.aws_region = None
        self.logging_level = None
        self.statsd_host = None
        self.statsd_port = 8125
        self.load_core()

        # Setup component name as function of CI input
        # TODO: Set this up for mosaic v2
        self.ci_postfix = (
            f"-{self.commit_sha}"
            if self.commit_sha is not None and len(self.commit_sha) == 8
            else ""
        )
        self.component_name = self.raw_component_name + self.ci_postfix

        # Setup boto
        self.boto_session = None
        if init_boto_session:
            self.setup_boto_session(self.aws_region, access_key_id, access_key)

        # Setup API doc settings
        self.api_docs_settings = (
            api_docs_settings
            if api_docs_settings is not None
            else ApiDocSettings(service_name, f"{service_name} service")
        )

        # Setup logging
        self.logger = log_util.get_logger(self.component_name, self.logging_level)

        # Setup metrics
        self.statsd_prefix = statsd_prefix or self.raw_component_name
        statsd_prefix = f"tic.{self.statsd_prefix.lower()}"
        # Normalize time to integers
        statsd_aggregation_time = int(statsd_aggregation_time)
        if not self.statsd_host:
            self.statsd = self.statsd_tcp = Nop()
        elif not statsd_aggregation_time or statsd_aggregation_time < 1:
            self.statsd = statsd.StatsClient(
                host=self.statsd_host, port=self.statsd_port, prefix=statsd_prefix
            )
            self.statsd_tcp = statsd.TCPStatsClient(
                host=self.statsd_host,
                port=self.statsd_port,
                prefix=statsd_prefix,
                timeout=10,
            )
        else:
            self.statsd = StatsDAggregator(
                host=self.statsd_host,
                port=self.statsd_port,
                prefix=statsd_prefix,
                flush_cadence_seconds=statsd_aggregation_time,
            )
            self.statsd_tcp = StatsDAggregator(
                host=self.statsd_host,
                port=self.statsd_port,
                prefix=statsd_prefix,
                flush_cadence_seconds=statsd_aggregation_time,
                tcp=True,
            )

        # Other setup
        self.port = port
        self._loaded = False
        self._printed_config = False

    def setup_boto_session(self, aws_region, access_key_id=None, access_key=None):
        if access_key_id and access_key:
            self.boto_session = boto3.Session(
                aws_access_key_id=access_key_id,
                aws_secret_access_key=access_key,
                region_name=aws_region,
            )
        else:
            self.boto_session = boto3.Session(region_name=aws_region)

    def pre_init(self, service_name):
        self.tic_environment = TICEnvironment[
            os.environ.get("TIC_ENVIRONMENT", TICEnvironment.dev.name)
        ]
        self.raw_component_name = service_name or self.__class__.__name__.lower()
        self.commit_sha = os.environ.get("CI_COMMIT_SHORT_SHA")

    def load_core(self):
        """
        Loads the core config values.
        This may be overwritten by a child.
        """

        # Try official region first and fallback to LocalStack
        self.aws_region = self.get("AWS_REGION")
        if self.aws_region is None:
            self.aws_region = self.get("AWS_DEFAULT_REGION")

        self.logging_level = self.get("LOG_LEVEL", default="INFO")
        self.statsd_host = self.get("STATSD_HOST")
        self.statsd_port = int(self.get("STATSD_PORT", 8125))

    @property
    def loaded(self):
        return self._loaded and tic_status.config_loaded

    @loaded.setter
    def loaded(self, _loaded):
        self._loaded = _loaded
        tic_status.config_loaded = _loaded

        if self._loaded and not self._printed_config:
            self.logger.info(f"Loaded Config for environment {self.tic_environment}")

    # Note that this gets overridden by TICConfiguration
    # Needs to not be static, since it's not static in children
    def get(self, config_key, default=None):
        """
        Retrieve elements from environment variables.

        :param config_key: The key to retrieve.
        :param default: The default value to provide.
        :return: The retrieved item.
        """

        return os.environ.get(config_key, default)
