import configparser

import glob
import os

# TIC configuration is internally represented by a python dict
# Configuration follows this structure:
# config["RootKey"]["SectionKey"]["ConfigKey"] = "value"
#
# Configuration can be stored in:
# - Environment variables
# - Configuration files
#
# Purpose of this module is to help with loading/saving configuration
# from/to configuration files


def load_config_from_files(environment, root_key=None, config_location=None):
    """
    Load configuration from INI config files with .conf extension
    Default locations are:
    ~/.tic/config
    /etc/tic/config

    If environment provided location is also:
    ~/.tic/config/<environment>

    :param environment: "dev","test","stage" or "prod"
    :return: dict of all config values
    """
    search_pattern = "*.conf"
    ret_config = {}

    # Default config file locations
    config_locations = [
        "/etc/tic/config",
        "/etc/tic/base_config",
        os.path.expanduser("~/.tic/config"),
    ]

    # Environment custom config file specified
    if environment:
        config_locations.append(
            os.path.join(os.path.expanduser("~/.tic/"), "config", environment.name)
        )

    # User custom config file specified
    if config_location:
        config_locations.append(config_location)

    # Specify config file to search for
    if root_key:
        print(f"only looking for RootKey [{root_key}]")

    for config_location in config_locations:
        config_files = glob.glob(os.path.join(config_location, search_pattern))
        for config_file in config_files:
            if os.path.isfile(config_file):
                try:
                    print(f"parsing config file [{config_file}]")

                    config = configparser.ConfigParser()
                    config.read(config_file)

                    # Test RootKey exists (for config file search, will be removed after config file found)
                    assert config.has_section("TIC")

                    # Case: We are not searching for a specific RootKey
                    if root_key is None:
                        print("\t+ utilizing this config file")
                        pass
                    # Case: We are searching for a specific RootKey, and this file doesn't contain the one we want
                    elif config["TIC"]["RootKey"] != root_key:
                        print("\t- skipping config file (RootKey doesn't match)")
                        print(f"\t- RootKey Desired: [{root_key}]")
                        print(f"\t- RootKey Found:   [{config.get('TIC', 'RootKey')}]")
                        continue
                    # Case: We are searching for a specific RootKey, and this file contains it
                    else:
                        print("\t+ utilizing this config file")
                        pass

                    # Remove TIC section from the config
                    config.remove_section("TIC")
                    assert config.has_section("TIC") is False

                    for section in config.sections():
                        if len(config.options(section)) == 0:
                            print(f"\t* WARN: empty section [{section}] found")
                        else:
                            # ex: {'Section1': {'key2': 'value2 in Section1', 'key1': 'value1 in Section1'}}
                            ret_config[section] = dict(config.items(section))

                except Exception as e:
                    print(f"\t- exception parsing config file ({str(e)})")
                    continue

    if ret_config:
        return {root_key: ret_config}
    else:
        return {}
