from time import sleep
from threading import Thread, Event

import statsd


class StatsDAggregator:
    def __init__(
        self,
        host: str,
        port: int,
        prefix: str = None,
        flush_cadence_seconds: int = 60,
        tcp: bool = False,
    ):
        if not tcp:
            self.statsd = statsd.StatsClient(host=host, port=port, prefix=prefix)
        else:
            self.statsd = statsd.TCPStatsClient(
                host=host,
                port=port,
                prefix=prefix,
                timeout=10,
            )
        self.agg_dict = dict()

        self.flush_cadence_seconds = flush_cadence_seconds
        self._stop = Event()
        self._start_metrics_thread()

    def _start_metrics_thread(self):
        # daemon thread will exit with the main program.
        self.loop_thread = Thread(target=self._metrics_loop, daemon=True)
        self.loop_thread.start()

    def _metrics_loop(self):
        while not self._stop.is_set():
            sleep(self.flush_cadence_seconds)
            self.flush()

    def stop_metrics_loop(self):
        self._stop.set()
        self.flush()

    def _aggregate(self, call_type: str, *args):
        self.agg_dict.setdefault(call_type, list())
        self.agg_dict[call_type].append(args)

    def incr(self, *args):
        self._aggregate("incr", *args)

    def decr(self, *args):
        self._aggregate("decr", *args)

    def timing(self, *args):
        self._aggregate("timing", *args)

    def gauge(self, *args):
        self._aggregate("gauge", *args)

    def set(self, *args):
        self._aggregate("set", *args)

    def flush(self):
        with self.statsd.pipeline() as pipe:
            for call_type, calls_args in self.agg_dict.items():
                pipeline_func = getattr(pipe, call_type)
                for call_args in calls_args:
                    pipeline_func(*call_args)
                self.agg_dict[call_type] = list()
