# coding: utf-8
from sqlalchemy import (
    Integer,
    BigInteger,
    Boolean,
    CheckConstraint,
    Column,
    DateTime,
    ForeignKey,
    Index,
    LargeBinary,
    SmallInteger,
    String,
    text,
)
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()
metadata = Base.metadata


class Magic(Base):
    __tablename__ = "magic"
    __table_args__ = (
        CheckConstraint("length(sha256_value) = 32"),
        {"schema": "ubs_state"},
    )

    id = Column(
        Integer,
        primary_key=True,
        server_default=text("nextval('magic_id_seq'::regclass)"),
    )
    value = Column(String(1024), nullable=False)
    sha256_value = Column(LargeBinary, nullable=False, unique=True)
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))


class Main(Base):
    __tablename__ = "main"
    __table_args__ = (CheckConstraint("length(sha256) = 32"), {"schema": "ubs_state"})

    id = Column(
        BigInteger,
        primary_key=True,
        server_default=text("nextval('ubs_state.main_id_seq'::regclass)"),
    )
    sha256 = Column(LargeBinary, nullable=False, unique=True)
    sha256_truncated = Column(LargeBinary, nullable=True)
    s3_key = Column(String(1024), nullable=True)
    truncated = Column(Boolean, nullable=True)
    magic_id = Column(ForeignKey("ubs_state.magic.id"), nullable=True)
    date_created = Column(
        DateTime(True), nullable=False, index=True, server_default=text("now()")
    )
    date_modified = Column(DateTime(True), nullable=True)

    magic = relationship("Magic")


class Status(Base):
    __tablename__ = "status"
    __table_args__ = {"schema": "ubs_state"}

    id = Column(
        SmallInteger,
        primary_key=True,
        server_default=text("nextval('ubs_state.status_id_seq'::regclass)"),
    )
    name = Column(String, nullable=False, unique=True)
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))


class Version(Base):
    __tablename__ = "version"
    __table_args__ = (
        Index("idx_version_major_minor", "major", "minor", unique=True),
        {"schema": "ubs_state"},
    )

    id = Column(
        SmallInteger,
        primary_key=True,
        server_default=text("nextval('ubs_state.version_id_seq'::regclass)"),
    )
    major = Column(SmallInteger, nullable=False)
    minor = Column(SmallInteger, nullable=False)
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))


class Partner(Base):
    __tablename__ = "partner"
    __table_args__ = {"schema": "ubs_state"}

    id = Column(
        SmallInteger,
        primary_key=True,
        server_default=text("nextval('partner_id_seq'::regclass)"),
    )
    name = Column(String(128), nullable=False)
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))


class Pipeline(Base):
    __tablename__ = "pipeline"
    __table_args__ = {"schema": "ubs_state"}

    id = Column(
        SmallInteger,
        primary_key=True,
        server_default=text("nextval('ubs_state.pipeline_id_seq'::regclass)"),
    )
    name = Column(String, nullable=False, unique=True)
    current_version_id = Column(ForeignKey("ubs_state.version.id"), nullable=False)
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))
    date_modified = Column(DateTime(True), nullable=False, server_default=text("now()"))

    current_version = relationship("Version")


class Stage(Base):
    __tablename__ = "stage"
    __table_args__ = {"schema": "ubs_state"}

    id = Column(
        SmallInteger,
        primary_key=True,
        server_default=text("nextval('ubs_state.stage_id_seq'::regclass)"),
    )
    name = Column(String, nullable=False, unique=True)
    current_version_id = Column(ForeignKey("ubs_state.version.id"), nullable=False)
    requires_file = Column(Boolean, nullable=False, server_default=text("false"))
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))
    date_modified = Column(DateTime(True), nullable=False, server_default=text("now()"))

    current_version = relationship("Version")


class MainPipelineLink(Base):
    __tablename__ = "main_pipeline_link"
    __table_args__ = (
        Index(
            "idx_main_pipeline_link_main_id_pipeline_id",
            "main_id",
            "pipeline_id",
            unique=True,
        ),
        {"schema": "ubs_state"},
    )

    id = Column(
        BigInteger,
        primary_key=True,
        server_default=text("nextval('ubs_state.main_pipeline_link_id_seq'::regclass)"),
    )
    main_id = Column(ForeignKey("ubs_state.main.id"), nullable=False)
    pipeline_id = Column(ForeignKey("ubs_state.pipeline.id"), nullable=False)
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))

    main = relationship("Main")
    pipeline = relationship("Pipeline")


class MainStageLink(Base):
    __tablename__ = "main_stage_link"
    __table_args__ = (
        Index(
            "idx_main_stage_link_main_id_stage_id", "main_id", "stage_id", unique=True
        ),
        {"schema": "ubs_state"},
    )

    id = Column(
        BigInteger,
        primary_key=True,
        server_default=text("nextval('ubs_state.main_stage_link_id_seq'::regclass)"),
    )
    main_id = Column(ForeignKey("ubs_state.main.id"), nullable=False)
    stage_id = Column(ForeignKey("ubs_state.stage.id"), nullable=False)
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))

    main = relationship("Main")
    stage = relationship("Stage")


class MainPartnerLink(Base):
    __tablename__ = "main_partner_link"
    __table_args__ = (
        Index(
            "idx_main_id_partner_id",
            "partner_id",
            "main_id",
            unique=True,
        ),
        {"schema": "ubs_state"},
    )

    id = Column(
        BigInteger,
        primary_key=True,
        server_default=text("nextval('ubs_state.main_partner_link_id_seq'::regclass)"),
    )
    partner_id = Column(ForeignKey("ubs_state.partner.id"), nullable=False, index=True)
    main_id = Column(ForeignKey("ubs_state.main.id"), nullable=False, index=True)
    response_state = Column(SmallInteger, nullable=False)
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))
    date_modified = Column(DateTime(True), nullable=False, server_default=text("now()"))

    main = relationship("Main")
    partner = relationship("Partner")


class PipelineStageLink(Base):
    __tablename__ = "pipeline_stage_link"
    __table_args__ = (
        Index(
            "idx_pipeline_stage_link_pipeline_id_stage_id",
            "pipeline_id",
            "stage_id",
            unique=True,
        ),
        {"schema": "ubs_state"},
    )

    id = Column(
        SmallInteger,
        primary_key=True,
        server_default=text(
            "nextval('ubs_state.pipeline_stage_link_id_seq'::regclass)"
        ),
    )
    pipeline_id = Column(ForeignKey("ubs_state.pipeline.id"), nullable=False)
    stage_id = Column(ForeignKey("ubs_state.stage.id"), nullable=False)
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))

    pipeline = relationship("Pipeline")
    stage = relationship("Stage")


class StageTarget(Base):
    __tablename__ = "stage_target"
    __table_args__ = (
        Index(
            "idx_stage_target_stage_id_target_stage_id",
            "stage_id",
            "target_stage_id",
            unique=True,
        ),
        {"schema": "ubs_state"},
    )

    id = Column(
        SmallInteger,
        primary_key=True,
        server_default=text("nextval('ubs_state.stage_target_id_seq'::regclass)"),
    )
    stage_id = Column(ForeignKey("ubs_state.stage.id"), nullable=False)
    target_stage_id = Column(ForeignKey("ubs_state.stage.id"), nullable=False)
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))
    date_modified = Column(DateTime(True), nullable=False, server_default=text("now()"))

    stage = relationship("Stage", primaryjoin="StageTarget.stage_id == Stage.id")
    target_stage = relationship(
        "Stage", primaryjoin="StageTarget.target_stage_id == Stage.id"
    )


class StageTopic(Base):
    __tablename__ = "stage_topic"
    __table_args__ = {"schema": "ubs_state"}

    id = Column(
        SmallInteger,
        primary_key=True,
        server_default=text("nextval('ubs_state.stage_topic_id_seq'::regclass)"),
    )
    stage_id = Column(ForeignKey("ubs_state.stage.id"), nullable=False, unique=True)
    sns_arn = Column(String, nullable=False)
    enabled = Column(Boolean, nullable=False, server_default=text("false"))
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))
    date_modified = Column(DateTime(True), nullable=False, server_default=text("now()"))

    stage = relationship("Stage")


class PipelineStatus(Base):
    __tablename__ = "pipeline_status"
    __table_args__ = (
        Index(
            "idx_pipeline_status_main_pipeline_link_id_version_id",
            "main_pipeline_link_id",
            "version_id",
            unique=True,
        ),
        {"schema": "ubs_state"},
    )

    id = Column(
        BigInteger,
        primary_key=True,
        server_default=text("nextval('ubs_state.pipeline_status_id_seq'::regclass)"),
    )
    main_pipeline_link_id = Column(
        ForeignKey("ubs_state.main_pipeline_link.id"), nullable=False
    )
    version_id = Column(ForeignKey("ubs_state.version.id"), nullable=False)
    status_id = Column(ForeignKey("ubs_state.status.id"), nullable=False)
    completion_date = Column(
        DateTime(True), nullable=False, server_default=text("now()")
    )
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))
    date_modified = Column(DateTime(True), nullable=False, server_default=text("now()"))

    main_pipeline_link = relationship("MainPipelineLink")
    status = relationship("Status")
    version = relationship("Version")


class StageStatus(Base):
    __tablename__ = "stage_status"
    __table_args__ = (
        Index(
            "idx_stage_status_main_stage_link_id_version",
            "main_stage_link_id",
            "version_id",
            unique=True,
        ),
        {"schema": "ubs_state"},
    )

    id = Column(
        BigInteger,
        primary_key=True,
        server_default=text("nextval('ubs_state.stage_status_id_seq'::regclass)"),
    )
    main_stage_link_id = Column(
        ForeignKey("ubs_state.main_stage_link.id"), nullable=False
    )
    version_id = Column(ForeignKey("ubs_state.version.id"), nullable=False)
    status_id = Column(ForeignKey("ubs_state.status.id"), nullable=False)
    completion_date = Column(
        DateTime(True), nullable=False, server_default=text("now()")
    )
    date_created = Column(DateTime(True), nullable=False, server_default=text("now()"))
    date_modified = Column(DateTime(True), nullable=False, server_default=text("now()"))

    main_stage_link = relationship("MainStageLink")
    status = relationship("Status")
    version = relationship("Version")
