from binascii import hexlify, unhexlify
from ...shared.ubs.models.models import CustIxLink
from ...shared.ubs.dtos.customer_access import CustomerAccess


def generate_s3_key(sha256):
    split_sha256 = [sha256[i : i + 2] for i in range(0, len(sha256), 2)]
    split_sha256.append(sha256)
    return "/".join(split_sha256) + ".zip"


def get_sha256_from_s3_key(key):
    return key.split("/")[-1].replace(".zip", "")


def hash_to_binary(hash_value):
    if type(hash_value) == str:
        hash_value = hash_value.encode("utf-8")
    return unhexlify(hash_value)


def binary_to_hash(binary_value, type=str):
    hash_value = hexlify(binary_value)
    if type == str:
        return hash_value.decode("utf-8")
    return hash_value


machine_architectures = [
    "none",  # ARCH_NONE 0x00
    "x86",  # ARCH_X86 0x01
    "amd64",  # ARCH_AMD64 0x02
    "arm64",  # ARCH_ARM64 0x04
]

signature_states = [
    "invalid",  # FILE_SIGNATURE_STATE_INVALID 0x00
    "signed",  # FILE_SIGNATURE_STATE_SIGNED 0x01
    "verified",  # FILE_SIGNATURE_STATE_VERIFIED 0x02
    "not_signed",  # FILE_SIGNATURE_STATE_NOT_SIGNED 0x04
    "unknown",  # FILE_SIGNATURE_STATE_UNKNOWN 0x08
    "chained",  # FILE_SIGNATURE_STATE_CHAINED 0x10
    "trusted",  # FILE_SIGNATURE_STATE_TRUSTED 0x20
    "os",  # FILE_SIGNATURE_STATE_OS 0x40
    "catalog_signed",  # FILE_SIGNATURE_STATE_CATALOG_SIGNED 0x80
]

signature_values = {
    "FILE_SIGNATURE_STATE_INVALID": 0x00,
    "FILE_SIGNATURE_STATE_SIGNED": 0x01,
    "FILE_SIGNATURE_STATE_VERIFIED": 0x02,
    "FILE_SIGNATURE_STATE_NOT_SIGNED": 0x04,
    "FILE_SIGNATURE_STATE_UNKNOWN": 0x08,
    "FILE_SIGNATURE_STATE_CHAINED": 0x10,
    "FILE_SIGNATURE_STATE_TRUSTED": 0x20,
    "FILE_SIGNATURE_STATE_OS": 0x40,
    "FILE_SIGNATURE_STATE_CATALOG_SIGNED": 0x80,
}

architecture_values = {
    "ARCH_NONE": 0x00,  # ARCH_NONE 0x00
    "ARCH_X86": 0x01,  # ARCH_X86 0x01
    "ARCH_AMD64": 0x02,  # ARCH_AMD64 0x02
    "ARCH_ARM64": 0x04,  # ARCH_ARM64 0x04
}


def convert_architectures_to_array(architecture):
    """
    Converts the architecture bit masked value to an array of strings of corresponding architectures
    @param architecture: The bitmask arch value to be converted to an array
    @return: An array of the strings of corresponding architectures
    """
    return convert_bitmask_to_array(architecture, machine_architectures)


def convert_sign_states_to_array(sign_state):
    """
    Converts the sign_state bit masked value to an array of strings of corresponding signature states
    @param sign_state: The bitmask sign_state value to be converted to an array
    @return: An array of the strings of corresponding signature states
    """
    return convert_bitmask_to_array(sign_state, signature_states)


def convert_bitmask_to_array(bitmask, conversion_array):
    """
    Converts the bitmask bit masked value to an array of strings of corresponding to the conversion_array
    @param bitmask: The bitmask value to be converted to an array
    @param conversion_array:
    @return: An array of the strings composed of elements of the array passed in
    """
    if bitmask is None:
        return []

    result = []
    for val in range(1, len(conversion_array)):

        if (0x01 << (val - 1)) & bitmask:
            result.append(conversion_array[val])

    if not result:
        result.append(conversion_array[0])

    return result


def get_batch_cust_link_and_status_code(repository, search_key, sha256_strs, reader):
    """
    calls get_cust_access and determines status code based on values returned
    @param repository: Repository instance
    @param search_key: Search key for the customer in question
    @param sha256_strs: The file hashes being searched for
    @return: status_code: 404 if cust_id is None; 404 if file_index is None; 200 if successful.
             cust_ix_link_id: id. limit access in calls to this method based on returned status code.
    """
    result = repository.get_batch_cust_access(search_key, sha256_strs, reader)

    if not result:
        message = "unauthorized"
        return [CustomerAccess(None, 404, None, message)]

    link_list = []

    for CustLinkDetail in result:
        cust_id = CustLinkDetail.cust_id
        cust_ix_link_id = CustLinkDetail.cust_ix_link_id
        file_index_id = CustLinkDetail.file_index_id
        sha256 = CustLinkDetail.sha256

        if file_index_id is None or cust_id is None:
            message = "{} was not found".format(sha256)
            link_list.append(CustomerAccess(sha256, 404, None, message))

        else:
            cust_ix_link = CustIxLink(
                cust_id=cust_id, file_index_id=file_index_id, id=cust_ix_link_id
            )

            message = "success"
            link_list.append(CustomerAccess(sha256, 200, cust_ix_link, message))

    return link_list
