import logging
import sys
import traceback
import flask

from pythonjsonlogger import jsonlogger
from flask import g
from tic import __version__

FORMAT = "%(asctime)s [%(name)s] [%(levelname)s] [%(request_id)s] [{0}]:\t%(message)s".format(
    __version__
)


def get_logger(component_name, log_level_name=None):

    # ELK Does not handle dashes well - replacing them with _ won't break them out into separate values
    component_name = component_name.replace("-", "_")

    logger = logging.getLogger(component_name)
    log_level = {
        "CRITICAL": logging.CRITICAL,
        "ERROR": logging.ERROR,
        "WARNING": logging.WARNING,
        "INFO": logging.INFO,
        "DEBUG": logging.DEBUG,
    }.get(log_level_name, logging.DEBUG)

    logger.setLevel(log_level)

    if not any(
        x.stream == sys.stdout
        for x in logger.handlers
        if isinstance(x, logging.StreamHandler)
    ):
        std_out_logger = logging.StreamHandler(sys.stdout)
        std_out_logger.setFormatter(jsonlogger.JsonFormatter(FORMAT))
        logger.addHandler(std_out_logger)

    logger.log(
        log_level,
        "Logging Configured at {0}:{1}".format(
            logging.getLevelName(log_level), log_level
        ),
    )

    logger = TicLoggerAdapter(logger, {"tic_version": __version__})

    return logger


def get_stdout_logger(component_name):
    logger = logging.getLogger(component_name)
    logger.setLevel(logging.DEBUG)
    stdOutLogger = logging.StreamHandler(sys.stdout)
    stdOutLogger.setFormatter(jsonlogger.JsonFormatter(FORMAT))
    logger.addHandler(stdOutLogger)
    return logger


# https://gist.github.com/danielrichman/1acbbd66166ab857c404
class TicLoggerAdapter(logging.LoggerAdapter):
    def __init__(self, logger, extra):
        self.extra = extra
        self.extra["tic_version"] = __version__

        super(TicLoggerAdapter, self).__init__(logger, self.extra)

    def process(self, msg, kwargs):
        try:
            add = {"request_id": kwargs.get("extra", {}).get("request_id", None)}
            message_id = kwargs.get("extra", {}).get("message_id", None)

            if flask.has_request_context():
                request_id = getattr(g, "request_id", None)
                if request_id:
                    add["request_id"] = request_id

                # add["flask_endpoint"] = flask.request.endpoint
                # add["remote_addr"] = flask.request.remote_addr
                # Example:
                # if getattr(flask.session, "ok", False):
                #     add["session_id"] = flask.session["session_id"]
                #     add["user_id"] = flask.session["user_id"]
            else:  # If there is no request context check for a message_id and use that as the request id.
                if message_id is not None:
                    add["request_id"] = message_id

            extra = kwargs.setdefault("extra", {})
            for key in add:
                extra.setdefault(key, add[key])

            kwargs["extra"] = self.extra
            kwargs["extra"].update(add)

        except Exception:
            traceback.print_exc()

        return msg, kwargs


class OptionalKeysFormatter(logging.Formatter):
    """
    Adapt LogRecords that haven't come from :class:`LoggerAdapter`

    This is a :class:`logging.Formatter` that fills in the keys that
    :class:`LoggerAdapter` adds with ``None`` if they are not already set,
    so that the same formatter may be used for records from
    :class:`LoggerAdapter` and regular loggers.
    """

    def __init__(self, format_string):
        super(OptionalKeysFormatter, self).__init__(fmt=format_string)

    fill_keys = ["request_id"]
    # fill_keys = ("request_path", "flask_endpoint", "remote_addr")  # Example: "session_id", "user_id")

    def format(self, record):
        for key in self.fill_keys:
            if not hasattr(record, key):
                setattr(record, key, None)

        return super(OptionalKeysFormatter, self).format(record)
