"""
Switch to using netaddr as it is what we use elsewhere and looks to be more mature/active. We also use it elsewhere in
the TIC already.

Added stricter validation of IP addresses. If a IPv4 address is supplied that isn't a dotted quad (1.1.1.1) this code
will no longer recognize the IP address.

Notes on IPv6 Network Ranges

IPv6 Addresses are 128-bit addresses that are allocated from IANA to RIR at a minimum unit of allocation of /12
    https://www.icann.org/resources/pages/allocation-ipv6-rirs-2012-02-25-en
"""

import re

from enum import Enum
from netaddr import IPAddress, IPNetwork, INET_PTON, AddrFormatError, ipv6_compact


_domain_regex = re.compile(r"^((?!-)[A-Za-z0-9-]{1,63}(?<!-)\.)+[A-Za-z]{2,30}$")


class AuthorityType(Enum):
    unknown = ""
    ipv4 = "ipv4"
    ipv6 = "ipv6"
    ipv6_subnet = "ipv6_subnet"
    domain = "domain"


def determine_authority_type(authority_value):
    authority_type = AuthorityType.unknown

    if authority_value is None:
        return authority_type, authority_value

    temp_authority = authority_value.strip()

    if _domain_regex.match(temp_authority):
        authority_type = AuthorityType.domain
        return authority_type, temp_authority

    try:
        # Possibly a IPv6 Subnet
        if "/" in temp_authority:
            ip_subnet = IPNetwork(temp_authority, version=6)
            authority_type = AuthorityType.ipv6_subnet
            temp_authority = str(ip_subnet)
        else:
            ip = IPAddress(temp_authority, flags=INET_PTON)
            if ip.version == 4:
                authority_type = AuthorityType.ipv4
            elif ip.version == 6:
                authority_type = AuthorityType.ipv6

                # Force the format of the IPv6 address into the compact format.
                temp_authority = ip.format(dialect=ipv6_compact)
    except AddrFormatError:
        pass

    return authority_type, temp_authority


def is_public_ipv6(ipv6_str):
    """
    A Public IPv6 Address or Network is defined as not being one of the following -
        http://www.iana.org/assignments/ipv6-address-space/ipv6-address-space.xhtml

    For more details see - Examples from https://en.wikipedia.org/wiki/Reserved_IP_addresses

    :param ip_address: IP address as a string
    :return: True if the IP is public - False if otherwise (including invalid IPv6 address)
    """

    try:
        # Possibly a IPv6 Subnet
        if "/" in ipv6_str:
            ip = IPNetwork(ipv6_str, version=6)
        else:
            ip = IPAddress(ipv6_str, version=6)
    except AddrFormatError:
        return False

    return (
        ip.is_unicast()
        and not ip.is_link_local()
        and not ip.is_loopback()
        and not ip.is_multicast()
        and not ip.is_private()
        and not ip.is_reserved()
    )


def is_public_ipv4(ip_address):
    """
    A Public IPv4 Address is defined as not being one of the following -
        Netmask (255.0.0.0/8) or hostmask (0.0.1.255)
        Not in a private range - 172.16.0.0/12, 192.168.0.0/16, 10.0.0.0/8
        Not a loopback address - 127.0.0.0/8
        Not a multicast
        Not a reserved range

    For more details see - Examples from https://en.wikipedia.org/wiki/Reserved_IP_addresses

    Use flags=INET_PTON to force a stricter validation.

    :param ip_address: IP address as a string
    :return: True if the IP is public - False if otherwise (including invalid IPv4 address)
    """
    try:
        ip = IPAddress(ip_address, version=4, flags=INET_PTON)
    except AddrFormatError:
        return False

    return (
        ip.is_unicast()
        and not ip.is_hostmask()
        and not ip.is_link_local()
        and not ip.is_loopback()
        and not ip.is_multicast()
        and not ip.is_netmask()
        and not ip.is_private()
        and not ip.is_reserved()
    )
