import json

from flask import jsonify, request

from .analysis_frontend_config import (
    UbsAnalysisFrontendConfiguration,
    UbsAnalysisFrontendConfig,
)
from .analysis_frontend_principal import UbsAnalysisFrontendPrincipal
from .shared.exceptions import ClientError
from tic.shared.aws.sqs_client import SqsClient
from tic.shared.aws.kinesis_client import KinesisClient
from tic.shared.service import TICService, TicPrincipal, create_ubs_data_permission
from tic.shared.ubs.dtos.analysis.state_update import UbsAnalysisStateUpdate
from tic.shared.ubs.enums.analysis.pipeline import UbsAnalysisPipeline
from tic.shared.ubs.enums.analysis.status import UbsAnalysisStatus
from tic.shared.ubs.repository.ubs_analysis_db import UbsAnalysisDbRepository
from tic.shared.utils.date_util import utc_now
from ...shared.utils.util import to_bool
from tic.shared.utils.parsers.hash import is_valid_sha256
from ...shared.ubs.dtos.analysis.pipeline_summary import PipelineSummary


class UbsAnalysisFrontend(TICService):
    def __init__(self, config, principal):
        super().__init__(config, principal)
        self._set_up_clients()
        self._set_up_routes()

    def _set_up_clients(self):
        self.sqs_client = SqsClient(self.boto_session)
        self.log_kinesis_client = KinesisClient(
            self.boto_session, self.config.logging_stream
        )
        self.ubs_analysis_repository = UbsAnalysisDbRepository(
            self.logger,
            self.config.ubap_writer_conn_str,
            self.config.ubap_reader_conn_str,
            connection_args=self.config.ssl_connection_args,
            statement_timeout=self.config.statement_timeout_ms,
        )

    def _set_up_routes(self):
        self.app.add_url_rule(
            "/v1/pipeline/start/<pipeline_name>/<sha256>",
            "start_pipeline",
            self.start_pipeline,
            methods=["POST"],
        )

    def _get_kinesis_log(self, pipeline_summary: PipelineSummary):
        """
        Get the kinesis log for the given pipeline_summary
        @param pipeline_summary: Summary for a given UBS Analysis Pipeline
        @return: JSON to log out to kinesis
        """
        summary_dict = pipeline_summary.to_dict(True, True)
        summary_dict["requested_by"] = self.tic_principal.customer_token
        summary_dict["requested_at"] = utc_now().strftime("%Y-%m-%d %H:%M:%S.%f")
        return json.dumps(summary_dict)

    @create_ubs_data_permission.require(http_exception=403)
    def start_pipeline(self, pipeline_name, sha256):
        start_time = utc_now()
        try:
            try:
                pipeline = UbsAnalysisPipeline[pipeline_name.upper()]
            except KeyError:
                raise ClientError(f"Invalid parameter {pipeline_name=}")
            if not is_valid_sha256(sha256):
                raise ClientError(f"Invalid parameter {sha256=}")

            force_flag = to_bool(request.args.get("force"))
            if force_flag:
                self.statsd.incr("force_restart_pipeline")

            with self.ubs_analysis_repository.ubs_writer_session() as writer:
                self.statsd.incr("write")
                pipeline_summary = self.ubs_analysis_repository.start_pipeline(
                    writer, sha256, pipeline, force_flag
                )
                self.logger.debug(
                    f"pipeline {pipeline_name} started for {sha256}. status={pipeline_summary.status}"
                )

            if pipeline_summary.status in (
                UbsAnalysisStatus.ERROR_RETRY,
                UbsAnalysisStatus.IN_PROGRESS,
                UbsAnalysisStatus.NOT_STARTED,
            ):

                # Use write instance for reading also to avoid replication latency issues.
                with self.ubs_analysis_repository.ubs_writer_session() as reader:
                    self.statsd.incr("read")
                    runnable_root_stages = self.ubs_analysis_repository.get_pipeline_executable_root_stages(
                        reader, sha256, pipeline
                    )

                    if len(runnable_root_stages) < 1:
                        self.logger.warning(
                            f"There are no root stages for pipeline {pipeline_name} and {sha256}."
                        )

                    for stage in runnable_root_stages:
                        self.logger.debug(
                            f"runnable stage {stage} for pipeline {pipeline_name} and {sha256}."
                        )

                sqs_messages = [
                    UbsAnalysisStateUpdate(
                        sha256=sha256,
                        name=stage["stage"],
                        version=stage["version"],
                        status=UbsAnalysisStatus.NOT_STARTED,
                    ).to_json()
                    for stage in runnable_root_stages
                ]
                self.sqs_client.send_message_batch(
                    self.config.controller_queue_url, sqs_messages
                )

                self.logger.debug(
                    f"Messages sent to {self.config.controller_queue_url}"
                )

            self.statsd.timing("call_time", utc_now() - start_time)
            self.log_kinesis_client.put_records(
                {None: [self._get_kinesis_log(pipeline_summary)]}
            )
            return jsonify(pipeline_summary.to_dict(True, True)), 200

        except ClientError as e:
            self.logger.warning(f"Client error: {str(e)}")
            self.statsd.incr("client_error")
            return jsonify({type(e).__name__: str(e)}), 400

        except Exception as e:
            self.logger.exception(f"Exception occurred: {str(e)}")
            self.statsd.incr("exception")
            return jsonify({type(e).__name__: str(e)}), 500
        finally:
            self.statsd.incr("call")


def start_service(port):
    config = UbsAnalysisFrontendConfiguration(port)
    principal = TicPrincipal(config)
    service = UbsAnalysisFrontend(config, principal)
    return service.app


def start_service_mcs2(port):
    config = UbsAnalysisFrontendConfig(port)
    principal = UbsAnalysisFrontendPrincipal()
    service = UbsAnalysisFrontend(config, principal)
    return service.app
