from tic.shared.base_config import BaseConfig
from tic.shared.service import TICConfiguration
from tic.shared.utils.db_util import build_postgresql_connection_string
from tic.shared.utils.util import to_bool


class UbsAnalysisFrontendConfiguration(TICConfiguration):
    def __init__(self, port):
        super().__init__("ubs_analysis", port, "ubs.analysis.frontend")
        self.ssl_connection_args = None
        self.ubs_writer_conn_str = None
        self.ubs_reader_conn_str = None
        self.controller_queue_url = None
        self.statement_timeout_ms = 100
        self.config_changed()

    def config_changed(self):
        self.ssl_connection_args = {
            "sslmode": self.get("ubs_db", "ssl_mode", default="verify-full"),
            "application_name": "ubs_analysis_frontend",
        }

        reader_host = self.get("ubs_db", "reader_host")
        ubap_reader_host = self.get("ubap_db", "reader_host")
        if self.commit_sha is not None and len(self.commit_sha) > 0:
            reader_host = reader_host.format(self.commit_sha)
            ubap_reader_host = reader_host.format(self.commit_sha)

        self.ubs_reader_conn_str = build_postgresql_connection_string(
            host=reader_host,
            port=self.get("ubs_db", "port"),
            db=self.get("ubs_db", "db"),
            username=self.get("ubs_db", "user"),
            password=self.get("ubs_db", "password"),
            decrypt=to_bool(self.get("ubs_db", "pw_encrypted")),
            aws_region=self.get("AWS", "aws_region"),
        )

        # create connection string to be used for cdc ubap services with separate user from ubs.
        self.ubap_reader_conn_str = build_postgresql_connection_string(
            host=ubap_reader_host,
            port=self.get("ubap_db", "port"),
            db=self.get("ubap_db", "db"),
            username=self.get("ubap_db", "user"),
            password=self.get("ubap_db", "password"),
            decrypt=to_bool(self.get("ubap_db", "pw_encrypted")),
            aws_region=self.get("AWS", "aws_region"),
        )

        writer_host = self.get("ubs_db", "writer_host")
        ubap_writer_host = self.get("ubap_db", "writer_host")
        if self.commit_sha is not None and len(self.commit_sha) > 0:
            writer_host = writer_host.format(self.commit_sha)
            ubap_writer_host = self.get("ubap_db", "writer_host")

        self.ubs_writer_conn_str = build_postgresql_connection_string(
            host=writer_host,
            port=self.get("ubs_db", "port"),
            db=self.get("ubs_db", "db"),
            username=self.get("ubs_db", "user"),
            password=self.get("ubs_db", "password"),
            decrypt=to_bool(self.get("ubs_db", "pw_encrypted")),
            aws_region=self.get("AWS", "aws_region"),
        )

        # create connection string to be used for cdc ubap services with separate user from ubs
        self.ubap_writer_conn_str = build_postgresql_connection_string(
            host=ubap_writer_host,
            port=self.get("ubap_db", "port"),
            db=self.get("ubap_db", "db"),
            username=self.get("ubap_db", "user"),
            password=self.get("ubap_db", "password"),
            decrypt=to_bool(self.get("ubap_db", "pw_encrypted")),
            aws_region=self.aws_region,
        )

        self.controller_queue_url = self.get("ubs_analysis", "state_update_queue_url")
        self.statement_timeout_ms = int(self.get("ubs_db", "statement_timeout_ms"))
        self.logging_stream = self.get("ubs_analysis", "logging_stream")


class UbsAnalysisFrontendConfig(BaseConfig):
    def __init__(self, port):
        super().__init__("ubs_analysis", port, "ubs.analysis.frontend")
        self.ssl_connection_args = None
        self.ubs_writer_conn_str = None
        self.ubs_reader_conn_str = None
        self.controller_queue_url = None

        reader_host = self.get("RDS_READER_HOST")
        # TODO: change these .get env vars to be ubap specific when ready for mcs2 implemenatation
        ubap_reader_host = self.get("RDS_READER_HOST")
        if self.commit_sha is not None and len(self.commit_sha) > 0:
            reader_host = reader_host.format(self.commit_sha)
            ubap_reader_host = ubap_reader_host.format(self.commit_sha)

        db = self.get("RDS_DB")
        port = int(self.get("RDS_PORT"))
        user = self.get("RDS_USER")
        password = self.get("RDS_PASSWORD")

        # ubap config
        ubap_writer_host = self.get("RDS_WRITER_HOST")
        ubap_port = self.get("RDS_PORT")
        ubap_db = self.get("RDS_DB")
        ubap_username = self.get("RDS_USER")
        ubap_password = self.get("RDS_PASSWORD")

        self.ubs_reader_conn_str = build_postgresql_connection_string(
            host=self.get("RDS_READER_HOST"),
            port=port,
            db=db,
            username=user,
            password=password,
            decrypt=False,
        )

        self.ubap_reader_conn_str = build_postgresql_connection_string(
            host=ubap_reader_host,
            port=ubap_port,
            db=ubap_db,
            username=ubap_username,
            password=ubap_password,
            decrypt=False,
        )

        # TODO: The writer isn't actually needed by this frontend
        self.ubs_writer_conn_str = build_postgresql_connection_string(
            host=self.get("RDS_WRITER_HOST"),
            port=port,
            db=db,
            username=user,
            password=password,
            decrypt=False,
        )

        self.ubap_writer_conn_str = build_postgresql_connection_string(
            host=ubap_writer_host,
            port=ubap_port,
            db=ubap_db,
            username=ubap_username,
            password=ubap_password,
            decrypt=False,
        )

        self.ssl_connection_args = {
            "sslmode": self.get("RDS_SSLMODE", default="verify-ca"),
            "application_name": "ubs_analysis_frontend",
        }
        sslrootcert = self.get("RDS_SSLROOTCERT")
        self.statement_timeout_ms = int(self.get("STATEMENT_TIMEOUT_MS", 100))
        if sslrootcert is not None:
            self.ssl_connection_args["sslrootcert"] = sslrootcert

        self.controller_queue_url = self.get("ANALYSIS_CONTROLLER_QUEUE_NAME")
        # self.statement_timeout_ms = int(self.get("ubs_db", "statement_timeout_ms"))
        self.logging_stream = self.get("LOGGING_STREAM_NAME")

        self.loaded = True
