import os
from io import BytesIO
import logging
import magic
from os import environ
from zipfile import ZipFile

import boto3

from ...shared.aws.kinesis_client import KinesisClient
from ...shared.aws.kms_client import KmsClient
from ...shared.aws.sqs_client import SqsClient
from ...shared.aws.s3_client import S3Client
from ...shared.ubs.dtos.analysis.bsv_timing import BsvTiming
from ...shared.ubs.dtos.analysis.state_update import UbsAnalysisStateUpdate
from ...shared.ubs.dtos.analysis.timing_collection import UbsAnalysisTimingCollection
from ...shared.ubs.dtos.analysis.stage_trigger_event import UbsAnalysisStageTriggerEvent
from ...shared.ubs.dtos.apc_upload_metadata import ApcUploadMetadata
from ...shared.ubs.dtos.response_psc_metadata import ResponsePscMetadata
from ...shared.ubs.enums.analysis.stage import UbsAnalysisStage
from ...shared.ubs.enums.analysis.status import UbsAnalysisStatus
from ...shared.ubs.repository.ubs_db import UbsDbRepository
from ...shared.ubs.models.models import FileIndex, FileDetail
from ...shared.ubs.util import generate_s3_key, hash_to_binary
from ...shared.utils.date_util import utc_now
from ...shared.utils.db_util import build_postgresql_connection_string
from ...shared.utils.hash_util import calculate_sha256, calculate_md5
from ...shared.utils.magic_util import is_invalid_type, executable_types, get_file_magic
from ...shared.utils.parsers.hash import is_valid_sha256
from ...shared.statsd_client import StatsdClient


class BinaryStorageValidation(object):
    def __init__(self):
        self.logger = None
        self.aws_region = None
        self.permanent_storage_bucket = None
        self.investigation_bucket = None
        self.kms_client = None
        self.s3_client = None
        self.sqs_client = None
        self.timing_kinesis_client = None
        self.invalid_file_log_enabled = None
        self.repository = None
        self.statsd = None
        self.statsd_prefix = "tic.pscr.ubs"
        self.state_update_queue_url = None
        self.stage_version = None
        self.valid_magic_prefixes = executable_types

        self.path_map = {
            "defense/apc_upload": {"dto": ApcUploadMetadata},
            "pscr/submit": {"dto": ResponsePscMetadata},
            "tau/submit": {"dto": ResponsePscMetadata},
        }

        self.exception_string = "Failed to validate or copy to permanent storage"
        self._set_up_clients()

    def _set_up_clients(self):
        self.aws_region = environ["AWS_DEFAULT_REGION"]
        self.stage_version = environ["version"]
        self.permanent_storage_bucket = environ["permanent_storage_bucket"]
        self.investigation_bucket = environ["investigation_bucket"]
        self.kms_key_alias = environ["kms_key_alias"]
        self.invalid_file_log_enabled = environ["invalid_file_log_enabled"]

        self.logger = logging.getLogger()
        self.logger.setLevel(logging.INFO)

        boto_session = boto3.Session(region_name=self.aws_region)
        self.kms_client = KmsClient(boto_session)

        self.fips_s3_endpoint = (
            "https://" + environ["AWS_ENDPOINT_OVERRIDE_S3"]
            if "AWS_ENDPOINT_OVERRIDE_S3" in os.environ
            else None
        )
        self.logger.info(f"fips s3 endpoint: {self.fips_s3_endpoint}")
        self.s3_client = S3Client(boto_session, endpoint_url=self.fips_s3_endpoint)
        self.sqs_client = SqsClient(boto_session)
        self.timing_kinesis_client = KinesisClient(
            boto_session, environ["timing_persistence_kinesis_stream"]
        )

        self.magic = magic.Magic(magic_file="/opt/share/magic")
        self.is_govcloud = (
            True
            if "is_govcloud" in os.environ and environ["is_govcloud"] == "true"
            else False
        )

        default_statsd_port = (
            environ["default_statsd_port"]
            if "default_statsd_port" in os.environ
            else None
        )
        # State Update Queue does not exist in gov cloud, remove gov cloud check if queue is added in the future
        if not self.is_govcloud:
            self.state_update_queue_url = environ["state_update_queue_url"]
        self.statsd = StatsdClient(
            prefix=self.statsd_prefix,
            is_govcloud=self.is_govcloud,
            port=default_statsd_port,
            logger=self.logger,
        )

        # TODO Figure out if we can have a session to the database maintained here.
        self.repository = UbsDbRepository(
            self.logger,
            build_postgresql_connection_string(
                environ["ubs_db_writer_host"],
                environ["ubs_db_port"],
                environ["ubs_db_db_name"],
                environ["ubs_db_user"],
                environ["ubs_db_pw"],
                environ["ubs_db_pw_encrypted"],
                self.aws_region,
            ),
            build_postgresql_connection_string(
                environ["ubs_db_reader_host"],
                environ["ubs_db_port"],
                environ["ubs_db_db_name"],
                environ["ubs_db_user"],
                environ["ubs_db_pw"],
                environ["ubs_db_pw_encrypted"],
                self.aws_region,
            ),
            connection_args={
                "sslmode": "verify-ca",
                "sslrootcert": "/tmp/root.crt",
                "application_name": "binary_storage_validation",
            },
        )

    @staticmethod
    def determine_download_file(known_size, new_object_size, truncated):
        """
        @param known_size: size of file that we currently have in binary store
        @param new_object_size: size of new file in ingestion store
        @param truncated: if true, file in binary store is truncated
        @return:
        """
        # We know nothing about this binary - ingest it.
        if known_size is None or truncated is None:
            return True
        elif truncated is True and known_size < new_object_size:
            return True
        else:
            return False

    def main(self, s3_bucket, s3_key):
        timing_collection = UbsAnalysisTimingCollection()
        timing_collection.bsv = BsvTiming()
        timing_collection.date_started = utc_now()
        try:
            fragments = s3_key.split("/")
            if "pscr" in set(x.lower() for x in fragments):
                org_key = fragments[-2]
                path = "/".join(
                    fragments[:-2]
                )  # because org key is provided in the path
                self.statsd.incr("bsv.upload.th")
            elif "tau" == fragments[0]:
                org_key = "tau"
                path = "/".join(fragments[:-1])  # org key doesn't seem to be provided
                self.statsd.incr("bsv.upload.tau")
            else:
                if len(fragments) == 4:  # including org key
                    org_key = fragments[-2]
                    path = "/".join(
                        fragments[:-2]
                    )  # because org key is provided in the path
                else:  # excluding org key
                    org_key = None
                    path = "/".join(fragments[:-1])
                self.statsd.incr("bsv.upload.apc")

            sha256 = fragments[-1].replace(".zip", "")
            timing_collection.sha256 = sha256
            target_s3_key = generate_s3_key(sha256)
            trigger_event = UbsAnalysisStageTriggerEvent(
                sha256=sha256, stage=UbsAnalysisStage.BINARY_STORAGE_VALIDATION
            )

            self._update_state(
                trigger_event, UbsAnalysisStatus.IN_PROGRESS, self.is_govcloud
            )

            # If it isn't a valid sha256 - we don't know what we have - this is a bug - don't query the database.
            if not is_valid_sha256(sha256):
                raise InvalidSha256()

            if path not in self.path_map:
                self.logger.warning(
                    "No path defined for uploads to {0}. Please implement.".format(path)
                )
                self._update_state(
                    trigger_event, UbsAnalysisStatus.DECLINE_EXECUTE, self.is_govcloud
                )
                return

            with self.repository.ubs_reader_session() as reader:
                (
                    known_size,
                    truncated,
                    _,
                    existing_magic,
                ) = self.repository.get_file_details_and_magic(sha256, reader)

            # download the file from the bucket
            with BytesIO() as f:
                # TODO: do this in streaming rather than full file in case of huge file.
                # Download compressed zip file from S3
                self.s3_client.download_fileobj(
                    file_obj=f, bucket_name=s3_bucket, key_name=s3_key
                )
                try:
                    incoming_file, metadata_json = self.extract_zip(f)
                except InvalidZipError:
                    self.handle_invalid_zip_error(
                        f, target_s3_key, org_key, sha256, s3_bucket, trigger_event
                    )
                    return
            timing_collection.bsv.pulled_file_and_existing_data = utc_now()

            sha256_truncated = calculate_sha256(incoming_file)

            # determine the size of the binary
            incoming_file_size = len(incoming_file)

            # load DTOs to pass to S3 insert and DB insert
            dto_class = self.path_map[path]["dto"]
            metadata_dto = dto_class.from_json(metadata_json)

            metadata_dto.org_key = org_key
            metadata_dto.sha256 = sha256

            trigger_event.s3_key = target_s3_key
            trigger_event.sha256_truncated = sha256_truncated
            trigger_event.magic = existing_magic or get_file_magic(
                incoming_file, magic=self.magic
            )

            # TAU can upload files of an invalid binary type.
            if org_key != "tau" and is_invalid_type(
                trigger_event.magic, self.valid_magic_prefixes
            ):
                valid_magic = False
                log_data = metadata_dto.get_log_data()
                log_data["magic"] = trigger_event.magic
                log_data["computed_sha256"] = sha256_truncated
                log_data["file_size"] = incoming_file_size
                log_data["s3_key"] = s3_key
                log_data["bucket"] = s3_bucket
                log_data["region"] = self.aws_region
                if path == "defense/apc_upload":
                    self.logger.info(
                        f"File provided for APC: {sha256=}, {metadata_dto.org_key=}, {metadata_dto.device_id=}, {valid_magic=}"
                    )
                self.handle_invalid_file_type(log_data)

            valid_magic = True
            if path == "defense/apc_upload":
                self.logger.info(
                    f"File provided for APC: {sha256=}, {metadata_dto.org_key=}, {metadata_dto.device_id=}, {valid_magic=}"
                )

            # determine whether we want the file or not
            file_is_desired = self.determine_download_file(
                known_size, incoming_file_size, truncated
            )

            # create bool to indicate whether to mark stage complete after sending timing.
            mark_complete = False
            if file_is_desired:
                self.handle_desired_file(incoming_file, trigger_event, metadata_dto)
                mark_complete = True
                timing_collection.bsv.stored_file = utc_now()

            if path == "defense/apc_upload":
                self._handle_apc_link_preexisting_file_to_customer(metadata_dto)
                mark_complete = True
                self._update_state(
                    trigger_event,
                    UbsAnalysisStatus.COMPLETE,
                    self.is_govcloud,
                    UbsAnalysisStage.APC_UPLOAD_ALLOWED,
                )
            self._update_state(
                trigger_event,
                UbsAnalysisStatus.COMPLETE
                if mark_complete
                else UbsAnalysisStatus.DECLINE_EXECUTE,
                self.is_govcloud,
            )
            self._update_state(
                trigger_event,
                UbsAnalysisStatus.COMPLETE,
                self.is_govcloud,
                pseudostage=UbsAnalysisStage.FILE_EXISTS,
            )

        except InvalidFileTypeError as e:
            self.statsd.incr("bsv.failure.non_pe_file")
            self.logger.exception(
                "InvalidFileTypeError: Upload of {0} denied due to magic filter of type {1}".format(
                    e.sha256_truncated, trigger_event.magic
                )
            )
            self._update_state(trigger_event, UbsAnalysisStatus.ERROR, self.is_govcloud)
        except InvalidMetadataError:
            self.logger.exception(
                "{} InvalidMetadataError {} in {}".format(
                    self.exception_string, s3_key, s3_bucket
                )
            )
            # TODO what happens here if a new file is uploaded?
            self._update_state(trigger_event, UbsAnalysisStatus.ERROR, self.is_govcloud)
        except InvalidSha256:
            self.logger.exception(
                "{} InvalidSha256 {} in {}".format(
                    self.exception_string, s3_key, s3_bucket
                )
            )
            self._update_state(trigger_event, UbsAnalysisStatus.ERROR, self.is_govcloud)
        except CorruptedFileError:
            self.statsd.incr("bsv.failure.corrupted_file")
            self.logger.exception(
                "{} CorruptedFileError {} in {}".format(
                    self.exception_string, s3_key, s3_bucket
                )
            )
            # TODO: what happens here if a new file is uploaded?
            self._update_state(trigger_event, UbsAnalysisStatus.ERROR, self.is_govcloud)
        except Exception as e:
            self.statsd.incr("bsv.failure.uncaught")
            self.logger.exception(
                "Failed to validate or copy to permanent storage {0} in {1}".format(
                    s3_key, s3_bucket
                )
            )
            # TODO: this should be sql or connection issues- retryable. Confirm.
            self._update_state(
                trigger_event, UbsAnalysisStatus.ERROR_RETRY, self.is_govcloud
            )
            raise e
        finally:
            if timing_collection.sha256 is not None:
                timing_collection.date_completed = utc_now()
                self.timing_kinesis_client.put_records(
                    {None: [timing_collection.to_json(jsonl=True)]}
                )

    def handle_invalid_file_type(
        self,
        log_data,
    ):
        try:
            self.s3_client.delete_files(log_data["bucket"], [log_data["s3_key"]])
        except Exception as e:
            self.logger.exception(
                f'Unable to delete file from submission bucket: {log_data["computed_sha256"]}'
            )
            raise e  # re-raise - we don't want to ignore this!

        raise InvalidFileTypeError(log_data["computed_sha256"])

    def handle_corrupted_file_error(self, incoming_file, metadata_dto):
        corrupted_file_path = f"corrupted_file/{metadata_dto.sensor_version}/{metadata_dto.org_key}/{metadata_dto.sha256}.zip"
        # Create zipfile in memory and upload.
        zipped_file = BytesIO()
        with ZipFile(zipped_file, "a", False) as zip_file:
            for file_name, data in [
                ("filedata", BytesIO(incoming_file)),
                ("metadata", BytesIO(bytes(metadata_dto.to_json(), "utf-8"))),
            ]:
                zip_file.writestr(file_name, data.getvalue())

        zipped_file.seek(0)
        self.s3_client.upload_fileobj(
            zipped_file, self.investigation_bucket, corrupted_file_path
        )

    def handle_invalid_zip_error(
        self, f, target_s3_key, org_key, sha256, s3_bucket, trigger_event
    ):
        zip_sha256 = calculate_sha256(f.getvalue())
        appended_s3_key = target_s3_key + "." + zip_sha256 + ".zip"

        # Note: We used to reduce S3 PUTs if object pre-existed. This was
        # a premature optimization and the code path wasn't used. It's been removed.
        self.kms_client.encrypt_data_with_boto(
            f.getvalue(),
            self.investigation_bucket,
            "invalid_zip_format/{}/{}".format(org_key, appended_s3_key),
            self.kms_key_alias,
        )
        self.statsd.incr("bsv.failure.invalid_zip_format")
        self._update_state(trigger_event, UbsAnalysisStatus.ERROR, self.is_govcloud)
        self.logger.exception(
            "{} InvalidZipError Original sha256 {} Zip Sha256 {} in {}".format(
                self.exception_string, sha256, zip_sha256, s3_bucket
            )
        )

    def handle_desired_file(self, incoming_file, trigger_event, metadata_dto):
        """
        Call functions that handle insert / update
        """

        self.validate_and_persist(trigger_event, incoming_file, metadata_dto)
        self.update_db(incoming_file, trigger_event, metadata_dto)

    def validate_and_persist(self, trigger_event, incoming_file, metadata_dto):
        """
        Upload file to permanent storage in S3
        """
        if not metadata_dto.is_valid():
            raise InvalidMetadataError()

        if metadata_dto.is_corrupted(trigger_event.sha256_truncated):
            if isinstance(metadata_dto, ResponsePscMetadata):
                self.handle_corrupted_file_error(incoming_file, metadata_dto)
            raise CorruptedFileError()

        file_truncated = trigger_event.sha256_truncated != metadata_dto.sha256

        if file_truncated and (
            metadata_dto.original_size is None or metadata_dto.original_md5 is None
        ):
            raise InvalidMetadataError(
                "No information provided about execution on disk for truncated file"
            )

        # Create zipfile in memory and upload.
        zipped_file = BytesIO()
        with ZipFile(zipped_file, "a", False) as zip_file:
            zip_file.writestr("filedata", incoming_file)
        zipped_file.seek(0)

        self.s3_client.upload_fileobj(
            zipped_file, self.permanent_storage_bucket, trigger_event.s3_key
        )

    def update_db(self, incoming_file, trigger_event, metadata_dto):
        """
        Update the DB with the new file & relevant relationships
        """
        trigger_event.is_test_file = metadata_dto.is_test()

        # retrieve values to insert into db
        file_truncated = trigger_event.sha256_truncated != metadata_dto.sha256
        truncated_md5 = calculate_md5(incoming_file)
        truncated_size = len(incoming_file)

        if not file_truncated:
            metadata_dto.original_md5 = truncated_md5
            metadata_dto.original_size = truncated_size

        cust = metadata_dto.get_cust()

        file_index = FileIndex(
            sha256=hash_to_binary(metadata_dto.sha256),
            md5=hash_to_binary(metadata_dto.original_md5),
            size=metadata_dto.original_size,
        )

        file_details = FileDetail(
            sha256=hash_to_binary(trigger_event.sha256_truncated),
            md5=hash_to_binary(truncated_md5),
            s3_key=trigger_event.s3_key,
            size=truncated_size,
            truncated=file_truncated,
            test=trigger_event.is_test_file,
        )

        with self.repository.ubs_writer_session() as writer:
            self.repository.put_upload_metadata(
                cust, file_index, file_details, trigger_event.magic, writer
            )

    def _handle_apc_link_preexisting_file_to_customer(self, metadata_dto):
        cust = metadata_dto.get_cust()

        with self.repository.ubs_writer_session() as writer:
            self.repository.insert_update_cust_and_link(
                session=writer,
                file_index=self.repository.get_file_index(
                    writer, hash_to_binary(metadata_dto.sha256)
                ),
                cust=cust,
            )

    @staticmethod
    def extract_zip(zip_file):
        with ZipFile(zip_file) as zf:
            if len(zf.namelist()) != 2:
                raise InvalidZipError()

            elif "metadata" in zf.namelist()[0]:
                return zf.read(zf.namelist()[1]), zf.read(zf.namelist()[0])

            return zf.read(zf.namelist()[0]), zf.read(zf.namelist()[1])

    def _update_state(
        self,
        trigger_event,
        status: UbsAnalysisStatus,
        is_govcloud,
        pseudostage: UbsAnalysisStage = None,
    ):
        """
        Update state by sending a message to the UBS Analysis Controller
        Only send messages on the commercial side (no op on govcloud side)
        @param trigger_event: Contains sha256, current stage information
        @param status: The status to update the stage to
        @param is_govcloud: Boolean of if this is running in govcloud currently
        @param pseudostage: The pseudostage to update- else updates current running stage
        """
        if is_govcloud:
            return

        # Only send to state queue when running on commercial side
        self.sqs_client.send_message(
            self.state_update_queue_url,
            UbsAnalysisStateUpdate(
                sha256=trigger_event.sha256,
                name=pseudostage or trigger_event.stage,
                version=self.stage_version,
                status=status,
            ).to_json(),
        )


# Handling these guys appropriately - log the exception and return.
class InvalidZipError(Exception):
    pass


class InvalidMetadataError(Exception):
    pass


class CorruptedFileError(Exception):
    pass


class InvalidSha256(Exception):
    pass


class InvalidFileTypeError(Exception):
    def __init__(self, sha256_truncated):
        super().__init__()
        self.sha256_truncated = sha256_truncated
