import json
from datetime import datetime, timedelta, timezone

import dateutil.parser
from flask import request
from werkzeug.exceptions import BadRequest

from tic.shared.service import TICService, TicPrincipal, read_ubs_data_permission
from tic.shared.utils.date_util import to_iso_8601
from tic.ubs.frontend.ubs_frontend_config import UbsFrontendConfig
from tic.ubs.frontend.ubs_frontend_configuration import UbsFrontendConfiguration
from tic.ubs.frontend.ubs_frontend_principal import UbsFrontendPrincipal
from tic.ubs.frontend.ubs_unique_hashes import get_unique_binaries_for_date
from .dtos.device_results import DeviceResults
from .dtos.download_request_body import DownloadRequestBody
from .dtos.download_results import DownloadResults, FoundResult
from .dtos.existence_request_body import ExistenceRequestBody
from .dtos.existence_results import ExistingResult
from .dtos.filepath_results import FilepathResults, Filepath
from .dtos.metadata_results import MetadataResults, BulkMetadataResult
from .dtos.signature_results import SignatureResults, Signature, SignState
from .exceptions import ClientError
from ...shared.aws.s3_client import S3Client
from ...shared.dtos.psc_api_response import (
    LegacyPscApiResponse,
    PscApiErrorCodes,
    PscApiErrorResponse,
)
from ...shared.ubs.models.models import Cust
from ...shared.ubs.repository.ubs_db import UbsDbRepository, TableCustDeviceLink
from ...shared.ubs.util import (
    convert_architectures_to_array,
    convert_sign_states_to_array,
    binary_to_hash,
    get_batch_cust_link_and_status_code,
)
from ...shared.utils.date_util import utc_now
from ...shared.utils.magic_util import is_invalid_type, executable_types
from ...shared.utils.parsers.hash import is_valid_sha256


class UbsFrontendService(TICService):
    def __init__(self, config, principal):
        super().__init__(config, principal)
        self.app.config["JSON_AS_ASCII"] = False  # Enable Unicode support

        if config == "swagger":
            self._set_up_routes()
            return

        self.ubs_repository = None
        self.ubs_repository_cache = None
        self.s3_client = None

        self._set_up_clients()
        self._set_up_routes()

        self.valid_magic_prefixes = executable_types

        self.MAX_REPORT_DELTA = timedelta(days=8)

        @self.app.after_request
        def always_return_header_content_type_application_json(response):
            response.headers["Content-Type"] = "application/json"
            return response

    def _set_up_clients(self):
        self.s3_client = S3Client(
            self.boto_session, endpoint_url=self.config.s3_endpoint_url
        )
        self.ubs_repository = UbsDbRepository(
            self.logger,
            self.config.ubs_writer_conn_str,
            self.config.ubs_reader_conn_str,
            connection_args=self.config.ssl_connection_args,
            statement_timeout=self.config.statement_timeout_ms,
        )

    def _set_up_routes(self):
        # If you add a new customer endpoint - be sure to add it to the psc.ubs project allow list
        # https://gitlab.bit9.local/cb-defense/psc.ubs/
        # These endpoints are for both customer and internal usage
        self.app.add_url_rule(
            f"{self.config.route_prefix}/v1/orgs/<orgKey>/file/_download",
            "download",
            self.get_download_links,
            methods=["POST"],
        )

        self.app.add_url_rule(
            f"{self.config.route_prefix}/v1/orgs/<orgKey>/sha256/<sha256>/metadata",
            "metadata",
            self.get_metadata,
            methods=["GET"],
        )

        self.app.add_url_rule(
            f"{self.config.route_prefix}/v1/orgs/<orgKey>/sha256/<sha256>/summary/file_path",
            "filepath",
            self.get_filepath,
            methods=["GET"],
        )

        self.app.add_url_rule(
            f"{self.config.route_prefix}/v1/orgs/<orgKey>/sha256/<sha256>/summary/device",
            "device",
            self.get_device,
            methods=["GET"],
        )
        self.app.add_url_rule(
            f"{self.config.route_prefix}/v1/orgs/<orgKey>/sha256/<sha256>/summary/signature",
            "signature",
            self.get_signature,
            methods=["GET"],
        )

        # These endpoints are for internal usage only.
        self.app.add_url_rule(
            f"{self.config.route_prefix}/v1/file/existence",
            "existence",
            self.get_existence,
            methods=["POST"],
        )
        self.app.add_url_rule(
            f"{self.config.route_prefix}/v1/metadata/_bulk",
            "bulk_metadata",
            self.get_metadata_bulk,
            methods=["POST"],
        )
        self.app.add_url_rule(
            f"{self.config.route_prefix}/v1/orgs/<orgKey>/bulk_file_access",
            "bulk_file_access",
            self.get_file_access_bulk,
            methods=["POST"],
        )
        self.app.add_url_rule(
            f"{self.config.route_prefix}/v1/orgs/<orgKey>/unique_hashes/<date>",
            "unique_hashes",
            self.get_recent_unique_hashes,
            methods=["GET"],
        )

    @read_ubs_data_permission.require(http_exception=403)
    def get_download_links(self, orgKey):
        """
        Get presigned download links
        ---
          consumes:
            - application/json
          produces:
            - application/json
          parameters:
            - in: path
              name: orgKey
              required: true
              type: string
            - in: body
              name: sha256
              required: true
              schema:
                type: object
                required:
                  - sha256
                properties:
                  sha256:
                    type: array
                    items:
                      type: string
            - in: body
              name: expiration_seconds
              required: false
              default: 3600
              type: integer
          responses:
            '200':
              description: success - all files found
              schema:
                type: object
                properties:
                  found:
                    description: List of presigned URLs to download files
                    type: array
                    items:
                      type: object
                      properties:
                        sha256:
                          type: string
                        url:
                          type: string
                  not_found:
                    description: List of sha256s not found
                    type: array
                    items:
                      type: string
                  error:
                    description: List of sha256s with errors
                    type: array
                    items:
                      type: string
            '404':
              description: files not found
        """

        status_code = 200
        response = PscApiErrorResponse()
        error_hashes = set()
        request_start_time = utc_now()

        try:
            download_request = DownloadRequestBody.from_dict(
                request.get_json(force=True)
            )
            try:
                download_request.validate_and_normalize_request()
            except Exception as e:
                raise ClientError(str(e))

            results = DownloadResults()
            org_key = orgKey

            if len(download_request.sha256) == 0:
                raise ClientError("No sha256s provided")

            if org_key is None:
                raise ClientError("No org key provided")

            self.statsd.incr(
                "file.download.files_requested", len(download_request.sha256)
            )

            # a set of the hashes that the customer has seen
            with self.ubs_repository.ubs_reader_session() as session:

                start_time = utc_now()

                rets = get_batch_cust_link_and_status_code(
                    repository=self.ubs_repository,
                    search_key=Cust.get_search_key(org_key=org_key),
                    sha256_strs=download_request.sha256,
                    reader=session,
                )

                self.statsd.timing(
                    "file.download.query_file_info",
                    (utc_now() - start_time).total_seconds() * 1000,
                )

                if rets[0].status_code == 404:
                    self.statsd.incr("file.download.success.missing_link")
                    status_code = 404
                    response.error_code = PscApiErrorCodes(status_code)
                    response.message = rets[0].message
                    return response.to_json(), status_code

                # Get the file details for the linked sha256s
                found_sha256s = [
                    link.sha256 for link in rets if link.status_code == 200
                ]

                if len(found_sha256s) == 0:
                    self.statsd.incr(
                        "file.download.success.no_files", len(found_sha256s)
                    )
                    results.not_found = download_request.sha256
                    return response.to_json(), status_code

                start_time = utc_now()

                file_details_and_magic = (
                    self.ubs_repository.get_file_details_and_magic_bulk(
                        sha256s=found_sha256s, reader=session
                    )
                )

                self.statsd.timing(
                    "file.download.query_file_details",
                    (utc_now() - start_time).total_seconds() * 1000,
                )

                for sha256 in download_request.sha256:
                    try:
                        if sha256 in file_details_and_magic:
                            file_details_obj = file_details_and_magic[
                                sha256
                            ].file_details
                            file_magic = file_details_and_magic[
                                sha256
                            ].file_details_magic_link
                            if is_invalid_type(
                                file_magic.magic.value, self.valid_magic_prefixes
                            ):
                                self.statsd.incr("file.download.failure.non_pe_file")
                                self.logger.exception(
                                    "Upload of {0} denied due to magic filter of type {1}".format(
                                        sha256, file_magic.magic.value
                                    )
                                )
                                results.not_found.append(sha256)
                            else:
                                results.found.append(
                                    FoundResult(
                                        sha256=sha256,
                                        url=self.s3_client.generate_presigned_download_url(
                                            self.config.ubs_storage_bucket,
                                            file_details_obj.s3_key,
                                            download_request.expiration_seconds,
                                        ),
                                    )
                                )
                        else:
                            results.not_found.append(sha256)
                    except Exception as e:
                        error_hashes.add(sha256)
                        self.logger.warning(
                            "Download Link Unknown Error Generating Presigned URL: {0}".format(
                                str(e)
                            )
                        )

                if len(error_hashes) > 0:
                    self.statsd.incr(
                        "file.download.success.failed_hashes", len(error_hashes)
                    )

            self.statsd.incr("file.download.success")

            self.statsd.timing(
                "file.download.time",
                (utc_now() - request_start_time).total_seconds() * 1000,
            )

            if status_code == 200:
                response = results

        except (BadRequest, ClientError) as e:
            self.logger.warning("Invalid UBS download request: {0}".format(str(e)))
            self.statsd.incr("file.download.failure.invalid_request")
            status_code = 400
            response.error_code = PscApiErrorCodes(status_code)
            response.message = str(e)

        except Exception:
            self.logger.exception("Uncaught exception during UBS download request")
            self.statsd.incr("file.download.failure.exception")
            status_code = 500
            response.error_code = PscApiErrorCodes(status_code)
            response.message = "Server Error, Please try again later"

        finally:
            return response.to_json(), status_code

    def get_existence(self):
        request_start = utc_now()
        status_code = 200
        response = LegacyPscApiResponse()

        try:
            existence_request = ExistenceRequestBody.from_dict(
                request.get_json(force=True)
            )
            existence_request.validate_and_normalize_sha256()

            response.results = []

            with self.ubs_repository.ubs_reader_session() as reader:
                file_details_bulk = self.ubs_repository.get_file_details_bulk(
                    existence_request.sha256, reader
                )

                for sha256 in existence_request.sha256:
                    if sha256 in file_details_bulk:
                        response.results.append(
                            ExistingResult(
                                sha256=sha256, file_size=file_details_bulk[sha256].size
                            )
                        )
                    else:
                        response.results.append(
                            ExistingResult(sha256=sha256, file_size=0)
                        )
                response.status_code = 200
                response.message = "success"
                self.statsd.incr("file.existence.success")

        except ClientError as e:
            self.logger.warning(
                "Invalid UBS file existence request: {0}".format(str(e))
            )
            self.statsd.incr("file.existence.failure.invalid_request")
            status_code = 400

            response = PscApiErrorResponse()
            response.message = str(e)
            response.error_code = PscApiErrorCodes(status_code)

        except Exception:
            self.logger.exception(
                "Failed to check UBS DB for existence of {0}".format(sha256)
            )
            self.statsd.incr("file.existence.failure.db_failed")
            status_code = 500

            response = PscApiErrorResponse()
            response.message = "Server Error, Please try again later"
            response.error_code = PscApiErrorCodes(status_code)

        finally:
            self.statsd.timing(
                "file.existence.latency",
                (utc_now() - request_start).total_seconds() * 1000,
            )
            self.statsd.incr("file.existence.requests")
            return response.to_json(), status_code

    @read_ubs_data_permission.require(http_exception=403)
    def get_metadata_bulk(self):
        """
        Internal API to return bulk metadata results
        """
        request_start = utc_now()
        status_code = 200
        response = []

        try:
            metadata_bulk_request = request.get_json(force=True)

            for sha256 in metadata_bulk_request:
                if not is_valid_sha256(sha256):
                    raise ClientError("{} is not a valid sha256".format(sha256))

            # Check that the org_key is valid
            with self.ubs_repository.ubs_reader_session() as reader:
                metadata_results = self.ubs_repository.get_metadata_bulk(
                    metadata_bulk_request, reader
                )

                for r in metadata_results:
                    result_md = BulkMetadataResult(r.sha256)

                    result_md.charset_id = r.charset_id
                    result_md.comments = r.comments
                    result_md.company_name = r.company_name.name
                    result_md.copyright = r.copyright.name
                    result_md.file_description = r.file_description
                    result_md.file_version = r.file_version
                    result_md.internal_name = r.internal_name.name
                    result_md.lang_id = r.lang_id
                    result_md.original_filename = r.original_filename
                    result_md.private_build = r.private_build
                    result_md.product_description = r.product_description
                    result_md.product_name = r.product_name.name
                    result_md.product_version = r.product_version
                    result_md.special_build = r.special_build
                    result_md.trademark = r.trademark.name

                    response.append(result_md.to_dict())

            self.statsd.incr("metadata.bulk.success")
        except ClientError as e:
            status_code = 400
            response = PscApiErrorResponse()
            response.message = str(e)
            response.error_code = PscApiErrorCodes(status_code)

            self.logger.warning("Invalid UBS Bulk metadata request: {0}".format(str(e)))
            self.statsd.incr("metadata.bulk.failure.invalid_request")
            response = response.to_dict(True, True)
        except Exception:
            status_code = 500
            response = PscApiErrorResponse()
            response.message = "Server Error, Please try again later"
            response.error_code = PscApiErrorCodes(status_code)

            self.logger.exception("Uncaught exception during UBS Bulk metadata request")
            self.statsd.incr("metadata.bulk.failure.exception")

            # We might have too many and this might fail
            self.logger.info(
                "sha256 associated with the exception {}".format(
                    metadata_bulk_request.sha256
                )
            )
            response = response.to_dict(True, True)
        finally:
            if self.config.statsd_metrics_enabled:
                self.statsd.timing(
                    "metadata.bulk.latency",
                    (utc_now() - request_start).total_seconds() * 1000,
                )
                self.statsd.incr("metadata.bulk.requests")
            return json.dumps(response, ensure_ascii=False), status_code

    @staticmethod
    def check_valid_sha256s(sha256s):
        if not sha256s:
            raise ClientError("No sha256s provided")

        for sha256 in sha256s:
            if not is_valid_sha256(sha256):
                raise ClientError("{} is not a valid sha256".format(sha256))

        return sha256s

    @staticmethod
    def handle_exception(error_code, message):
        response = PscApiErrorResponse()
        response.message = message
        response.error_code = PscApiErrorCodes(error_code)
        return response.to_dict(True, True), error_code

    @staticmethod
    def check_valid_org_key(org_key):
        if not org_key:
            raise ClientError("No org key provided")
        return org_key

    @read_ubs_data_permission.require(http_exception=403)
    def get_file_access_bulk(self, orgKey):
        """
        Internal API to return bulk file access results.  We want to see if a customer has access to the supplied
        sha256s.
        """
        request_start = utc_now()
        status_code = 200

        try:
            search_key = Cust.get_search_key(self.check_valid_org_key(orgKey))
            access_bulk_request = self.check_valid_sha256s(request.get_json(force=True))

            with self.ubs_repository.ubs_reader_session() as reader:
                accessible = self.ubs_repository.get_file_access_bulk(
                    search_key, access_bulk_request, reader
                )
                response = {
                    "approved": accessible,
                    "denied": [
                        sha256
                        for sha256 in access_bulk_request
                        if sha256 not in accessible
                    ],
                }

            self.statsd.incr("file_access.bulk.success")

        except ClientError as e:
            response, status_code = self.handle_exception(400, str(e))
            self.logger.warning(
                "Invalid UBS Bulk file access request: {0}".format(str(e))
            )
            self.statsd.incr("file_access.bulk.failure.invalid_request")

        except Exception:
            response, status_code = self.handle_exception(
                500, "Server Error, Please try again later"
            )
            self.logger.exception(
                "Unknown exception during UBS Bulk file access request"
            )
            self.statsd.incr("file_access.bulk.failure.exception")

        self.statsd.timing(
            "file_access.bulk.latency",
            (utc_now() - request_start).total_seconds() * 1000,
        )
        self.statsd.incr("file_access.bulk.requests")

        return json.dumps(response, ensure_ascii=False), status_code

    @read_ubs_data_permission.require(http_exception=403)
    def get_recent_unique_hashes(self, orgKey, date):
        try:
            report_date = dateutil.parser.parse(date).replace(tzinfo=timezone.utc)
            if datetime.now(
                tz=timezone.utc
            ) - self.MAX_REPORT_DELTA > report_date or report_date > datetime.now(
                tz=timezone.utc
            ) - timedelta(
                days=1
            ):
                response, status_code = self.handle_exception(
                    400,
                    "Date provided exceeds allowed range. min: {0}, max: {1}".format(
                        (
                            datetime.now(tz=timezone.utc) - self.MAX_REPORT_DELTA
                        ).isoformat(),
                        (datetime.now(tz=timezone.utc) - timedelta(days=1)).isoformat(),
                    ),
                )
            else:
                with self.ubs_repository.ubs_reader_session() as reader:
                    response = get_unique_binaries_for_date(report_date, orgKey, reader)
                    status_code = 200
        except TimeoutError:
            response, status_code = self.handle_exception(
                429,
                "Report failed to build in time, "
                "please retry after a few minutes to allow report to build",
            )
        return json.dumps(response, ensure_ascii=False), status_code

    @read_ubs_data_permission.require(http_exception=403)
    def get_metadata(self, orgKey, sha256):
        """
        Get metadata
        ---
          produces:
            - application/json
          description: ''
          responses:
            '200':
              description: success - metadata found
            '404':
              description: metadata not found
          parameters:
            - in: path
              name: orgKey
              required: true
              type: string
            - in: path
              name: sha256
              required: true
              type: string
        """

        status_code = 200
        response = PscApiErrorResponse()

        try:
            if not is_valid_sha256(sha256):
                raise ClientError("{} is not a valid sha256".format(sha256))

            sha256 = sha256.lower()
            results = MetadataResults(sha256)

            org_key = orgKey

            if org_key is None:
                raise ClientError("No org key provided")

            # Check that the org_key is valid
            with self.ubs_repository.ubs_reader_session() as reader:
                cust_access = get_batch_cust_link_and_status_code(
                    repository=self.ubs_repository,
                    search_key=Cust.get_search_key(org_key=org_key),
                    sha256_strs=[sha256],
                    reader=reader,
                )[0]

                if cust_access.cust_ix_link is None:
                    self.statsd.incr("summary.metadata.success.missing_link")
                    status_code = 404
                    response.message = "{} not found".format(sha256)
                    response.error_code = PscApiErrorCodes(status_code)
                    return response.to_json(), status_code

                metadata = self.ubs_repository.get_metadata(sha256, reader)

                if metadata is None:
                    self.statsd.incr("summary.metadata.success.no_metadata")
                    status_code = 404
                    response.message = "{} not found".format(sha256)
                    response.error_code = PscApiErrorCodes(status_code)
                    return response.to_json(), status_code

                # Determine if the file is available and the file_size
                file_details = self.ubs_repository.get_file_details(sha256, reader)
                if file_details:
                    results.available_file_size = file_details.size
                    results.file_available = True

                else:
                    results.available_file_size = 0
                    results.file_available = False

                results.sha256 = sha256
                results.file_size = metadata.file_index.size
                if metadata.file_index.md5 is not None:
                    results.md5 = binary_to_hash(metadata.file_index.md5)

                # Update the response with the metadata
                results.os_type = metadata.os_type
                results.lang_id = metadata.lang_id
                results.charset_id = metadata.charset_id
                results.internal_name = metadata.internal_name.name
                results.product_name = metadata.product_name.name
                results.company_name = metadata.company_name.name
                results.copyright = metadata.copyright.name
                results.trademark = metadata.trademark.name
                results.file_description = metadata.file_description
                results.file_version = metadata.file_version
                results.comments = metadata.comments
                results.original_filename = metadata.original_filename
                results.product_description = metadata.product_description
                results.product_version = metadata.product_version
                results.private_build = metadata.private_build
                results.special_build = metadata.special_build

                # Iterate over the machine_architectures list as a bitmask to populate the architecture array
                if metadata.architecture is not None:
                    results.architecture = convert_architectures_to_array(
                        metadata.architecture
                    )

                self.statsd.incr("summary.metadata.success.completed")

            if status_code == 200:
                response = results

        except ClientError as e:
            self.logger.warning("Invalid UBS metadata request: {0}".format(str(e)))
            self.statsd.incr("summary.metadata.failure.invalid_request")
            status_code = 400
            response.message = str(e)
            response.error_code = PscApiErrorCodes(status_code)

        except Exception:
            self.logger.exception(
                "Uncaught exception during UBS metadata request for {0}".format(sha256)
            )
            self.statsd.incr("summary.metadata.failure.exception")
            status_code = 500
            response.message = "Server Error, Please try again later"
            response.error_code = PscApiErrorCodes(status_code)

        finally:
            return response.to_json(), status_code

    @read_ubs_data_permission.require(http_exception=403)
    def get_filepath(self, orgKey, sha256):
        status_code = 200

        response = PscApiErrorResponse()
        try:
            org_key = orgKey
            rows = request.args.get("rows", "5")

            if not is_valid_sha256(sha256):
                raise ClientError("{} is not a valid sha256".format(sha256))

            sha256 = sha256.lower()
            results = FilepathResults(sha256)

            if org_key is None:
                raise ClientError("No org key provided")

            # Verify that the rows is within the correct range
            if rows is None:
                rows = 5
            try:
                rows = int(rows)
            except Exception:
                raise ClientError("rows must be an integer")
            if rows < 1 or rows > 100:
                raise ClientError("rows out of range")

            # Check that the org_key is valid
            with self.ubs_repository.ubs_reader_session() as reader:
                cust_access = get_batch_cust_link_and_status_code(
                    repository=self.ubs_repository,
                    search_key=Cust.get_search_key(org_key=org_key),
                    sha256_strs=[sha256],
                    reader=reader,
                )[0]

                status_code = cust_access.status_code
                response.message = cust_access.message
                results.sha256 = sha256
                if cust_access.cust_ix_link is None:

                    self.statsd.incr("summary.filepath.success.missing_link")
                    status_code = 404
                    response.error_code = PscApiErrorCodes(status_code)
                    return response.to_json(), status_code

                # Retrieve the filepath information for the file
                cust_file_paths = self.ubs_repository.get_file_path_array(
                    cust_access.cust_ix_link, rows, reader
                )

                if len(cust_file_paths) == 0:
                    self.statsd.incr("summary.filepath.success.no_filepaths")
                    status_code = 404
                    response.error_code = PscApiErrorCodes(status_code)
                    response.message = "Filepath Data not yet available for {}, please try again later".format(
                        sha256
                    )
                    return response.to_json(), status_code

                results.total_file_path_count = (
                    self.ubs_repository.get_total_file_path_count(
                        cust_access.cust_ix_link, reader
                    )
                )
                results.file_path_count = self.ubs_repository.get_file_path_count(
                    cust_access.cust_ix_link, reader
                )

                file_paths = []
                for cust_file_path in cust_file_paths:
                    file_path = Filepath()
                    file_path.count = cust_file_path.num_occurrences
                    file_path.file_path = cust_file_path.filepath.path

                    if isinstance(cust_file_path.first_seen, datetime):
                        file_path.first_seen_timestamp = to_iso_8601(
                            cust_file_path.first_seen
                        )
                    file_paths.append(file_path)

                results.file_paths = file_paths

            self.statsd.incr("summary.filepath.success.completed")

            if status_code == 200:
                response = results

        except ClientError as e:
            self.logger.warning("Invalid UBS file-path request: {0}".format(str(e)))
            self.statsd.incr("summary.filepath.failure.invalid_request")
            status_code = 400
            response.message = str(e)
            response.error_code = PscApiErrorCodes(status_code)
        except Exception:
            self.logger.exception(
                "Uncaught exception during UBS file-path request for {0}".format(sha256)
            )
            self.statsd.incr("summary.filepath.failure.exception")
            status_code = 500
            response.message = "Server Error, Please try again later"
            response.error_code = PscApiErrorCodes(status_code)
        finally:
            return response.to_json(), status_code

    @read_ubs_data_permission.require(http_exception=403)
    def get_device(self, orgKey, sha256):
        response = PscApiErrorResponse()
        status_code = 200

        try:
            org_key = orgKey

            if not is_valid_sha256(sha256):
                raise ClientError("{} is not a valid sha256".format(sha256))

            sha256 = sha256.lower()
            results = DeviceResults(sha256)

            if org_key is None:
                raise ClientError("No org key provided")

            # Check that the org_key is valid
            with self.ubs_repository.ubs_reader_session() as reader:
                cust_access = get_batch_cust_link_and_status_code(
                    repository=self.ubs_repository,
                    search_key=Cust.get_search_key(org_key=org_key),
                    sha256_strs=[sha256],
                    reader=reader,
                )[0]

                results.sha256 = sha256

                if cust_access.cust_ix_link is None:
                    self.statsd.incr("summary.device.success.missing_link")
                    status_code = 404
                    response.error_code = PscApiErrorCodes(status_code)
                    response.message = cust_access.message
                    return response.to_json(), status_code

                # get current cust_device_link table
                curr_cdl, next_cdl = self.ubs_repository.get_cust_device_link_tables(
                    reader
                )
                cdl = TableCustDeviceLink(curr_cdl)

                # Retrieve the device info
                results.num_devices = self.ubs_repository.get_num_devices(
                    cust_access.cust_ix_link, reader, cdl
                )

                # Exit if no devices
                if not results.num_devices or results.num_devices == 0:
                    self.statsd.incr("summary.device.success.no_devices")
                    status_code = 404
                    response.error_code = PscApiErrorCodes(status_code)
                    response.message = "Device Data not yet available for {}, please try again later".format(
                        sha256
                    )
                    return response.to_json(), status_code

                # If only one device, first seen and last seen device are the same
                first = self.ubs_repository.get_first_seen_device(
                    cust_access.cust_ix_link, reader, cdl
                )
                if results.num_devices == 1:
                    last = first
                else:
                    last = self.ubs_repository.get_last_seen_device(
                        cust_access.cust_ix_link, reader, cdl
                    )

                results.first_seen_device_id = first.device.device_id
                results.first_seen_device_name = first.device.device_name
                if isinstance(first.first_seen, datetime):
                    results.first_seen_device_timestamp = to_iso_8601(first.first_seen)

                results.last_seen_device_id = last.device.device_id
                results.last_seen_device_name = last.device.device_name
                if isinstance(last.first_seen, datetime):
                    results.last_seen_device_timestamp = to_iso_8601(last.first_seen)

            self.statsd.incr("summary.device.success.completed")

            if status_code == 200:
                response = results

        except ClientError as e:
            self.logger.warning("Invalid UBS device request: {0}".format(str(e)))
            self.statsd.incr("summary.device.failure.invalid_request")
            status_code = 400
            response.message = str(e)
            response.error_code = PscApiErrorCodes(status_code)
        except Exception:
            self.logger.exception(
                "Uncaught exception during UBS device request for {0}".format(sha256)
            )
            self.statsd.incr("summary.device.failure.exception")
            status_code = 500
            response.message = "Server Error, Please try again later"
            response.error_code = PscApiErrorCodes(status_code)
        finally:
            return response.to_json(), status_code

    @read_ubs_data_permission.require(http_exception=403)
    def get_signature(self, orgKey, sha256):
        response = PscApiErrorResponse()
        status_code = 200

        try:

            org_key = orgKey
            rows = request.args.get("rows", "5")

            if not is_valid_sha256(sha256):
                raise ClientError("{} is not a valid sha256".format(sha256))

            sha256 = sha256.lower()
            results = SignatureResults(sha256)

            if org_key is None:
                raise ClientError("No org key provided")

            # Verify that the rows is within the correct range
            if rows is None:
                rows = 5
            try:
                rows = int(rows)
            except Exception:
                raise ClientError("rows must be an integer")
            if rows < 1 or rows > 100:
                raise ClientError("rows out of range")

            # Check that the org_key is valid
            with self.ubs_repository.ubs_reader_session() as reader:
                cust_access = get_batch_cust_link_and_status_code(
                    repository=self.ubs_repository,
                    search_key=Cust.get_search_key(org_key=org_key),
                    sha256_strs=[sha256],
                    reader=reader,
                )[0]

                status_code = cust_access.status_code
                response.message = cust_access.message
                results.sha256 = sha256

                if status_code == 404:
                    response.error_code = PscApiErrorCodes(status_code)

                if cust_access.cust_ix_link is None:
                    self.statsd.incr("summary.signature.success.missing_link")
                    status_code = 404
                    response.error_code = PscApiErrorCodes(status_code)
                    return response.to_json(), status_code

                # Retrieve the signature information for the file
                sig_result_links = self.ubs_repository.get_signature_array(
                    cust_access.cust_ix_link, rows, reader
                )

                # Exit if no signature information
                if len(sig_result_links) == 0:
                    self.statsd.incr("summary.signature.success.no_signatures")
                    status_code = 404
                    response.error_code = PscApiErrorCodes(status_code)
                    response.message = "Signature Data not yet available for {}, please try again later".format(
                        sha256
                    )
                    return response.to_json(), status_code

                results.total_signatures_count = (
                    self.ubs_repository.get_total_signature_count(
                        cust_access.cust_ix_link, reader
                    )
                )
                results.signatures_count = self.ubs_repository.get_signature_count(
                    cust_access.cust_ix_link, reader
                )

                # TODO Update to be O(n)
                signatures = []
                for sig_result_link in sig_result_links:
                    sig_results = Signature()

                    sig_results.is_catalog_signature = (
                        sig_result_link.is_catalog_signature
                    )
                    sig_results.publisher_name = (
                        sig_result_link.signature.digsig_publisher_name.name
                    )
                    sig_results.issuer_name = (
                        sig_result_link.signature.digsig_issuer_name.name
                    )

                    if isinstance(sig_result_link.signature.digsig_sign_time, datetime):
                        sig_results.sign_timestamp = to_iso_8601(
                            sig_result_link.signature.digsig_sign_time
                        )

                    # Count is the sum of num_occurences with the same cust_ix_link and signature
                    sig_count = 0

                    sign_states = []
                    for sig in sig_result_links:

                        # Add the sign state if the signature matches
                        if sig.signature_id == sig_result_link.signature_id:
                            sig_count += sig.num_occurrences
                            sign_state = SignState()

                            sign_state.device_count = sig.num_occurrences

                            # Iterate over the signature_state list as a bitmask to populate the sign_state array
                            if sig.sig_state is not None:
                                sign_state.sign_state = convert_sign_states_to_array(
                                    sig.sig_state
                                )

                            if isinstance(sig.first_seen, datetime):
                                sign_state.first_seen_timestamp = to_iso_8601(
                                    sig.first_seen
                                )

                            sign_states.append(sign_state)

                    sig_results.sign_states = sign_states
                    sig_results.count = sig_count
                    signatures.append(sig_results)

                results.signatures = signatures

                self.statsd.incr("summary.signature.success.completed")

            if status_code == 200:
                response = results
        except ClientError as e:
            self.logger.warning("Invalid UBS signature request: {0}".format(str(e)))
            self.statsd.incr("summary.signature.failure.invalid_request")
            status_code = 400
            response.message = str(e)
            response.error_code = PscApiErrorCodes(status_code)

        except Exception:
            self.logger.exception(
                "Uncaught exception during UBS signature request for {0}".format(sha256)
            )
            self.statsd.incr("summary.signature.failure.exception")
            status_code = 500
            response.message = "Server Error, Please try again later"
            response.error_code = PscApiErrorCodes(status_code)

        finally:
            return response.to_json(), status_code


def start_service(port):
    config = UbsFrontendConfiguration(port)
    principal = TicPrincipal(config)
    service = UbsFrontendService(config, principal)
    return service.app


# Entry point for mosaic2
def start_service_mcs2(port):
    config = UbsFrontendConfig(port)
    principal = UbsFrontendPrincipal()
    service = UbsFrontendService(config, principal)
    return service.app
