"""
  UBS metadata ingestion dedupe lambda function -
  ingest UBS metadata from Kinesis metadata ingestion stream (randomly sharded),
  invoke metadata_ingestion_dedupe main function to dedupe
  Failed/errored records are put back to the Kinesis metadata ingest retry stream,
  failed tried records are put to MetadataDedupeLambdaDeadLetterQueue.
"""
# Native imports
from base64 import b64decode
import json
import logging
import os
import shutil

from aws_kinesis_agg.deaggregator import iter_deaggregate_records
import snappy

# Program imports
from tic.shared.utils.lambda_utils.kinesis_lambda_util import KinesisLambdaUtil
from tic.ubs.metadata.dedupe.metadata_ingestion_dedupe import MetadataIngestionDedupe
import tic.shared.ubs.repository

logger = logging.getLogger()
logger.setLevel(logging.INFO)

rds_cert = (
    "/rds-combined-ca-bundle.pem"
    if "is_govcloud" in os.environ
    else "/rds-ca-2019-root.pem"
)

root_crt = os.path.dirname(tic.shared.ubs.repository.__file__) + rds_cert
shutil.copyfile(root_crt, os.path.expanduser("/tmp/root.crt"))
logger.debug(f"rds_cert: {rds_cert}")

metadata_ingestion_dedupe = MetadataIngestionDedupe()


def handler(event, context):
    metadata_ingestion_dedupe.statsd.incr("kinesis.batch")
    metadata_ingestion_dedupe.statsd.incr("kinesis.record", len(event["Records"]))

    raw_records = [
        snappy.decompress(b64decode(record["kinesis"]["data"]))
        for record in iter_deaggregate_records(event["Records"])
        if record
    ]

    if len(raw_records) == 0:
        logger.warning("raw records length is 0, exit now")
        return

    metadata_ingestion_dedupe.statsd.incr("post.deaggregation.record", len(raw_records))

    kinesis_lambda_util = KinesisLambdaUtil(
        metadata_ingestion_dedupe.dlq_name,
        metadata_ingestion_dedupe.sqs_client,
        logger,
        metadata_ingestion_dedupe.statsd,
    )

    try:
        records = [json.loads(record) for record in raw_records]
    except Exception as ex:
        kinesis_lambda_util.exception_handler(ex, in_raw_records=raw_records)
        return

    try:
        metadata_ingestion_dedupe.main(records, kinesis_lambda_util)
    except Exception as ex:
        kinesis_lambda_util.exception_handler(ex, in_records=records)
