"""
  UBS metadata ingestion dedupe lambda function main functions -
  ingest UBS metadata from Kinesis metadata ingestion stream (randomly sharded),
  dedupe the records in the stream based on (sha256, org_key, device_id, device_name),
   put the deduped records to Kinesis metadata deduped stream, partitioned by "org_key-sha256".
  Exceptions are thrown upon failed puts/replayable records, decoding/non-replayable records are thrown to DLQ
"""

import logging
import os
from os import environ
import traceback

import boto3

from tic.shared.statsd_client import StatsdClient
from tic.ubs.metadata.shared.dtos.metadata import IngestionObject
from tic.shared.aws.kinesis_client import KinesisClient
from tic.shared.ubs.repository.ubs_db import UbsDbRepository, TableCustDeviceLink
from tic.shared.utils.db_util import build_postgresql_connection_string
from tic.shared.aws.s3_client import S3Client
from tic.shared.aws.sqs_client import SqsClient
from tic.shared.utils.exception_util import LambdaException
from tic.shared.utils.date_util import utc_now


class MetadataIngestionDedupe(object):
    def __init__(self):
        self.logger = None
        self.aws_region = None
        self.dlq_name = None
        self.dlq_msg_pack_size = None
        self.boto_session = None
        self.repository = None
        self.s3_client = None
        self.sqs_client = None
        self.result_kinesis_client = None
        self.curr_cust_device_link = None
        self.next_cust_device_link = None
        self.statsd_prefix = "tic.ubs.dedupe"

        self.metadata_internal_version = 1
        self._set_up_clients()

    def _set_up_clients(self):
        self.dlq_name = environ["dlq_name"]
        self.aws_region = environ["AWS_DEFAULT_REGION"]

        self.boto_session = boto3.Session(region_name=self.aws_region)
        self.logger = logging.getLogger()
        self.logger.setLevel(logging.INFO)
        self.result_kinesis_client = KinesisClient(
            self.boto_session, environ["deduped_events_stream"]
        )

        self.repository = self._get_ubs_repository()

        self.fips_s3_endpoint = (
            "https://" + environ["AWS_ENDPOINT_OVERRIDE_S3"]
            if "AWS_ENDPOINT_OVERRIDE_S3" in os.environ
            else None
        )
        self.logger.info(f"fips s3 endpoint: {self.fips_s3_endpoint}")
        self.s3_client = S3Client(
            boto_session=self.boto_session, endpoint_url=self.fips_s3_endpoint
        )
        self.sqs_client = SqsClient(self.boto_session)
        self.metadata_internal_version = int(environ["metadata_internal_version"])
        self.dlq_msg_pack_size = int(environ["dlq_msg_pack_size"])

        is_govcloud = True if "is_govcloud" in os.environ else False
        default_statsd_port = (
            environ["default_statsd_port"]
            if "default_statsd_port" in os.environ
            else None
        )
        self.statsd = StatsdClient(
            prefix=self.statsd_prefix,
            is_govcloud=is_govcloud,
            port=default_statsd_port,
            logger=self.logger,
        )

    def _get_ubs_repository(self):
        return UbsDbRepository(
            self.logger,
            None,
            build_postgresql_connection_string(
                environ["ubs_db_reader_host"],
                environ["ubs_db_port"],
                environ["ubs_db_db_name"],
                environ["ubs_db_user"],
                environ["ubs_db_pw"],
                environ["ubs_db_pw_encrypted"],
                self.aws_region,
            ),
            connection_args={
                "sslmode": "verify-ca",
                "sslrootcert": "/tmp/root.crt",
                "application_name": "metadata_dedupe",
            },
            statement_timeout=int(environ["statement_timeout_ms"]),
        )

    def _output_to_kinesis(self, ingestion_objects):
        # put the deduped results to the second/deduped Kinesis stream
        e = LambdaException()

        output = dict()  # partition_key : [raw_record]
        for ingestion_object in ingestion_objects:
            key = ingestion_object[0]
            partition_key = str(key[1]) + "-" + str(key[0])  # orgkey-sha256
            record = ingestion_object[1].to_json(minimize=True)
            if partition_key in output:
                output[partition_key].append(record)
            else:
                output[partition_key] = [record]

        # returns list of records {"Data": record value, "PartitionKey": pk}
        unsent_records = self.result_kinesis_client.put_records_aggregated(output)

        for item in unsent_records:
            e.add(
                "failed to put to metadata deduped stream",
                "_output_to_kinesis add unsent_records to LambdaException.data",
                None,
                item["Data"],
            )

        if len(unsent_records) > 0:
            self.logger.exception("failed to put to metadata deduped stream")
            self.statsd.incr("failed.puts.records", len(unsent_records))

        return e

    def _get_cust_device_link_tables(self):
        with self.repository.ubs_reader_session() as reader:
            (
                self.curr_cust_device_link,
                self.next_cust_device_link,
            ) = self.repository.get_cust_device_link_tables(reader)

    def main(self, records, err_handler):
        """
        Run this "service".
        Ingests metadata from Kinesis, dedupe, then put into dedupe Kinesis stream
        error/exceptions induced retries will be handled by the metadata dedupe Lambda handler

        @param err_handler: the error_handler to handle common error/exception by both lambdas
        @param records: the metadata provided for dedupe ingestion
        """
        self._get_cust_device_link_tables()

        (ingestion_objects, e_non_replayable,) = self._prep_records(
            records
        )  # dedupe

        # send mal-formatted records to DLQ
        if len(e_non_replayable.data) > 0:
            self.logger.exception("mal-formatted data caught when preparing records")
            err_handler.send_lambda_exception_to_dlq(e_non_replayable.data)
            self.statsd.incr("lambda.handler.exception.non_retryable.invocation", 1)
            self.statsd.incr(
                "lambda.handler.exception.non_retryable.records",
                len(e_non_replayable.data),
            )

        # put the deduped results to the second/deduped Kinesis stream
        e = self._output_to_kinesis(ingestion_objects)

        if len(e.data) > 0:
            raise e

    def _dedupe_device_links(self, tuple_mapping):
        cdl = TableCustDeviceLink(self.curr_cust_device_link)

        with self.repository.ubs_reader_session() as reader:
            start_time = utc_now()
            existing_cust_device_links = self.repository.get_duplicate_device_links(
                reader, tuple_mapping, cdl
            )
            self.statsd.incr(
                "duplicate.cust.device.link", len(existing_cust_device_links)
            )
            self.statsd.timing(
                "duplicate.cust.device.link.execution.time", utc_now() - start_time
            )
            ingestion_objects = [
                (
                    tuple_check,
                    ingestion_object,
                )
                for tuple_check, ingestion_object in tuple_mapping.items()
                if tuple_check not in existing_cust_device_links
            ]
        return ingestion_objects

    def _prep_records(self, in_records):
        device_cust_mapping = dict()
        e = LambdaException()
        self.logger.debug(f"total records: {len(in_records)}")
        for in_record in in_records:
            try:
                record = IngestionObject.from_dict(in_record, True, True)
                if record.event.process_op is not None:
                    main_module = record.event.process_op.target.main_module
                elif record.event.modload_op is not None:
                    # target is the file info in modload, for some reason.
                    main_module = record.event.modload_op.target
                else:
                    raise Exception("Record has no operation details")

                # map hash, org_key, device_id pair to associated record for later deduping
                # NOTE: we can get records with (sha256, org_key, device_id) dupes that will overwrite each other
                key = (
                    main_module.sha256,
                    record.device_meta.org_key,
                    record.device_meta.device_id,
                    record.device_meta.device_name,
                )
                device_cust_mapping[key] = record
            except Exception as ex:
                e.add(str(ex), traceback.format_exc(), in_record, None)

        # dedupe built dict based on key, return values
        ingestion_objects = self._dedupe_device_links(device_cust_mapping)

        return ingestion_objects, e
