import logging
import os
from contextlib import contextmanager
from enum import Enum
from io import BytesIO
from os import environ
import traceback
import json

import boto3

from tic.shared.aws.s3_client import S3Client, S3StorageClass
from tic.shared.aws.sqs_client import SqsClient
from tic.shared.statsd_client import StatsdClient
from tic.shared.ubs.models.file_path import FilePathInsert
from tic.shared.ubs.models.sp_models import IndexCustomerMetadata, SignatureData
from tic.shared.ubs.models.static_metadata_insert import StaticMetadataInsert
from tic.shared.ubs.models.device_insert import DeviceInsert
from tic.shared.ubs.repository.ubs_db import UbsDbRepository, TableCustDeviceLink
from tic.shared.utils.db_util import build_postgresql_connection_string
from tic.shared.utils.exception_util import LambdaException
from tic.shared.utils.date_util import utc_now, to_iso_8601
from tic.ubs.metadata.shared.dtos.metadata import IngestionObject

MAX_LANG_CHAR_ID = 2147483647


class DbSessionType(Enum):
    WRITER = 1
    READER = 2

    @staticmethod
    def get_session(repository, session_type):
        if session_type == DbSessionType.WRITER:
            return repository.ubs_writer_session()
        return repository.ubs_reader_session()


@contextmanager
def timed_block(statsd, label):
    start_time = utc_now()
    try:
        yield
        statsd.timing(f"{label}.time", utc_now() - start_time)
    except Exception:
        statsd.timing(f"{label}_failed.time", utc_now() - start_time)


@contextmanager
def timed_db_session(repository, statsd, label, session_type=DbSessionType.WRITER):
    with timed_block(statsd, label):
        with DbSessionType.get_session(repository, session_type) as session:
            yield session


class MetadataIngestion(object):
    def __init__(self):
        self.logger = None
        self.aws_region = None
        self.dlq_name = None
        self.invalid_file_bucket = None
        self.dlq_msg_pack_size = None
        self.repository = None
        self.s3_client = None
        self.sqs_client = None
        self.statsd = None
        self.statsd_tcp = None
        self.curr_cust_device_link = None
        self.next_cust_device_link = None
        self.statsd_prefix = "tic.ubs.ingestion"

        self.metadata_internal_version = 1
        self._set_up_clients()

    def _set_up_clients(self):
        self.dlq_name = environ["dlq_name"]
        self.aws_region = environ["AWS_DEFAULT_REGION"]
        self.boto_session = boto3.Session(region_name=self.aws_region)
        self.logger = logging.getLogger()
        self.logger.setLevel(logging.INFO)

        # parallelize inserts of metadata, device/filepath/signatures, s3 put of invalid does not need a repository
        self.repository = self._get_ubs_repository()

        self.fips_s3_endpoint = (
            "https://" + environ["AWS_ENDPOINT_OVERRIDE_S3"]
            if "AWS_ENDPOINT_OVERRIDE_S3" in os.environ
            else None
        )
        self.fips_sqs_endpoint = (
            "https://" + environ["AWS_ENDPOINT_OVERRIDE_SQS"]
            if "AWS_ENDPOINT_OVERRIDE_SQS" in os.environ
            else None
        )
        self.logger.info(f"fips s3 endpoint: {self.fips_s3_endpoint}")
        self.logger.info(f"fips sqs endpoint: {self.fips_sqs_endpoint}")
        self.s3_client = S3Client(
            boto_session=self.boto_session, endpoint_url=self.fips_s3_endpoint
        )
        self.sqs_client = SqsClient(
            self.boto_session, endpoint_url=self.fips_sqs_endpoint
        )
        self.metadata_internal_version = int(environ["metadata_internal_version"])
        self.dlq_msg_pack_size = int(environ["dlq_msg_pack_size"])
        self.invalid_file_bucket = environ["invalid_events_bucket"]

        is_govcloud = True if "is_govcloud" in os.environ else False
        default_statsd_port = (
            environ["default_statsd_port"]
            if "default_statsd_port" in os.environ
            else None
        )
        self.statsd = StatsdClient(
            prefix=self.statsd_prefix,
            is_govcloud=is_govcloud,
            port=default_statsd_port,
            logger=self.logger,
        )

    def _get_ubs_repository(self):
        return UbsDbRepository(
            self.logger,
            build_postgresql_connection_string(
                environ["ubs_db_writer_host"],
                environ["ubs_db_port"],
                environ["ubs_db_db_name"],
                environ["ubs_db_user"],
                environ["ubs_db_pw"],
                environ["ubs_db_pw_encrypted"],
                self.aws_region,
            ),
            build_postgresql_connection_string(
                environ["ubs_db_reader_host"],
                environ["ubs_db_port"],
                environ["ubs_db_db_name"],
                environ["ubs_db_user"],
                environ["ubs_db_pw"],
                environ["ubs_db_pw_encrypted"],
                self.aws_region,
            ),
            connection_args={
                "sslmode": "verify-ca",
                "sslrootcert": "/tmp/root.crt",
                "application_name": "metadata_ingestion",
            },
            statement_timeout=int(environ["statement_timeout_ms"]),
        )

    def _get_cust_device_link_tables(self):
        with self.repository.ubs_reader_session() as reader:
            (
                self.curr_cust_device_link,
                self.next_cust_device_link,
            ) = self.repository.get_cust_device_link_tables(reader)

    def main(self, records, err_handler):
        """
        Run this "service".
        Ingests already deduped metadata from dedupe Kinesis stream, output to UBS DB

        @param records: the metadata provided for ingestion
        """
        self._get_cust_device_link_tables()

        (ingestion_objects, e_non_replayable) = self._prep_records(records)

        # send mal-formatted records to DLQ
        if len(e_non_replayable.data) > 0:
            self.logger.exception("mal-formatted data caught when preparing records")
            err_handler.send_lambda_exception_to_dlq(e_non_replayable.data)
            self.statsd.incr("lambda.handler.exception.non_retryable.invocation", 1)
            self.statsd.incr(
                "lambda.handler.exception.non_retryable.records",
                len(e_non_replayable.data),
            )

        insertion_objects, e = self._parse_ingestion_objects(ingestion_objects)

        valid_file_index = insertion_objects["file_indices"]["valid"]
        invalid_file_index = insertion_objects["file_indices"]["invalid"]
        if len(invalid_file_index) > 0:
            self.logger.warning(f"send {len(invalid_file_index)} invalid records to s3")
            start_time = utc_now()
            self._upload_invalid_to_s3(
                self.s3_client,
                self.invalid_file_bucket,
                invalid_file_index,
                self.statsd,
            )
            self.statsd.timing("invalid.file.index.upload.time", utc_now() - start_time)

        if len(valid_file_index) == 0:
            if len(e.data) > 0:
                self.logger.exception(
                    "exceptions caught upon parsing records, throw to DLQ"
                )
                err_handler.send_lambda_exception_to_dlq(e.data)
                self.statsd.incr("lambda.handler.exception.non_retryable.invocation", 1)
                self.statsd.incr(
                    "lambda.handler.exception.non_retryable.records", len(e.data)
                )
            return

        # indices have to exist before inserting other data
        with timed_db_session(
            self.repository, self.statsd, "insert.file.index.execution"
        ) as writer:
            self.repository.add_file_indexes(writer, valid_file_index)

        self._insert_device_specific_data(
            self.repository, insertion_objects, self.statsd
        )

        self._insert_static_metadata(
            self.repository, insertion_objects["metadata"], self.statsd
        )

        # after all insertions of valid data, send non-replayble records to dlq.
        if len(e.data) > 0:
            self.logger.exception("throw lambda exceptions to DLQ")
            err_handler.send_lambda_exception_to_dlq(e.data)
            self.statsd.incr("lambda.handler.exception.non_retryable.invocation", 1)
            self.statsd.incr(
                "lambda.handler.exception.non_retryable.records", len(e.data)
            )

    @staticmethod
    def _upload_invalid_to_s3(s3_client, s3_bucket, invalid_file_index, statsd_client):
        invalid_records_in_batch = ""
        for record in invalid_file_index:
            invalid_records_in_batch += record + "\n"

        upload_obj = BytesIO(invalid_records_in_batch.encode())

        key_name = f'invalid_file_index/{utc_now().strftime("%Y/%m/%d/%H/%M/") + to_iso_8601(utc_now())}'

        s3_client.upload_fileobj(
            upload_obj, s3_bucket, key_name, storage_class=S3StorageClass.STANDARD_IA
        )

        statsd_client.incr("invalid_file_index.invalid_input", len(invalid_file_index))

    @staticmethod
    def _insert_static_metadata(repository, sha256_metadata_map, statsd):
        with timed_db_session(
            repository, statsd, "get.metadata.version.execution", DbSessionType.READER
        ) as reader:
            metadata_in_ubs = repository.get_metadata_version_bulk(
                list(sha256_metadata_map.keys()), reader
            )

        static_metadata_dtos = list()
        for (metadata_sha256, metadata_object) in sha256_metadata_map.items():
            if (
                metadata_sha256 not in metadata_in_ubs
                or metadata_object.internal_version
                > metadata_in_ubs[metadata_sha256]["internal_version"]
                or metadata_object.policy_version
                > metadata_in_ubs[metadata_sha256]["policy_version"]
            ):
                static_metadata_dtos.append(metadata_object)

        with timed_db_session(
            repository, statsd, "static.metadata.insertion"
        ) as writer:
            if len(static_metadata_dtos) > 0:
                repository.add_static_metadata(static_metadata_dtos, writer)
                statsd.incr(
                    "static.metadata.insertion.record", len(static_metadata_dtos)
                )
                statsd.incr("static.metadata.insertion.batch")

    def _insert_device_specific_data(self, repository, insertion_objects, statsd):
        # inserts devices, filepaths, signatures in that order. All must be done together because
        # in the same transaction, if we fail we do not end up with duplicates in any table.
        with timed_db_session(repository, statsd, "insert_device.execution") as writer:
            # returns any hashes already seen on devices.
            with timed_block(statsd, "device.insertion"):
                excluded_hashes = set(
                    repository.add_devices(
                        list(insertion_objects["devices"]),
                        writer,
                        TableCustDeviceLink(self.curr_cust_device_link),
                    )
                )

            with timed_block(statsd, "next_device.insertion"):
                # update next/future cust_device_link table
                repository.add_devices(
                    list(insertion_objects["devices"]),
                    writer,
                    TableCustDeviceLink(self.next_cust_device_link),
                )

            # skip if exclude (seen on this device before)
            paths = [
                path
                for path in insertion_objects["file_paths"].values()
                if path.sha256 not in excluded_hashes
            ]

            if paths:
                with timed_block(statsd, "file.path.insertion"):
                    repository.add_file_path(writer, paths)

            # insert signatures if there are any
            batch_signature_pairs = insertion_objects["signatures"]
            if batch_signature_pairs:
                # This can be a list instead of a set because the dict already deduped it.
                # Makes testing easier because order.
                dtos = list()
                # TODO: added same excluded_hashes check to make these operate the same way, but adopts the same concern
                for pair in batch_signature_pairs.values():
                    if pair["dto"].sha256 not in excluded_hashes:
                        pair["dto"].num_occurences = len(pair["ids"])
                        dtos.append(pair["dto"])
                if dtos:
                    with timed_block(statsd, "add.signature.execution"):
                        repository.add_signatures(writer, dtos)

    def _prep_records(self, records):
        ingestion_objects = []
        e = LambdaException()
        self.logger.debug(f"number of records to prep: {len(records)}")
        for rec in records:
            try:
                record = IngestionObject.from_dict(rec, True, True)

                if record.event.process_op is None and record.event.modload_op is None:
                    raise Exception("Record has no operation details")

                ingestion_objects.append(record)
            except Exception as ex:
                e.add(str(ex), traceback.format_exc(), rec, None)

        return ingestion_objects, e

    def _parse_ingestion_objects(self, ingestion_objects):
        # Set up for the finally-- return empty for anything that's not assigned.
        # Fixes case where only record in batch of 1 has something invalid.
        file_index_dtos = list()
        batch_signature_pairs = dict()
        device_dtos = set()
        sha256_with_metadata = dict()
        file_paths = dict()
        process_op_record_count = 0
        modload_op_record_count = 0
        invalid_file_index_count = 0
        valid_file_index_count = 0

        batch_sha256s = set()
        e = LambdaException()
        invalid_file_index = list()

        for ingestion_object in ingestion_objects:
            try:
                if ingestion_object.event.process_op is not None:
                    main_module = ingestion_object.event.process_op.target.main_module
                    process_op_record_count += 1
                elif ingestion_object.event.modload_op is not None:
                    # target is the file info in modload, for some reason.
                    main_module = ingestion_object.event.modload_op.target
                    modload_op_record_count += 1

                # validate langId and charsetId
                if main_module.file_metadata:
                    if main_module.file_metadata.lang_id:
                        if main_module.file_metadata.lang_id > MAX_LANG_CHAR_ID:
                            main_module.file_metadata.lang_id = -1
                            self.logger.warning("langId is out of range")
                            invalid_file_index_count += 1
                    else:
                        main_module.file_metadata.lang_id = -1
                        self.logger.warning("langId is missing")
                        invalid_file_index_count += 1

                    if main_module.file_metadata.charset_id:
                        if main_module.file_metadata.charset_id > MAX_LANG_CHAR_ID:
                            main_module.file_metadata.charset_id = -1
                            self.logger.warning("charsetId is out of range")
                            invalid_file_index_count += 1
                    else:
                        main_module.file_metadata.charset_id = -1
                        self.logger.warning("charsetId is missing")
                        invalid_file_index_count += 1

                # should only attempt to insert one copy of file_index for each sha256
                if main_module.sha256 not in batch_sha256s:
                    # check MD5
                    if main_module.md5 is None:
                        invalid_file_index.append(
                            json.dumps(
                                ingestion_object.to_dict(True, True),
                                separators=(",", ":"),
                                ensure_ascii=False,
                            )
                        )
                        self.logger.warning("invalid record: md5 is missing")
                        invalid_file_index_count += 1
                        continue

                    # verify data validity
                    try:
                        IndexCustomerMetadata(
                            main_module.sha256,
                            main_module.md5,
                            main_module.file_size,
                            ingestion_object.device_meta.org_key,
                        ).to_tuple()
                    except Exception as ex:
                        invalid_file_index.append(
                            json.dumps(
                                ingestion_object.to_dict(True, True),
                                separators=(",", ":"),
                                ensure_ascii=False,
                            )
                        )
                        self.logger.error(f"invalid record found - {str(ex)}")
                        invalid_file_index_count += 1
                        continue

                    # should only attempt to insert one copy of file_index for each sha256
                    file_index_dtos.append(
                        IndexCustomerMetadata(
                            main_module.sha256,
                            main_module.md5,
                            main_module.file_size,
                            ingestion_object.device_meta.org_key,
                        )
                    )
                    batch_sha256s.add(main_module.sha256)
                    valid_file_index_count += 1

                # update signature
                self._update_signature_dtos(
                    ingestion_object, main_module, batch_signature_pairs
                )

                # update device
                self._add_device_dto(ingestion_object, main_module, device_dtos)

                # update metadata
                self._add_static_metadata_dto(
                    ingestion_object, main_module, sha256_with_metadata
                )

                # update file path
                self._add_filepath_dto(ingestion_object, main_module, file_paths)

            except Exception as ex:
                e.add(
                    str(ex),
                    traceback.format_exc(),
                    json.dumps(
                        ingestion_object.to_dict(True, True),
                        separators=(",", ":"),
                        ensure_ascii=False,
                    ),
                    None,
                )
                continue

        self.statsd.incr("process.op.record", process_op_record_count)
        self.statsd.incr("modload.op.record", modload_op_record_count)
        self.statsd.incr("file.path.record", len(file_paths))
        self.statsd.incr("signature.record", len(batch_signature_pairs))
        self.statsd.incr("valid.file.index.record", valid_file_index_count)
        self.statsd.incr("invalid.file.index.record", invalid_file_index_count)
        self.statsd.incr("insertion_object_batch")

        return (
            {
                "file_indices": {
                    "valid": file_index_dtos,
                    "invalid": invalid_file_index,
                },
                "signatures": batch_signature_pairs,
                "devices": device_dtos,
                "metadata": sha256_with_metadata,
                "file_paths": file_paths,
            },
            e,
        )

    @staticmethod
    def _update_signature_dtos(record, main_module, batch_signature_pairs):
        for signature in main_module.digital_signatures:
            signature_data = SignatureData(
                sha256=main_module.sha256,
                org_key=record.device_meta.org_key,
                publisher_name=signature.publisher,
                issuer_name=signature.issuer,
                first_seen=record.event.header.device_timestamp,
                sign_timestamp=signature.sign_timestamp,
                sign_state=signature.sign_state,
                num_occurrences=1,
            )

            pair_id = hash(signature_data)
            if pair_id in batch_signature_pairs:
                batch_signature_pairs[pair_id]["ids"].add(record.device_meta.device_id)
            else:
                batch_signature_pairs[pair_id] = {
                    "ids": {record.device_meta.device_id},
                    "dto": signature_data,
                }

    @staticmethod
    def _add_device_dto(record, main_module, device_dtos):
        # Note: we don't have to provide last_seen, now() is default for new devices' entries
        device_dtos.add(
            DeviceInsert(
                sha256=main_module.sha256,
                org_key=record.device_meta.org_key,
                device_id=record.device_meta.device_id,
                device_name=record.device_meta.device_name,
                first_seen=record.event.header.device_timestamp,
            )
        )

    @staticmethod
    def _add_filepath_dto(record, main_module, file_paths):
        file_path = FilePathInsert(
            main_module.sha256,
            record.device_meta.org_key,
            main_module.file_path,
            1,
            record.event.header.device_timestamp,
        )

        if hash(file_path) in file_paths:
            file_paths[hash(file_path)].num_occurences += 1
        else:
            file_paths[hash(file_path)] = file_path

    def _add_static_metadata_dto(self, record, main_module, sha256_with_metadata):
        policy_version = 0
        if (
            len(record.event.header.triggers) > 0
            and record.event.header.triggers[0].policy_revision is not None
        ):
            policy_version = record.event.header.triggers[0].policy_revision

        metadata = StaticMetadataInsert(
            architecture=main_module.architecture,
            file_version=main_module.file_metadata.file_version,
            lang_id=main_module.file_metadata.lang_id,
            charset_id=main_module.file_metadata.charset_id,
            trademark=main_module.file_metadata.trademark,
            special_build=main_module.file_metadata.special_build,
            private_build=main_module.file_metadata.private_build,
            sha256=main_module.sha256,
            os_type=record.device_meta.os_type,
            internal_name=main_module.file_metadata.internal_name,
            product_name=main_module.file_metadata.product_name,
            company_name=main_module.file_metadata.company_name,
            copyright=main_module.file_metadata.copyright,
            file_description=main_module.file_metadata.file_description,
            comments=main_module.file_metadata.comments,
            original_filename=main_module.file_metadata.original_filename,
            product_description=main_module.file_metadata.product_description,
            product_version=main_module.file_metadata.product_version,
            sensor_version=record.event.header.sensor_version,
            internal_version=self.metadata_internal_version,
            policy_version=policy_version,
        )

        if (
            metadata.sha256 not in sha256_with_metadata
            or metadata > sha256_with_metadata[metadata.sha256]
        ):
            sha256_with_metadata[metadata.sha256] = metadata
