from datetime import datetime
from tic.shared.dto import Dto
from tic.shared.utils.date_util import match_str_strptime_to_datetime, to_iso_8601


class IngestionObject(Dto):
    decode_names = {"deviceMeta": "device_meta"}
    encode_names = {"device_meta": "deviceMeta"}

    decode_settings = {
        "device_meta": lambda x: None
        if x is None
        else DeviceMeta.from_dict(x, True, True),
        "event": lambda x: None if x is None else Event.from_dict(x, True, True),
    }

    def __init__(self, device_meta=None, event=None):
        self.device_meta = device_meta
        self.event = event


class DeviceMeta(Dto):
    decode_names = {
        "orgId": "org_key",
        "deviceId": "device_id",
        "deviceName": "device_name",
        "osType": "os_type",
    }
    encode_names = {
        "org_key": "orgId",
        "device_id": "deviceId",
        "device_name": "deviceName",
        "os_type": "osType",
    }

    def __init__(self, org_key=None, device_id=None, device_name=None, os_type=None):
        self.org_key = org_key
        self.device_id = device_id
        self.device_name = device_name
        self.os_type = os_type


class Event(Dto):
    decode_names = {
        "processOp": "process_op",
        "modloadOp": "modload_op",
    }
    encode_names = {"process_op": "processOp", "modload_op": "modloadOp"}

    decode_settings = {
        "modload_op": lambda x: None
        if x is None
        else ModloadOp.from_dict(x, True, True),
        "process_op": lambda x: None
        if x is None
        else ProcessOp.from_dict(x, True, True),
        "header": lambda x: Header() if x is None else Header.from_dict(x, True, True),
    }

    def __init__(self, process_op=None, modload_op=None, header=None):
        # self.header = header
        self.process_op = process_op
        self.modload_op = modload_op
        self.header = header


class Header(Dto):
    decode_names = {
        "deviceTimestamp": "device_timestamp",
        "sensorVersion": "sensor_version",
    }
    encode_names = {
        "device_timestamp": "deviceTimestamp",
        "sensor_version": "sensorVersion",
    }

    encode_settings = {
        "device_timestamp": lambda x: to_iso_8601(x),
        "triggers": lambda x: None
        if len(x) == 0
        else [y.to_dict(True, True) for y in x],
    }

    decode_settings = {
        "device_timestamp": lambda x: datetime.utcnow()
        if x is None
        else match_str_strptime_to_datetime(x),
        "triggers": lambda x: list()
        if x is None
        else [
            Trigger.from_dict(y, True, True) for y in x if y.get("policyTags") == "UBS"
        ],
    }

    def __init__(self, device_timestamp=None, sensor_version=None, triggers=None):
        self.device_timestamp = device_timestamp
        self.sensor_version = sensor_version
        self.triggers = list() if triggers is None else triggers


class Trigger(Dto):
    decode_names = {"policyRevision": "policy_revision", "policyTags": "policy_tags"}
    encode_names = {"policy_revision": "policyRevision", "policy_tags": "policyTags"}

    decode_settings = {"policy_revision": lambda x: -1 if x is None else x}

    def __init__(self, policy_revision=None, policy_tags=None):
        self.policy_revision = policy_revision
        self.policy_tags = policy_tags


class ProcessOp(Dto):
    decode_settings = {
        "target": lambda x: None if x is None else Target.from_dict(x, True, True)
    }

    def __init__(self, target=None):
        self.target = target


class ModloadOp(Dto):
    decode_settings = {
        "target": lambda x: None if x is None else MainModule.from_dict(x, True, True)
    }

    def __init__(self, target=None):
        self.target = target


class Target(Dto):
    decode_names = {"mainModule": "main_module"}
    encode_names = {"main_module": "mainModule"}

    decode_settings = {
        "main_module": lambda x: None
        if x is None
        else MainModule.from_dict(x, True, True)
    }

    def __init__(self, main_module=None):
        self.main_module = main_module


class MainModule(Dto):
    decode_names = {
        "path": "file_path",
        "md5Hash": "md5",
        "sha256Hash": "sha256",
        "fileMetadata": "file_metadata",
        "digitalSignatures": "digital_signatures",
        "fileSize": "file_size",
    }
    encode_names = {
        "file_path": "path",
        "md5": "md5Hash",
        "sha256": "sha256Hash",
        "file_metadata": "fileMetadata",
        "digital_signatures": "digitalSignatures",
        "file_size": "fileSize",
    }

    encode_settings = {
        "digital_signatures": lambda x: None
        if len(x) == 0
        else [y.to_dict(True, True) for y in x]
    }

    decode_settings = {
        "file_metadata": lambda x: FileMetadata()
        if x is None
        else FileMetadata.from_dict(x, True, True),
        "digital_signatures": lambda x: list()
        if x is None
        else [DigitalSignature.from_dict(y, True, True) for y in x],
    }

    def __init__(
        self,
        file_path=None,
        md5=None,
        sha256=None,
        architecture=None,
        file_size=0,
        file_metadata=None,
        digital_signatures=None,
    ):
        self.file_path = file_path
        self.md5 = md5
        self.sha256 = sha256
        self.architecture = architecture
        self.file_metadata = FileMetadata() if file_metadata is None else file_metadata
        self.digital_signatures = (
            list() if digital_signatures is None else digital_signatures
        )
        self.file_size = file_size


class FileMetadata(Dto):
    decode_names = {
        "internalName": "internal_name",
        "productName": "product_name",
        "companyName": "company_name",
        "fileDescription": "file_description",
        "originalFilename": "original_filename",
        "productDescription": "product_description",
        "productVersion": "product_version",
        "privateBuild": "private_build",
        "specialBuild": "special_build",
        "langId": "lang_id",
        "charsetId": "charset_id",
        "fileVersion": "file_version",
    }
    encode_names = {
        "internal_name": "internalName",
        "product_name": "productName",
        "company_name": "companyName",
        "file_description": "fileDescription",
        "original_filename": "originalFilename",
        "product_description": "productDescription",
        "product_version": "productVersion",
        "private_build": "privateBuild",
        "special_build": "specialBuild",
        "lang_id": "langId",
        "charset_id": "charsetId",
        "file_version": "fileVersion",
    }

    def __init__(
        self,
        internal_name=None,
        product_name=None,
        company_name=None,
        copyright=None,
        trademark=None,
        file_description=None,
        comments=None,
        original_filename=None,
        product_description=None,
        product_version=None,
        private_build=None,
        special_build=None,
        lang_id=None,
        charset_id=None,
        file_version=None,
    ):
        self.internal_name = internal_name
        self.product_name = product_name
        self.company_name = company_name
        self.copyright = copyright
        self.trademark = trademark
        self.file_description = file_description
        self.comments = comments
        self.original_filename = original_filename
        self.product_description = product_description
        self.product_version = product_version
        self.private_build = private_build
        self.special_build = special_build
        self.lang_id = lang_id
        self.charset_id = charset_id
        self.file_version = file_version


class DigitalSignature(Dto):
    decode_names = {"signTimestamp": "sign_timestamp", "state": "sign_state"}
    encode_names = {"sign_timestamp": "signTimestamp", "sign_state": "state"}

    encode_settings = {
        "sign_timestamp": lambda x: None if x is None else to_iso_8601(x)
    }

    decode_settings = {
        "sign_timestamp": lambda x: None
        if x is None
        else match_str_strptime_to_datetime(x)
    }

    def __init__(self, publisher=None, issuer=None, sign_state=8, sign_timestamp=None):
        self.publisher = publisher
        self.issuer = issuer
        self.sign_state = sign_state
        self.sign_timestamp = sign_timestamp
